/* 
    @(#)MFImplicitMF.h	1.21
    03/07/24 11:20:52
   
    PROGRAM NAME:  Manifold
   
    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.
   
    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   February 19, 1999                     */
/*              October 6, 2004   Added ProjectFromCenter */

#ifndef __MFIMPLICITMF_H__
#define __MFIMPLICITMF_H__
#include <MFBase.h>

#include <MFKVector.h>
#include <MFNVector.h>
#include <MFNKMatrix.h>
#include <MFNSpace.h>
#include <IMFFlow.h>

struct MFImplicitMFSt;
typedef struct MFImplicitMFSt *MFImplicitMF;

void MFRefImplicitMF(MFImplicitMF);
void MFFreeImplicitMF(MFImplicitMF);
char *MFImplicitMFId(MFImplicitMF);
int MFIMF_N(MFImplicitMF);
int MFIMF_K(MFImplicitMF);
int MFIMFProject(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector);
int MFIMFProjectFromCenter(MFImplicitMF,MFNVector,MFNKMatrix,MFKVector,MFNVector);
MFNKMatrix MFIMFTangentSpace(MFImplicitMF,MFNVector);
MFNKMatrix MFIMFTangentSpaceWithGuess(MFImplicitMF,MFNVector,MFNKMatrix);
double MFIMFScale(MFImplicitMF,MFNVector,MFNKMatrix);
MFNSpace MFIMFNSpace(MFImplicitMF);
void MFIMFEvaluate(MFImplicitMF,MFNVector,MFNVector);
void MFIMFApplyJacobian(MFImplicitMF,MFNVector,MFNKMatrix,MFNKMatrix);
void MFIMFApplySecDer(MFImplicitMF,MFNVector,MFNVector,MFNVector,MFNVector);

MFImplicitMF MFIMFCreateAlgebraicExpression(char*,char*);
MFImplicitMF MFIMFCreateAlgebraicExpressionWithRadius(char*,char*,double);
MFImplicitMF MFIMFCreateAlgebraicSubroutine(int,int,void (*)(int*,double*,int*,double*),void (*)(int*,double*,int*,double*));
MFImplicitMF MFIMFCreateAlgebraicSubroutineWithRadius(int,int,void (*)(int*,double*,int*,double*),void (*)(int*,double*,int*,double*),double);
MFImplicitMF MFIMFCreateTorus(double,double,double,double,double);
MFImplicitMF MFIMFCreateSphere(double,double,double,double);
MFImplicitMF MFIMFCreateCircle(double,double,double);
MFImplicitMF MFIMFCreatePendula(int,double,double,int);
MFImplicitMF MFIMFCreatePendulaFour(int,int);
MFImplicitMF MFIMFCreatePlane(void);
MFImplicitMF MFIMFCreateNSpace(int);
MFImplicitMF MFIMFCreateNSpaceWithRadius(int,double);
MFImplicitMF MFIMFCreateCSTR(double,double,double,double,double,double,double,double,double,double,double,double);
MFImplicitMF MFIMFCreateFlat(int,int,double*,double*);
MFImplicitMF MFIMFCreatePolygonIn3SpaceWithRadius(int,double*,double);
MFImplicitMF MFIMFCreateEdgeIn3SpaceWithRadius(double*,double*,double);
MFImplicitMF MFIMFCreateSwallow(void);

void MFWriteImplicitMF(FILE*,MFImplicitMF);
MFImplicitMF MFReadImplicitMF(FILE*);

typedef void (*MFTPBVPFFUNCTION)(double,int,double*,int,double*,double*,double*,double*);
typedef void (*MFTPBVPAFUNCTION)(int,int,double*,double*,int,double*,double*,double*,double*,double*);
typedef void (*MFTPBVPLFUNCTION)(int,double,int,double*,int,double*,double*,double*,double*);
typedef void (*MFTPBVPMFUNCTION)(int,int,double*,double*,double*);
void MFTPBVPSetEpsilon(MFImplicitMF,double);
/*
typedef void (*MFTPBVPAFUNCTION)(int,double*,double*,int,double*,double*,double*,double*,double*);
*/

/*
MFImplicitMF MFIMFCreateTPBVP(int,int,int,int,int,MFTPBVPFFUNCTION,MFTPBVPFFUNCTION,MFTPBVPFFUNCTION,MFTPBVPAFUNCTION,MFTPBVPAFUNCTION,MFTPBVPAFUNCTION,MFTPBVPLFUNCTION,MFTPBVPLFUNCTION,MFTPBVPLFUNCTION,MFTPBVPMFUNCTION,MFTPBVPMFUNCTION);
*/
MFImplicitMF MFIMFCreateTPBVP(int,int,int,int,MFTPBVPFFUNCTION,MFTPBVPFFUNCTION,MFTPBVPFFUNCTION,int,MFTPBVPAFUNCTION,MFTPBVPAFUNCTION,MFTPBVPAFUNCTION,int,MFTPBVPLFUNCTION,MFTPBVPLFUNCTION,MFTPBVPLFUNCTION,MFTPBVPMFUNCTION,MFTPBVPMFUNCTION);
MFNVector MFTPBVPIntegrateForInitialSolution(MFImplicitMF,double*,double*,double*);
MFNVector MFTPBVPIntegrateForTangent(MFImplicitMF,MFNVector,double*,double*);
int MFIMFStop(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix);

int MFIMFProjectToSave(MFImplicitMF,MFNVector,double*);
int MFIMFProjectToDraw(MFImplicitMF,MFNVector,double*);
int MFIMFProjectToBB(MFImplicitMF,MFNVector,double*);

MFImplicitMF MFIMFCreateBaseClass(int,int,char*);
void MFIMFSetData(MFImplicitMF,void*);
void *MFIMFGetData(MFImplicitMF);
void MFIMFSetWriteData(MFImplicitMF,void (*)(FILE*,void*));
void MFIMFSetFreeData(MFImplicitMF,void (*)(void*));
void MFIMFSetProjectFromCenter(MFImplicitMF,int (*)(int,int,MFNVector,MFNKMatrix, MFKVector,MFNVector,void*,int*));
void MFIMFSetProject(MFImplicitMF,int (*)(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*));
void MFIMFSetTangent(MFImplicitMF,int (*)(int,int,MFNVector,MFNKMatrix,void*));
void MFIMFSetTangentWithGuess(MFImplicitMF,int (*)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*));
void MFIMFSetEvaluate(MFImplicitMF,void (*)(int,MFNVector,MFNVector,void*));
void MFIMFSetApplyJacobian(MFImplicitMF,void (*)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*));
void MFIMFSetApplySecDer(MFImplicitMF,void (*)(int,int,MFNVector,MFNVector,MFNVector,MFNVector,void*));
void MFIMFSetScale(MFImplicitMF,double (*)(int,int,MFNVector,MFNKMatrix,void*));
void MFIMFSetStop(MFImplicitMF,int (*)(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*));
void MFIMFSetR(MFImplicitMF,double);
double MFIMFGetR(MFImplicitMF);
void MFIMFSetRMin(MFImplicitMF,double);
double MFIMFGetRMin(MFImplicitMF);
void MFIMFSetProjectForSave(MFImplicitMF,int (*)(MFNVector,double*,void*));
void MFIMFSetProjectForDraw(MFImplicitMF,int (*)(MFNVector,double*,void*));
void MFIMFSetProjectForBB(MFImplicitMF,int (*)(MFNVector,double*,void*));
void MFIMFSetData(MFImplicitMF,void*);
void *MFIMFGetData(MFImplicitMF);

void MFIMFSetSingular(MFImplicitMF,int (*)(int,int,MFNVector,MFNKMatrix,MFNVector,void*));
int MFIMFSingular(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector);

void MFIMFSetSetStability(MFImplicitMF,void (*)(MFImplicitMF,MFNVector,MFNKMatrix,void*));
void MFIMFSetStability(MFImplicitMF,MFNVector,MFNKMatrix);

MFImplicitMF MFIMFCreateOrthogonalFoliation(IMFFlow);

#endif
