/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i;
  int n,k;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFContinuationMethod H;

  k=1;
  M=MFIMFCreateNSpaceWithRadius(k,.1);
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,.5);

  u0=MFCreateNVector(n);
  for(i=0;i<k;i++)
   {
    MFNVSetC(u0,i, 0.);
   }

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.05);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,0);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"Line");

  S=MFComputeAtlas(H,M,Omega,u0);

/* Use this is Dump to Plotfile is set to zero */
/*
  fid=fopen("Line.atlas","w");
  MFWriteAtlas(fid,S);
  fclose(fid);
  printf("Done writing Atlas\n");fflush(stdout);
*/

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
