/* 
    @(#)ComputeSpherePacking.c	1.8
    03/02/28 15:20:59
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  char *vars;
  char *expr;
  char *digits[]={"0","1","2","3","4","5","6","7","8","9"};
  char name[80]="";
  MFContinuationMethod H;

  n=4;
  if(argc>1)sscanf(argv[1],"%d",&n);
  if(n<1||n>9)
   {
    fprintf(stderr,"%s currently only works for 0<n<10. You supplied n=%d\n",n);
    fflush(stderr);
    return 12;
   }
  vars=(char*)malloc((2+3*n)*sizeof(char));
  strcpy(vars,"[");
  for(i=0;i<n;i++)
   {
    if(i>0)strcat(vars,",");
    strcat(vars,"v");
    strcat(vars,digits[i]);
   }
  strcat(vars,"]");

  expr=(char*)malloc((5+6*n)*sizeof(char));
  strcpy(expr,"[");
  for(i=0;i<n;i++)
   {
    if(i>0)strcat(expr,"+");
    strcat(expr,"v");
    strcat(expr,digits[i]);
    strcat(expr,"**2");
   }
  strcat(expr,"-1.]");

  printf("Variables : %s\n",vars);
  fflush(stdout);
  printf("Expression: %s\n",expr);
  fflush(stdout);

  M=MFIMFCreateAlgebraicExpression(vars,expr);
  MFIMFSetR(M,sqrt(3.)/2.1);
  Omega=MFNRegionCreateHyperCube(n,3.);

  u0=MFCreateNVector(n);
  MFNVSetC(u0,0, 1.);
  for(i=1;i<n;i++)MFNVSetC(u0,i, 0.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,1.);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,0);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  strcpy(name,"SpherePacking");
  strcat(name,digits[n]);
  MFHendersonsMethodSetFilename(H,name);

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  printf("Done computing manifold\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);
  free(vars);
  free(expr);

  return(0);
 }
