#include <IMF.h>
#include <IMFExpansion.h>
#include <IMFFlow.h>
#include <IMFFixedPt.h>
#include <math.h>
#include <IMFLorenz.h>
#include <IMFSphereOnExpansion.h>
#include <IMFExpansionSpace.h>
#include <IMFIntegrateFat.h>
#include <IMFInterpolation.h>
#include <stdio.h>
#include <MFAtlas.h>
#include <MFChart.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

extern int IMFNInterper;
extern FILE *IMFInterper;
extern int IMFNInterpee;
extern FILE *IMFInterpee;
extern int IMFNInterpT;
extern FILE *IMFInterpT;
extern int IMFNCircle;
extern FILE *IMFCircle;
extern int IMFNTraj;
extern FILE *IMFTraj;
FILE *IMFInterp;
int IMFNInterp;

/* #define DOINTERPANIM
   #define DOTRAJ */

void MFAtlasPageOutChartsNotNearBoundary(MFAtlas,int,int,char*);
void MFAtlasSetNearRtn(MFAtlas,int (*)(MFAtlas,MFChart,MFChart));
int IMFIsNear(MFAtlas,MFChart,MFChart);
MFChart MFAtlasChart(MFAtlas,int);

double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

MFAtlas IMFComputeStableInvariantManifold(IMFFlow L,char *name, MFNVector u0,MFNRegion Omega, double eps, double dt, double tmax, int maxInterp, int maxCharts, double R0, int nSkip)
 {
  IMFFlow F;
  MFAtlas A;

  F=IMFCreateBackwardFlow(L);
  A=IMFComputeUnstableInvariantManifold(F,name,u0,Omega,eps,dt,tmax,maxInterp,maxCharts,R0,nSkip);
  IMFFreeFlow(F);
  return A;
 }

MFAtlas IMFComputeUnstableInvariantManifold(IMFFlow F,char *name, MFNVector u0,MFNRegion Omega, double eps, double dt, double tmax, int maxInterp, int maxCharts, double R0, int nSkip)
 {
  IMFExpansion u,a;
  double s0[2],s1[2];
  MFKVector s;
  MFNVector ug,v;
  MFNVector ustar;
  MFNKMatrix Phi0;
  MFNKMatrix Phi;
  MFNKMatrix Psi;
  MFImplicitMF c;
  int i,j;
  IMFExpansion cE,U;
  MFContinuationMethod H;
  MFAtlas A;
  MFAtlas I;
  MFImplicitMF M;
  double R,r;
  char sname[1024];
  FILE *fid;
  MFNKMatrix TS;
  MFNVector ui;
  MFNVector ut;
  double t;
  int chart;
  MFNVector sigma;
  MFKVector zero;
  int n,k;

  Phi=IMFGetBasisForUnstableInvariantSubspace(u0,F);
  Psi=IMFGetBasisForOrthogonalComplement(Phi);

  n=MFNKMatrixN(Phi);
  k=MFNKMatrixK(Phi);
  printf("n=%d, k=%d\n",n,k);fflush(stdout);
  u=IMFCreateExpansion(n,k);
  a=IMFCreateExpansion(k,k);
  IMFFindExpansionNearFixedPt(u0,Phi,Psi,F,u,a);
  zero=MFCreateKVector(k-1);
  printf("Expansion at fixed point:\n");IMFPrintExpansion(stdout,u);fflush(stdout);

  R=IMFExpansionR(u,eps);if(R!=R||R>R0)R=R0;
  r=.5*R;
/* For U1!    R=R0; r=.05; */
  c=IMFCreateSphereOnExpansion(u,F,eps,R,r);

  s =MFCreateKVector(2);
  MFKVSetC(s,0,R);
  MFKVSetC(s,1,0.);

  ug=MFCreateNVector(n);
  IMFEvaluateExpansion(u,s,ug);

  ut=MFCreateNVector(n+k);
  for(i=0;i<n;i++)MFNVSetC(ut,i,MFNV_C(ug,i));
  for(i=0;i<k;i++)MFNVSetC(ut,n+i,MFKV_C(s,i));
  MFFreeNVector(ug);

  ustar=MFCreateNVector(n+k);

  Phi0=MFIMFTangentSpace(c,ut);
  if(MFIMFProject(c,ut,Phi0,ustar))
   {

    H=MFCreateHendersonsMethod();
    MFHendersonsMethodSetEpsilon(H,.8);
    MFHendersonsMethodSetMaxCharts(H,-1);
    MFHendersonsMethodSetVerbose(H,1);
    MFHendersonsMethodSetPage(H,0);
    MFHendersonsMethodSetDumpToPlotFile(H,0);
    MFHendersonsMethodSetDumpToCenterFile(H,0);
    MFHendersonsMethodSetFilename(H,"SphereOnU0");

    A=MFComputeAtlas(H,c,Omega,ustar);
    MFFreeNVector(ut);
    MFCloseAtlas(H,A);
    MFFreeHendersonsMethod(H);
    printf("Done computing Atlas\n");fflush(stdout);
   }else{
    printf("Projection of initial point failed\n");fflush(stdout);
    return (MFAtlas)NULL;
   }

  M=IMFCreateFlat(n,k);
  I=MFCreateAtlas(M);
  MFAtlasSetNearRtn(I,IMFIsNear);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetFilename(H,name);

  TS=IMFExpansionTS(u);
  ui=IMFCreateExpansionNVector(u,-100.,u0,-1,1);
  MFAtlasAddChartWithAll(I,ui,TS,R);
  printf("0) Fixed Point. ");MFPrintNVector(stdout,ui);printf(", R=%lf\n",R);fflush(stdout);
  MFFreeNVector(ui);
  MFFreeNKMatrix(TS);

  s0[0]=-r; s1[0]= r;
  s0[1]=-r; s1[1]= r;
  for(i=0;i<MFAtlasNumberOfCharts(A);i++)
   {
    cE=IMFSphereOnExpansionGetLocal(A,MFAtlasCenterOfChart(A,i));
    U=IMFInflateExpansionWithFlow(cE,F);

    ui=IMFCreateExpansionNVector(U,0.,MFAtlasCenterOfChart(A,i),-1,1);
    IMFExpansionNVSetChart0(ui,i);
    IMFExpansionNVSetS0(ui,zero);
    printf("*** Point on c\n");fflush(stdout);
    MFPrintNVector(stdout,MFAtlasCenterOfChart(A,i));printf("\n");fflush(stdout);
    IMFExtendAtlasAlongFatTraj(I,F,ui,dt,0.,tmax,eps,Omega,maxCharts,maxCharts,1.,nSkip);
    nSkip=0;
    if(MFAtlasNumberOfCharts(I)%1000==0){printf("Paging\n");MFAtlasPageOutChartsNotNearBoundary(I,1,0,name);}
    MFFreeNVector(ui);

    IMFFreeExpansion(cE);
    IMFFreeExpansion(U);
    if(MFAtlasNumberOfCharts(I)>maxCharts)
     {
      printf("Maximum number of charts exceeded (%d>%d).\n",MFAtlasNumberOfCharts(I),maxCharts);fflush(stdout);
      goto DoneCover;
     }
   }

  printf("*** Done covering c\n");fflush(stdout);
#ifdef DOINTERPOLATIONPT
  sprintf(sname,"%s.intpt",name);
  IMFInterp=fopen(sname,"w");
  IMFNInterp=0;
#endif

  i=0;
  while(i<maxInterp&& (ui=IMFGetInterpolationPoint(I,F,A,tmax,(MFNRegion)NULL))!=(MFNVector)NULL && MFAtlasNumberOfCharts(I)<maxCharts)
   {
    printf("*** Interpolated Point %d\n",i);fflush(stdout);
#ifdef DOINTERPOLATIONPT
    for(j=0;j<3;j++)fprintf(IMFInterp," %lf",IMFExpansionU(IMFExpansionNVGetE(ui))[j]);
    fprintf(IMFInterp,"\n");fflush(IMFInterp);
    IMFNInterp++;
#endif
    if(i<maxInterp-1)IMFExtendAtlasAlongFatTraj(I,F,ui,dt,IMFExpansionNVGetT(ui),tmax,eps,Omega,maxCharts,maxCharts,1.,0);
    if(MFAtlasNumberOfCharts(I)%1000==0){printf("Paging\n");MFAtlasPageOutChartsNotNearBoundary(I,1,0,name);}
    MFFreeNVector(ui);
    i++;
    if(MFAtlasNumberOfCharts(I)>maxCharts)
     {
      printf("Maximum number of charts exceeded (%d>%d).\n",MFAtlasNumberOfCharts(I),maxCharts);fflush(stdout);
      goto DoneCover;
     }
   }

  if(i<maxInterp)
    {printf("*** No more interpolation points, total was %d\n",i);fflush(stdout);}
   else
    {printf("*** Max number of interpolation points reached, %d\n",i);fflush(stdout);}

DoneCover:

  MFFreeImplicitMF(c);
  MFFreeKVector(s);

  printf("Flushing invariant manifold\n");
  MFCloseAtlas(H,I);
  MFFreeHendersonsMethod(H);

  MFFreeNKMatrix(Phi0);

  MFFreeNVector(ustar);
  MFFreeNKMatrix(Phi);
  MFFreeNKMatrix(Psi);

  IMFFreeExpansion(u);
  IMFFreeExpansion(a);
  MFFreeImplicitMF(M);
  MFFreeAtlas(A);
  free(zero);

#ifdef DOINTERPOLATIONPT
  if(IMFInterp!=(FILE*)NULL)
   {
    fclose(IMFInterp);

    sprintf(sname,"i%s.dx",name);
    IMFInterp=fopen(sname,"w");
    sprintf(sname,"%s.intpt",name);
    if(IMFNInterp>0)
     {
      fprintf(IMFInterp,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file %s.dx\n",sname,IMFNInterp);
     }else{
      fprintf(IMFInterp,"object \"Vertices\" class array type float rank 1 shape 3 items 1 data follows\n");
      fprintf(IMFInterp,"  0. 0. 0.\n");
     }
    fprintf(IMFInterper,"object \"interpolated points\" class field\n");
    fprintf(IMFInterper,"component \"positions\" value \"Vertices\"\n");
    fclose(IMFInterper);
   }
#endif

#ifdef DOINTERPANIM

  if(IMFInterper!=(FILE*)NULL)
   {
    fclose(IMFInterper);
    IMFInterper=fopen("gInterp1.dx","w");
    if(IMFNInterper>0)
     {
      fprintf(IMFInterper,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file Interp1.dx\n",IMFNInterper);
     }else{
      fprintf(IMFInterper,"object \"Vertices\" class array type float rank 1 shape 3 items 1 data follows\n");
      fprintf(IMFInterper,"  0. 0. 0.\n");
     }
    fprintf(IMFInterper,"object \"interpolated points\" class field\n");
    fprintf(IMFInterper,"component \"positions\" value \"Vertices\"\n");
    fclose(IMFInterper);
   }

  if(IMFInterpee!=(FILE*)NULL)
   {
    fclose(IMFInterpee);
    IMFInterpee=fopen("gInterp2.dx","w");
    if(IMFNInterpee>0)
     {
      fprintf(IMFInterpee,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file Interp2.dx\n",IMFNInterpee);
      fprintf(IMFInterpee,"object \"Lines\" class array type int rank 1 shape 2 items %d data follows\n",IMFNInterpee/2);
      for(i=0;i<IMFNInterpee;i+=2)
      fprintf(IMFInterpee,"   %d %d\n",i,i+1);
     }else{
      fprintf(IMFInterpee,"object \"Vertices\" class array type float rank 1 shape 3 items 2 data follows\n");
      fprintf(IMFInterpee,"  0. 0. 0.\n");
      fprintf(IMFInterpee,"  0. 0. 0.001\n");
      fprintf(IMFInterpee,"object \"Lines\" class array type int rank 1 shape 2 items 1 data follows\n");
      fprintf(IMFInterpee,"   %d %d\n",0,1);
     }
    fprintf(IMFInterpee,"attribute \"ref\" string \"positions\"\n");
    fprintf(IMFInterpee,"attribute \"element type\" string \"lines\"\n");

    fprintf(IMFInterpee,"object \"interpolation points\" class field\n");
    fprintf(IMFInterpee,"component \"positions\" value \"Vertices\"\n");
    fprintf(IMFInterpee,"component \"connections\" value \"Lines\"\n");
    fclose(IMFInterpee);
   }

  if(IMFInterpT!=(FILE*)NULL)
   {
    fclose(IMFInterpT);
    IMFInterpT=fopen("gInterp3.dx","w");
    if(IMFNInterpT>0)
     {
      fprintf(IMFInterpT,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file Interp3.dx\n",IMFNInterpT
);
      fprintf(IMFInterpT,"object \"Lines\" class array type int rank 1 shape 2 items %d data follows\n",IMFNInterpT-1);
      for(i=0;i<IMFNInterpT-1;i++)
      fprintf(IMFInterpT,"   %d %d\n",i,i+1);
     }else{
      fprintf(IMFInterpT,"object \"Vertices\" class array type float rank 1 shape 3 items 2 data follows\n");
      fprintf(IMFInterpT,"  0. 0. 0.\n");
      fprintf(IMFInterpT,"  0. 0. 0.001\n");
      fprintf(IMFInterpT,"object \"Lines\" class array type int rank 1 shape 2 items 1 data follows\n");
      fprintf(IMFInterpT,"   %d %d\n",0,1);
     }
    fprintf(IMFInterpT,"attribute \"ref\" string \"positions\"\n");
    fprintf(IMFInterpT,"attribute \"element type\" string \"lines\"\n");

    fprintf(IMFInterpT,"object \"interpolation points\" class field\n");
    fprintf(IMFInterpT,"component \"positions\" value \"Vertices\"\n");
    fprintf(IMFInterpT,"component \"connections\" value \"Lines\"\n");
    fclose(IMFInterpT);
   }

  if(IMFCircle!=(FILE*)NULL)
   {
    fclose(IMFCircle);
    IMFCircle=fopen("gIMFCircle.dx","w");
    fprintf(IMFCircle,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file IMFCircle.dx\n",IMFNCircle);
    fprintf(IMFCircle,"object \"Lines\" class array type int rank 1 shape 2 items %d data follows\n",IMFNCircle/2);
    for(i=0;i<IMFNCircle;i+=2)
    fprintf(IMFCircle,"   %d %d\n",i,i+1);
    fprintf(IMFCircle,"attribute \"ref\" string \"positions\"\n");
    fprintf(IMFCircle,"attribute \"element type\" string \"lines\"\n");

    fprintf(IMFCircle,"object \"interpolation points\" class field\n");
    fprintf(IMFCircle,"component \"positions\" value \"Vertices\"\n");
    fprintf(IMFCircle,"component \"connections\" value \"Lines\"\n");
    fclose(IMFCircle);
   }
#endif

#ifdef DOTRAJ
  if(IMFTraj!=(FILE*)NULL)
   {
    fclose(IMFTraj);
    IMFTraj=fopen("gIMFTraj.dx","w");
    fprintf(IMFTraj,"object \"Vertices\" class array type float rank 1 shape 3 items %d data file IMFTraj.dx\n",IMFNTraj);
    fprintf(IMFTraj,"object \"Lines\" class array type int rank 1 shape 2 items %d data follows\n",IMFNTraj/2);
    for(i=0;i<IMFNTraj;i+=2)
    fprintf(IMFTraj,"   %d %d\n",i,i+1);
    fprintf(IMFTraj,"attribute \"ref\" string \"positions\"\n");
    fprintf(IMFTraj,"attribute \"element type\" string \"lines\"\n");

    fprintf(IMFTraj,"object \"trajectories\" class field\n");
    fprintf(IMFTraj,"component \"positions\" value \"Vertices\"\n");
    fprintf(IMFTraj,"component \"connections\" value \"Lines\"\n");
    fclose(IMFTraj);
   }
#endif

  return I;
 }

MFAtlas IMFComputeInvariantManifold(IMFFlow F,char *name, MFAtlas c,MFNRegion Omega, double eps, double dt, double tmax, int maxInterp, int maxCharts, double Rmax)
 {
  double s0[2],s1[2];
  double *ddu;
  MFKVector s;
  MFNVector ug,v;
  MFNKMatrix Phi;
  MFNKMatrix Psi;
  int i;
  IMFExpansion cE,U;
  MFContinuationMethod H;
  MFAtlas A;
  MFAtlas I;
  MFImplicitMF M;
  double R,r;
  char sname[1024];
  FILE *fid;
  MFNKMatrix TS;
  MFNVector ui;
  double t;
  int chart;
  MFNVector sigma;
  MFKVector zero;
  double *u0;
  int n,k;

  n=MFAtlasN(c);
  k=MFAtlasK(c);
  M=IMFCreateFlat(n,k+1);
  MFIMFSetR(M,.1);
  I=MFCreateAtlas(M);
  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetFilename(H,name);
  ddu=(double*)malloc(n*sizeof(double));if(ddu==(double*)NULL)abort();
  zero=MFCreateKVector(k-1);

  s0[0]=-r; s1[0]= r;
  s0[1]=-r; s1[1]= r;
  for(i=0;i<n;i++)ddu[0]=0.;
  for(i=0;i<MFAtlasNumberOfCharts(c);i++)
   {
    u0=MFNV_CStar(MFAtlasChartCenter(c,i));
    Phi=MFAtlasChartTangentSpace(c,i);
    cE=IMFCreateExpansion(n,k);
    IMFExpansionSetDerivatives(cE,u0,MFNKM_CStar(Phi),ddu,(double*)NULL); /* Assumes it's a straight line */
    U=IMFInflateExpansionWithFlow(cE,F);

    ui=IMFCreateExpansionNVector(U,0.,MFAtlasCenterOfChart(c,i),-1,1);
    IMFExpansionNVSetChart0(ui,i);
    IMFExpansionNVSetS0(ui,zero);
    printf("*** Point on c\n");fflush(stdout);
    IMFExtendAtlasAlongFatTraj(I,F,ui,dt,0.,tmax,eps,Omega,maxCharts,maxCharts,.2,0);
    if(MFAtlasNumberOfCharts(I)%1000==0)MFAtlasPageOutChartsNotNearBoundary(I,1,0,name);
    MFFreeNVector(ui);

    IMFFreeExpansion(cE);
    IMFFreeExpansion(U);
   }

  printf("*** Done covering c\n");fflush(stdout);

  i=0;
  while(i<maxInterp&& (ui=IMFGetInterpolationPoint(I,F,A,tmax,(MFNRegion)NULL))!=(MFNVector)NULL && MFAtlasNumberOfCharts(I)<maxCharts)
   {
    printf("*** Interpolated Point %d\n",i);fflush(stdout);
    IMFExtendAtlasAlongFatTraj(I,F,ui,dt,IMFExpansionNVGetT(ui),tmax,eps,Omega,maxCharts,maxCharts,Rmax,0);
    if(MFAtlasNumberOfCharts(I)%1000==0)MFAtlasPageOutChartsNotNearBoundary(I,1,0,name);
    MFFreeNVector(ui);
    i++;
   }

  if(i<maxInterp)
    {printf("*** No more interpolation points, total was %d\n",i);fflush(stdout);}
   else
    {printf("*** Max number of interpolation points reached, %d\n",i);fflush(stdout);}

  MFFlushAtlas(H,I);

  MFFreeHendersonsMethod(H);
  MFFreeImplicitMF(M);

  free(ddu);

  return I;
 }

int IMFIsNear(MFAtlas A,MFChart c0,MFChart c1)
 {
  MFNVector u0,u1;
  double t0,t1;

  u0=MFChartCenter(c0);
  u1=MFChartCenter(c1);

  t0=IMFExpansionNVGetT(u0);
  t1=IMFExpansionNVGetT(u1);
/*printf("IMFIsNeart, nCharts=%d, t0=%lf, t1=%lf\n",MFAtlasNumberOfCharts(A),t0,t1);fflush(stdout);*/
  if(fabs(t0-t1)>30. && t0>0. && t1>0. )return 0;

  return 1;
 }

MFAtlas IMFComputeStableInvariantManifold2(IMFFlow L,char *name, MFNVector u0,MFNRegion Omega, double eps, double dt, double tmax, int maxInterp, int maxCharts, double R0, int nSkip)
 {
  IMFFlow F;
  MFAtlas A;

  F=IMFCreateBackwardFlow(L);
  A=IMFComputeUnstableInvariantManifold2(F,name,u0,Omega,eps,dt,tmax,maxInterp,maxCharts,R0,nSkip);
  IMFFreeFlow(F);
  return A;
 }

MFAtlas IMFComputeUnstableInvariantManifold2(IMFFlow F,char *name, MFNVector u0,MFNRegion Omega, double eps, double dt, double tmax, int maxInterp, int maxCharts, double R0, int nSkip)
 {
  IMFExpansion u,a;
  double s0[2],s1[2];
  MFKVector s;
  MFNVector ug,v;
  MFNVector ustar;
  MFNKMatrix Phi0;
  MFNKMatrix Phi;
  MFNKMatrix Psi;
  MFImplicitMF c;
  int i,j,l;
  IMFExpansion cE,U;
  MFContinuationMethod H;
  MFAtlas A;
  MFAtlas I;
  MFImplicitMF M;
  double R,r;
  double Rf,Rmax;
  char sname[1024];
  FILE *fid;
  MFNKMatrix TS;
  MFNVector ui;
  MFNVector ut;
  double t;
  int chart;
  MFNVector sigma;
  int n,k;
  int iring,maxRings;
  double epsilon=1.;

  int nCharts;
  int *chartList=(int*)NULL;
  int mCharts;
  MFChart *cList=(MFChart*)NULL;
  MFChart nchart;
  double dist,d;
  MFKVector zero;

  Rmax=1.;

  Phi=IMFGetBasisForUnstableInvariantSubspace(u0,F);
  Psi=IMFGetBasisForOrthogonalComplement(Phi);

  n=MFNKMatrixN(Phi);
  k=MFNKMatrixK(Phi);
  printf("n=%d, k=%d\n",n,k);fflush(stdout);
  u=IMFCreateExpansion(n,k);
  a=IMFCreateExpansion(k,k);
  IMFFindExpansionNearFixedPt(u0,Phi,Psi,F,u,a);
  zero=MFCreateKVector(k-1);

  R=IMFExpansionR(u,eps);if(R!=R||R>R0)R=R0;
  r=.5*R;
  c=IMFCreateSphereOnExpansion(u,F,eps,R,r);

  s =MFCreateKVector(2);
  MFKVSetC(s,0,R);
  MFKVSetC(s,1,0.);

  ug=MFCreateNVector(n);
  IMFEvaluateExpansion(u,s,ug);

  ut=MFCreateNVector(n+k);
  for(i=0;i<n;i++)MFNVSetC(ut,i,MFNV_C(ug,i));
  for(i=0;i<k;i++)MFNVSetC(ut,n+i,MFKV_C(s,i));
  MFFreeNVector(ug);

  ustar=MFCreateNVector(n+k);

  Phi0=MFIMFTangentSpace(c,ut);
  if(MFIMFProject(c,ut,Phi0,ustar))
   {

    H=MFCreateHendersonsMethod();
    MFHendersonsMethodSetEpsilon(H,.8);
    MFHendersonsMethodSetMaxCharts(H,-1);
    MFHendersonsMethodSetVerbose(H,0);
    MFHendersonsMethodSetPage(H,0);
    MFHendersonsMethodSetDumpToPlotFile(H,0);
    MFHendersonsMethodSetDumpToCenterFile(H,0);
    MFHendersonsMethodSetFilename(H,"SphereOnU0");

    A=MFComputeAtlas(H,c,Omega,ustar);
    MFFreeNVector(ut);
    MFCloseAtlas(H,A);
    MFFreeHendersonsMethod(H);
    printf("Done computing Atlas\n");fflush(stdout);
   }else{
    printf("Projection of initial point failed\n");fflush(stdout);
    return (MFAtlas)NULL;
   }

  M=IMFCreateFlat(n,k);
  I=MFCreateAtlas(M);
/*MFAtlasSetNearRtn(I,IMFIsNear);*/

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetFilename(H,name);

  TS=IMFExpansionTS(u);
  ui=IMFCreateExpansionNVector(u,-100.,u0,-1,1);
  chart=MFAtlasAddChartWithAll(I,ui,TS,R);
  printf("%d) Fixed Point. ",chart);MFPrintNVector(stdout,ui);printf(", R=%lf\n",R);fflush(stdout);
  MFFreeNVector(ui);
  MFFreeNKMatrix(TS);

/* --------------------------------------------------------------------- */

  s0[0]=-r; s1[0]= r;
  s0[1]=-r; s1[1]= r;
  for(i=0;i<MFAtlasNumberOfCharts(A);i++)
   {
    cE=IMFSphereOnExpansionGetLocal(A,MFAtlasCenterOfChart(A,i));
    U=IMFInflateExpansionWithFlow(cE,F);

    ui=IMFCreateExpansionNVector(U,0.,MFAtlasCenterOfChart(A,i),-1,1);
    IMFExpansionNVSetChart0(ui,i);
    IMFExpansionNVSetS0(ui,zero);

    TS=IMFExpansionTS(U);
    R=IMFExpansionR(U,.5*epsilon);
    Rf=IMFFlowR(F,.5*epsilon,ui,TS);
    if(Rf<R)R=Rf;
    if(R>Rmax||R!=R)R=Rmax;
    chart=MFAtlasAddChartWithAll(I,ui,TS,R);
    printf("%d) Point on c ",chart);MFPrintNVector(stdout,ui);printf(", R=%lf\n",R);fflush(stdout);

    MFFreeNKMatrix(TS);
    MFFreeNVector(ui);
    IMFFreeExpansion(cE);
    IMFFreeExpansion(U);
    if(MFAtlasNumberOfCharts(I)>maxCharts)
     {
      printf("Maximum number of charts exceeded (%d>%d).\n",MFAtlasNumberOfCharts(I),maxCharts);fflush(stdout);
      goto DoneCover;
     }
   }

  printf("*** Done covering c\n");fflush(stdout);

/* --------------------------------------------------------------------- */

  i=0;
  iring=0;
  maxRings=1;
  while(iring<maxRings&&i<maxInterp&&MFAtlasNumberOfCharts(I)<maxCharts&&MFAtlasNumberOfChartsWithBoundary(I)>0)
   {
    nCharts=MFAtlasNumberOfChartsWithBoundary(I);
    chartList=(int*)realloc((void*)chartList,nCharts*sizeof(int));if(chartList==(int*)NULL)abort();
    mCharts=0;
    cList=(MFChart*)realloc((void*)cList,nCharts*sizeof(MFChart));if(cList==(MFChart*)NULL)abort();

    for(j=0;j<nCharts;j++)
     chartList[j]=MFAtlasChartWithBoundary(I,j);

    printf("Extend ring %d\n",iring);fflush(stdout);
    for(j=0;j<nCharts;j++)
     {
      ui=MFAtlasChartCenter(I,chartList[j]);
      nSkip=1;
      printf("-------Front point %d/%d\n",j,nCharts);fflush(stdout);
      nchart=IMFStepAlongFatTraj(I,F,ui,MFAtlasChartRadius(I,chartList[j]),dt,eps,Omega,Rmax);

      if(nchart!=(MFChart)NULL)
       {
        cList[mCharts]=nchart;
        mCharts++;
       }

      printf("-------\n");fflush(stdout);
     }
    iring++;

    for(j=0;j<mCharts;j++)
     {
      for(l=0;l<j;l++)
       {
        d=MFNSpaceDistance(MFIMFNSpace(MFAtlasMF(I)),MFChartCenter(cList[j]),MFChartCenter(cList[l]))/MFChartRadius(cList[j]);
        if(l==0||d<dist)dist=d;
       }
      if(1||j==0||dist>1.)
        MFAtlasAddChartWithAll(I,MFChartCenter(cList[j]),MFChartTangentSpace(cList[j]),MFChartRadius(cList[j]));
     }

    for(j=0;j<mCharts;j++)MFFreeChart(cList[j]);

    for(j=0;j<nCharts;j++)
     {
      if(nCharts>1&&!MFChartHasBoundary(MFAtlasChart(I,chartList[j])) )
       {
        if(j!=nCharts-1)chartList[j]=chartList[nCharts-1];
        nCharts--;
        if(j>0)j--;
       }
     }

    while(0&&i<maxInterp&& (ui=IMFGetInterpolationPointOnList(I,F,A,tmax,(MFNRegion)NULL,nCharts,chartList))!=(MFNVector)NULL && MFAtlasNumberOfCharts(I)<maxCharts)
     {
      printf("*** Interpolated Point %d\n",i);fflush(stdout);

      U=IMFExpansionNVGetE(ui);
      TS=IMFExpansionTS(U);
      R=IMFExpansionR(U,epsilon);
      Rf=IMFFlowR(F,epsilon,ui,TS);
      if(Rf<R)R=Rf;
      if(R>Rmax||R!=R)R=Rmax;
      MFAtlasAddChartWithAll(I,ui,TS,R);

      MFFreeNKMatrix(TS);
      IMFFreeExpansion(cE);
      IMFFreeExpansion(U);
      MFFreeNVector(ui);
      if(MFAtlasNumberOfCharts(I)>maxCharts)
       {
        printf("Maximum number of charts exceeded (%d>%d).\n",MFAtlasNumberOfCharts(I),maxCharts);fflush(stdout);
        goto DoneCover;
       }
      i++;

      for(j=0;j<nCharts;j++)
       {
        if(nCharts>1&&!MFChartHasBoundary(MFAtlasChart(I,chartList[j])) )
         {
          if(j!=nCharts-1)chartList[j]=chartList[nCharts-1];
          nCharts--;
          if(j>0)j--;
         }
       }
     }
   }

  if(i<maxInterp)
    {printf("*** No more boundary, total was %d interpolated, %d charts\n",i,MFAtlasNumberOfCharts(I));fflush(stdout);}
   else
    {printf("*** Max number of interpolation points reached, %d\n",i);fflush(stdout);}

/* --------------------------------------------------------------------- */

DoneCover:

  MFFreeImplicitMF(c);
  MFFreeKVector(s);

  printf("Flushing invariant manifold\n");
  MFCloseAtlas(H,I);
  MFFreeHendersonsMethod(H);

  MFFreeNKMatrix(Phi0);

  MFFreeNVector(ustar);
  MFFreeNKMatrix(Phi);
  MFFreeNKMatrix(Psi);

  IMFFreeExpansion(u);
  IMFFreeExpansion(a);
  MFFreeImplicitMF(M);
  MFFreeAtlas(A);
  free(chartList);
  free(cList);
  free(zero);

  return I;
 }
