/* 
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   September 15, 2003                    */

#include <MFImplicitMF.h>
#include <IMFExpansionSpace.h>
#include <IMFExpansion.h>
#include <stdlib.h>
#include <string.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

static char MFFlatMFErrorMsg[256]="";

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFSetError(int,char*,char*,int,char*);

void MFFreeNSpaceData(void*);
int MFProjectNSpace(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentNSpace(int,int,MFNVector,MFNKMatrix,void*);
double MFScaleNSpace(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteNSpaceData(FILE*,void*);
MFImplicitMF MFReadNSpaceMF(FILE*);

int MFFlatMFProjectToSave(MFNVector,double*,void*);
int MFFlatMFProjectToDraw(MFNVector,double*,void*);
int MFFlatMFProjectForBB(MFNVector,double*,void*);

struct IMFFlatData
 {
  int n;
  int k;
 };

MFImplicitMF IMFCreateFlat(int n, int k)
 {
  static char RoutineName[]={"IMFCreateFlat"};
  MFImplicitMF nSpace;
  struct IMFFlatData *data;
  MFNSpace space;

  printf("%s, n=%d, k=%d\n",RoutineName,n,k);
  nSpace=MFIMFCreateBaseClass(n,k,"IMFFlat");

  space=IMFCreateExpansionSpace(n);
  MFIMFSetSpace(nSpace,space);
  MFFreeNSpace(space);

  MFIMFSetFreeData(nSpace,MFFreeNSpaceData);
  MFIMFSetProjectForSave(nSpace,MFFlatMFProjectToSave);
  MFIMFSetProjectForDraw(nSpace,MFFlatMFProjectToDraw);
  MFIMFSetProjectForBB(nSpace,MFFlatMFProjectForBB);

  data=(struct IMFFlatData*)malloc(sizeof(struct IMFFlatData));
  if(data==(struct IMFFlatData*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct IMFFlatData));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }
  data->n=n;
  data->k=k;
  MFIMFSetData(nSpace,(void*)data);
  printf("done %s\n",RoutineName);

  return nSpace;
 }

void MFFreeNSpaceData(void *d)
 {
  static char RoutineName[]={"MFFreeNSpaceData"};
  struct IMFFlatData *data;

  data=(struct IMFFlatData*)d;
  if(data!=(struct IMFFlatData*)NULL)free(data);
  return;
 }

int MFFlatMFProjectToSave(MFNVector u, double *x, void *d)
 {
  int i;
  struct IMFFlatData *data;

  data=(struct IMFFlatData*)d;

  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=(IMFExpansionU((IMFExpansionNVGetE(u))))[i];

  return 0;
 }

int MFFlatMFProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;
  struct IMFFlatData *data;

  data=(struct IMFFlatData*)d;

  if(x==(double*)NULL)return data->n;
/*if(x==(double*)NULL)return data->n+1;*/

  for(i=0;i<data->n;i++)x[i]=(IMFExpansionU((IMFExpansionNVGetE(u))))[i];
/*x[data->n]=.005*IMFExpansionNVGetT(u);*/

  return 0;
 }

int MFFlatMFProjectToDraw(MFNVector u, double *x, void *d)
 {
  int i,j;
  struct IMFFlatData *data;

  data=(struct IMFFlatData*)d;

  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)
   {
    x[i]=MFNV_C(u,i);
    if(x[i]!=x[i]){printf("NaN in FlatProjectToDraw\n");fflush(stdout);}
   }

  return 0;
 }
