#include <MFAtlas.h>
#include <MFChart.h>
#include <MFFortran.h>
#include <IMFFlow.h>
#include <IMFExpansion.h>
#include <IMFExpansionSpace.h>
#include <IMFSphereOnExpansion.h>
#include <IMFInterpolation.h>
#include <MFImplicitMF.h>
#include <math.h>
#include <stdio.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

/* #define DOINTERPANIM */

FILE *IMFInterper=(FILE*)NULL;
int IMFNInterper=0;
FILE *IMFInterpee=(FILE*)NULL;
int IMFNInterpee=0;
FILE *IMFInterpT=(FILE*)NULL;
int IMFNInterpT=0;
FILE *IMFCircle=(FILE*)NULL;
int IMFNCircle=0;

MFChart MFAtlasChart(MFAtlas,int);
double *MFKV_CStar(MFKVector);
double *MFNKM_CStar(MFNKMatrix);

MFNVector IMFInterpolateOnDualFace(MFAtlas,IMFFlow,MFNVector,int,int*);
MFNVector IMFInterpolateOnDualFace2(MFAtlas,IMFFlow,MFAtlas,MFKVector,MFNVector,int,int*);
void MFNullVectorSmall(int,double*,double*);
void MFSolveFullSmall(int,double*,double*);
double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
int MFChartChangedFlag(MFChart);
int MFAtlasAddChartToList(MFAtlas,MFNVector,MFNKMatrix,double);

#define kmax 10
void TestEm(int,int,int);
void TestF(double r, int nu, double *u, int np, double *p, double *u0,double *l0,
                               void (*f)(double,int,double*,int,double*,double*,double*,double*),
                               void (*fu)(double,int,double*,int,double*,double*,double*,double*),
                               void (*fl)(double,int,double*,int,double*,double*,double*,double*));
void TestA(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0,
                               void (*a)(int,int,double*,double*,int,double*,double*,double*,double*,double*),
                               void (*au)(int,int,double*,double*,int,double*,double*,double*,double*,double*),
                               void (*al)(int,int,double*,double*,int,double*,double*,double*,double*,double*));

MFNVector IMFGetInterpolationPoint(MFAtlas A,IMFFlow F, MFAtlas c, double tmax, MFNRegion Omega)
 {
/*
    edges whose index is all >m give points where edge crosses sphere
    clip points against all faces with index >m not in index set of edge
    mark intersection point
    index of edge is ind1,ind2
*/

  static char RoutineName[]={"IMFGetInterpolationPoint"};
  int verbose=1;

  MFPolytope P;
  int iBndChart,nBndCharts;
  int chart;
  int doit;

  int i,j,k,l;
  int n,m;
  int n1,n2;
  double d0,d1,t;
  double r;
  int good0;
  int good1;
  int inter[kmax];
  MFNVector result;
  double R;

  int ii,jj;
  int noncube,Mcube;
  double a[(kmax-1)*kmax],b[kmax-1],aa[(kmax-1)*(kmax-1)],asave[(kmax-1)*kmax];
  int iface[kmax-1];
  double d[kmax];
  double org[kmax],nrm[kmax];
  double oo;
  double *nn;
  double *f;
  MFKVector vf;
  MFNVector z;
  MFNVector vp;
  double *p;
  MFNVector vF;
  double *pF;
  int nv;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

#ifdef DOINTERPANIM
  if(IMFInterper!=(FILE*)NULL)
   {
    fclose(IMFInterper);
    IMFInterper=(FILE*)NULL;
   }
  if(IMFInterper==(FILE*)NULL)
   {
    IMFInterper=fopen("Interp1.dx","w");
    IMFNInterper=0;
   }
  if(IMFInterpee==(FILE*)NULL)
   {
    IMFInterpee=fopen("Interp2.dx","w");
    IMFNInterpee=0;
   }
  if(IMFInterpT==(FILE*)NULL)
   {
    IMFInterpT=fopen("Interp3.dx","w");
    IMFNInterpT=0;
   }
  if(IMFCircle!=(FILE*)NULL)
   {
    fclose(IMFCircle);
    IMFCircle=(FILE*)NULL;
   }
  if(IMFCircle==(FILE*)NULL)
   {
    IMFCircle=fopen("IMFCircle.dx","w");
    IMFNCircle=0;
   }
#endif

  nBndCharts=MFAtlasNumberOfChartsWithBoundary(A);
  if(verbose){printf("There are %d charts on the boundary list\n",nBndCharts);fflush(stdout);}

  k=MFAtlasK(A);
  if(k>kmax)
   {
    fprintf(stderr,"%s, I statically allocated some things here for k<10! Complain.\n");
    fflush(stderr);
    abort();
   }

  Mcube=1;
  for(i=0;i<k;i++)Mcube=Mcube*2;
  Mcube=Mcube-1;

  vf=MFCreateKVector(k);
  f=MFKV_CStar(vf);
  z=MFCreateNVector(MFAtlasN(A));
  vp=MFCreateNVector(MFAtlasN(A));
  p=MFNV_CStar(vp);
  vF=MFCreateNVector(MFAtlasN(A));
  pF=MFNV_CStar(vF);

/* Clean boundary list */

  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));

    d0=0.;
    for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
     {
      d1=MFPolytopeRadiusOfVertex(MFChartPolytope(MFAtlasChart(A,chart)),i);
      if(d1>d0)d0=d1;
     }

    R=MFChartRadius(MFAtlasChart(A,chart));
    if(MFChartIsSingular(MFAtlasChart(A,chart))
           ||d0<R
           ||IMFExpansionNVGetT(MFChartCenter(MFAtlasChart(A,chart)))>tmax
           ||(Omega!=(MFNRegion)NULL&&!MFNRegionInterior(Omega,IMFExpansionNVGetSigma(MFChartCenter(MFAtlasChart(A,chart))))) )
     {
      MFAtlasRemoveChartFromBoundaryList(A,iBndChart);
      iBndChart--;
      nBndCharts--;
     }
   }

#ifdef DOINTERPANIM
  nBndCharts=MFAtlasNumberOfChartsWithBoundary(A);
  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));
    R=MFChartRadius(MFAtlasChart(A,chart));

    if(0)
     { /* Draw Polygon */
      for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
       {
        n1=MFPolytopeVertexNumberOfIndices(P,i);
        for(j=i+1;j<MFPolytopeNumberOfVertices(P);j++)
         {
          n2=MFPolytopeVertexNumberOfIndices(P,j);
          m=n1;if(n2>m)m=n2;

          if(m>0)
           {
            n=MFPolytopeIntersectIndexSets(P,i,j,inter);
           }else{
            n=0;
           }

          if(n==k-1)
           {
            MFPolytopeVertex(P,i,vf);
            MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
            for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
            IMFNInterpee++;
            MFPolytopeVertex(P,j,vf);
            MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
            for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
            IMFNInterpee++;
           }
         }
       }
     }
    if(1)
     { /* Draw Ball */
      for(i=0;i<301;i++)
       {
        f[0]=R*cos(2*3.1415926*i/300.);
        f[1]=R*sin(2*3.1415926*i/300.);
        good0=1;
        for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
         {
          oo=MFPolytopeFaceOrigin(P,ii);
          nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
          d0=-oo;for(l=0;l<k;l++)d0+=f[l]*nn[l];
          if(d0>0)good0=0;
         }

        MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

        f[0]=R*cos(2*3.1415926*(i+1)/300.);
        f[1]=R*sin(2*3.1415926*(i+1)/300.);
        for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
         {
          oo=MFPolytopeFaceOrigin(P,ii);
          nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
          d0=-oo;for(l=0;l<k;l++)d0+=f[l]*nn[l];
          if(d0>0)good0=0;
         }

        if(good0)
         {
          for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);IMFNCircle++;
          MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
          for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
          IMFNCircle++;
         }
       }
     }
   }

  if(1)
   {
    for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
     {
      chart=MFAtlasChartWithBoundary(A,iBndChart);
      P=MFChartPolytope(MFAtlasChart(A,chart));

      R=MFChartRadius(MFAtlasChart(A,chart));
      for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
       {
        n1=MFPolytopeVertexNumberOfIndices(P,i);
        for(j=i+1;j<MFPolytopeNumberOfVertices(P);j++)
         {
          n2=MFPolytopeVertexNumberOfIndices(P,j);
          m=n1;if(n2>m)m=n2;

          if(m>0)
           {
            n=MFPolytopeIntersectIndexSets(P,i,j,inter);
           }else{
            n=0;
           }

          noncube=1;for(l=0;l<n;l++)noncube=noncube&&inter[l]>Mcube;

          if(n==k-1&&noncube)
           {
            for(ii=0;ii<k-1;ii++)
             {
              iface[ii]=-1;
              for(l=0;l<MFPolytopeNumberOfFaces(P);l++)
                if(MFPolytopeFaceIndex(P,l)==inter[ii])iface[ii]=l;
             }
  
            for(ii=0;ii<k-1;ii++)
             {
              b[ii]=MFPolytopeFaceOrigin(P,iface[ii]);
              nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
              d0;for(l=0;l<k;l++)d0+=nn[l]*nn[l];
              for(l=0;l<k;l++)
               {
                a[ii+(k-1)*l]=nn[l]/d0;
                asave[ii+(k-1)*l]=a[ii+(k-1)*l];
               }
              b[ii]=b[ii]/d0;
             }

            for(ii=0;ii<k-1;ii++)
             {
              for(jj=0;jj<k-1;jj++)
               {
                aa[ii+(k-1)*jj]=0;
                for(l=0;l<k;l++)aa[ii+(k-1)*jj]+=a[ii+(k-1)*l]*a[jj+(k-1)*l];
               }
             }

            MFNullVectorSmall(k,a,nrm);
            MFSolveFullSmall(k-1,aa,b);

            for(ii=0;ii<k;ii++)
             {
              org[ii]=0.;
              for(l=0;l<k-1;l++)org[ii]+=asave[l+(k-1)*ii]*b[l];
             }

            r=R*R;for(l=0;l<k;l++)r-=org[l]*org[l];
            if(r>0)
             {
              r=sqrt(r);

              good0=1;
              good1=1;
              for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
               {
                doit=1;for(l=0;l<k-1;l++)if(ii==iface[l])doit=0;
                if(doit)
                 {
                  oo=MFPolytopeFaceOrigin(P,ii);
                  nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
                  d0=-oo;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*nn[l];
                  if(d0>0)good0=0;
                  d1=-oo;for(l=0;l<k;l++)d1+=(org[l]-r*nrm[l])*nn[l];
                  if(d1>0)good1=0;
                 }
               }

              if(good0)
               {
                for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                IMFNInterper++;
                if(1||inter[0]<inter[1])
                 {
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                  IMFEvaluateFlow(F,vp,pF);
                  MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                  d0=0.;for(l=0;l<k;l++)d0+=f[l]*f[l];
                  for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l]+.2*f[l]/sqrt(d0);
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                 }
                for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

                IMFEvaluateFlow(F,vp,pF);
                MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
                t=-r/d1;
                if(t<0.)
                 {
                  good0=1;
                  for(ii=0;ii<k-1;ii++)
                   {
                    d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                    d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                    if(d0<0 || d0>d1)good0=0;
                   }
                  if(good0)
                   {
                    for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                    IMFNInterper++;
                   }
                 }
               }

              if(good1)
               {
                for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                IMFNInterper++;
                if(1||inter[0]<inter[1])
                 {
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                  IMFEvaluateFlow(F,vp,pF);
                  MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                  d0=0.;for(l=0;l<k;l++)d0+=f[l]*f[l];
                  for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l]+.2*f[l]/sqrt(d0);
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                 }
                for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

                IMFEvaluateFlow(F,vp,pF);
                MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
                t=r/d1;
                if(t<0.)
                 {
                  good1=1;
                  for(ii=0;ii<k-1;ii++)
                   {
                    d0=0.;for(l=0;l<k;l++)d0+=(org[l]-r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                    d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                    if(d0<0 || d0>d1)good1=0;
                   }
                  if(good1)
                   {
                    for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                    IMFNInterper++;
                   }
                 }
               }
             }
           }
         }
       }
     }
   }
#endif

  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));
    if(verbose){printf("The %dth chart on the boundary list is chart %d\n",iBndChart,chart);fflush(stdout);}

    nv=MFPolytopeNumberOfVertices(P);
/*  if(!MFChartChangedFlag(MFAtlasChart(A,chart)))nv=0;*/

    R=MFChartRadius(MFAtlasChart(A,chart));
    for(i=0;i<nv;i++)
     {
      n1=MFPolytopeVertexNumberOfIndices(P,i);
      for(j=i+1;j<nv;j++)
       {
        n2=MFPolytopeVertexNumberOfIndices(P,j);
        m=n1;if(n2>m)m=n2;

        if(m>0)
         {
          n=MFPolytopeIntersectIndexSets(P,i,j,inter);
         }else{
          n=0;
         }

        noncube=1;for(l=0;l<n;l++)noncube=noncube&&inter[l]>Mcube;

        if(n==k-1&&noncube)
         {
          if(verbose){printf("  Indices of this non-cube edge:\n");fflush(stdout);}
          for(ii=0;ii<k-1;ii++)
           {
            iface[ii]=-1;
            for(l=0;l<MFPolytopeNumberOfFaces(P);l++)
             {
              if(MFPolytopeFaceIndex(P,l)==inter[ii])iface[ii]=l;
             }
            if(verbose){printf("  %d (face %d)\n",inter[ii],iface[ii]);fflush(stdout);}
           }

          for(ii=0;ii<k-1;ii++)
           {
            b[ii]=MFPolytopeFaceOrigin(P,iface[ii]);
            nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
            d0;for(l=0;l<k;l++)d0+=nn[l]*nn[l];
            for(l=0;l<k;l++)
             {
              a[ii+(k-1)*l]=nn[l]/d0;
              asave[ii+(k-1)*l]=a[ii+(k-1)*l];
             }
            b[ii]=b[ii]/d0;
           }

          for(ii=0;ii<k-1;ii++)
           {
            for(jj=0;jj<k-1;jj++)
             {
              aa[ii+(k-1)*jj]=0;
              for(l=0;l<k;l++)aa[ii+(k-1)*jj]+=a[ii+(k-1)*l]*a[jj+(k-1)*l];
             }
           }

          MFNullVectorSmall(k,a,nrm);
          MFSolveFullSmall(k-1,aa,b);

          for(ii=0;ii<k;ii++)
           {
            org[ii]=0.;
            for(l=0;l<k-1;l++)org[ii]+=asave[l+(k-1)*ii]*b[l];
           }

          if(verbose)
           {
            printf("  Done finding edge equation org=(%lf",org[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",org[ii]);
            printf(")\n");fflush(stdout);
            printf("                             nrm=(%lf",nrm[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",nrm[ii]);
            printf(")\n");fflush(stdout);
    
            printf("Check: is org on all the faces?\n");fflush(stdout);
            d0=0.;for(l=0;l<k;l++)d0+=org[l]*nrm[l];
            printf("            nrm.org=%lf\n",d0);fflush(stdout);
            for(ii=0;ii<k-1;ii++)
             {
              oo=MFPolytopeFaceOrigin(P,iface[ii]);
              nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
              d0=-oo;for(l=0;l<k;l++)d0+=nn[l]*org[l];
              printf("   face %d: org.n-o=%lf\n",ii,d0);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=nn[l]*nrm[l];
              printf("            nrm.n=%lf\n",d0);fflush(stdout);
             }
           }

          r=R*R;for(l=0;l<k;l++)r-=org[l]*org[l];
          if(r>0)
           {
            r=sqrt(r);
            if(verbose)
             {
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("  This edge crosses the ball. |pt0|=%lf, R=%lf\n",d0,R);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("                              |pt1|=%lf, R=%lf\n",d0,R);fflush(stdout);
             }

            good0=1;
            good1=1;
            for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
             {
              doit=1;for(l=0;l<k-1;l++)if(ii==iface[l])doit=0;
              if(doit)
               {
                oo=MFPolytopeFaceOrigin(P,ii);
                nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
                d0=-oo;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*nn[l];
                if(d0>0)good0=0;
                d1=-oo;for(l=0;l<k;l++)d1+=(org[l]-r*nrm[l])*nn[l];
                if(d1>0)good1=0;
               }
             }

            if(verbose)
             {
              if(good0){printf("  The +r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good0)
             {
              for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=-r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good0=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good0=0;
                 }
                if(1||good0)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
/*                result=IMFInterpolateOnDualFace2(A,F,c,vf,z,chart,inter);*/
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
#ifdef DOINTERPANIM
                      for(j=0;j<MFNV_NC(vp);j++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
                      IMFNInterpee++;
                      for(j=0;j<MFNV_NC(result);j++)fprintf(IMFInterpee," %lf",IMFExpansionU((IMFExpansionNVGetE(result)))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
                      IMFNInterpee++;
#endif
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }

            if(verbose)
             {
              if(good1){printf("  The -r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good1)
             {
              for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good1=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]-r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good1=0;
                 }
                if(1||good1)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
/*                result=IMFInterpolateOnDualFace2(A,F,A,vf,z,chart,inter);*/
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
#ifdef DOINTERPANIM
                      for(j=0;j<MFNV_NC(vp);j++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
                      IMFNInterpee++;
                      for(j=0;j<MFNV_NC(result);j++)fprintf(IMFInterpee," %lf",IMFExpansionU((IMFExpansionNVGetE(result)))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
                      IMFNInterpee++;
#endif
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }
           }
         }
       }
     }
    MFChartResetChangedFlag(MFAtlasChart(A,chart));
   }

  result=(MFNVector)NULL;
  printf("No such points on the boundary!!!\n");

FreeAndReturn:
  MFFreeKVector(vf);
  MFFreeNVector(z);
  MFFreeNVector(vp);
  MFFreeNVector(vF);

  return result;
 }

void MFNullVectorSmall(int k,double *A,double *phi)
 {
/*Find the null vector of a (k-1)xk matrix */

  double t0,t1;

  if(k==2)
   {
    t1=sqrt(A[0]*A[0]+A[1]*A[1]);
    t0    = A[1]/t1;
    phi[1]=-A[0]/t1;
    phi[0]=t0;
   }else if(k==3)
   {
    t0=sqrt(A[0]*A[0]+A[2]*A[2]+A[4]*A[4]);
    t1=sqrt(A[1]*A[1]+A[3]*A[3]+A[5]*A[5]);
    phi[0]=(A[2]*A[5]-A[3]*A[4])/t0/t1;
    phi[1]=(A[4]*A[1]-A[5]*A[0])/t0/t1;
    phi[2]=(A[0]*A[3]-A[1]*A[2])/t0/t1;
   }else{  
    int i,j;
    char jobvl='N';
    char jobvr='A';
    int lda;
    double *s;
    double *U;
    int ldU;
    double *V;
    int ldV;
    double *work;
    int lwork;
    int info=0;
    int m;
    double t;

    lda=k-1;
    ldU=k-1;
    U=(double*)NULL;

    ldV=k;
    V=malloc(k*(k-1)*sizeof(double));
    if(V==(double*)NULL)abort();

    s=(double*)malloc(k*sizeof(double));
    if(s==(double*)NULL)abort();

    i=-1;
    info=0;
    jobvl='N';
    jobvr='A';
    F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&lda,&k,A,&lda,s,U,&ldU,V,&ldV,&t,&i,&info);
    lwork=t;

    work=malloc(lwork*sizeof(double));
    if(work==(double*)NULL)abort();

    info=0;
    jobvl='N';
    jobvr='A';
    F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&lda,&k,A,&lda,s,U,&ldU,V,&ldV,work,&lwork,&info);
  
    m=0;
    for(i=0;i<k;i++)
     {
      if(fabs(s[i])<1.e-7)
       {
        for(j=0;j<k;j++)phi[j]=V[i+ldV*j];
        m++;
       }
     }
    free(V);
    free(s);
    free(work);
   }
 }

void MFSolveFullSmall(int k,double *A,double *b)
 {
/*solves the kxk system of linear eqs.;*/
  double t;

  if(k==1)
   {
    b[0]=b[0]/A[0];
   }else if(k==2)
   {
    t   =(A[3]*b[0]-A[2]*b[1])/(A[0]*A[3]-A[1]*A[2]);
    b[1]=(A[0]*b[1]-A[1]*b[0])/(A[0]*A[3]-A[1]*A[2]);
    b[0]=t;
   }else{  
    MFSolveFull(k,A,b);
   }

  return;
 }

MFNVector IMFInterpolateOnDualFace(MFAtlas A, IMFFlow F, MFNVector z, int chart, int *inter)
 {
  int i,k,n;
  int j;
  MFNVector c[kmax];
  double a[(kmax-1)*(kmax-1)];
  double t[kmax];
  double d;
  MFNSpace space;
  MFNVector diffi;
  MFNVector diffj;
  MFNVector diffz;
  MFNVector result;
  MFNVector sigma;
  int nsigma;
  double time;
  IMFExpansion U,uu;
  IMFExpansion V0,V1;
  MFKVector vs,s0;
  double *s;

  double *u;
  double *du;
  double *ds;
  double *dds;
  int good;
  int verbose=0;

  k=MFAtlasK(A);
  n=MFAtlasN(A);
  space=MFIMFNSpace(MFAtlasMF(A));

  c[0]=MFAtlasChartCenter(A,chart);
  diffi=MFCreateNVector(n);
  diffj=MFCreateNVector(n);
  diffz=MFCreateNVector(n);
  vs=MFCreateKVector(k);
  s=MFKV_CStar(vs);
  s0=MFCreateKVector(k);
  
  if(0){printf("IMFInterpolateOnDualFace, z=");MFPrintNVector(stdout,z);printf("\n");fflush(stdout);}
  nsigma=MFNV_NC(IMFExpansionNVGetSigma(c[0]));

  good=0;
  for(i=0;i<k-1;i++)
   {
    j=MFAtlasHalfSpaceRightChart(A,inter[i]);
    if(j==chart)j=MFAtlasHalfSpaceLeftChart(A,inter[i]);
    c[i+1]=MFAtlasChartCenter(A,j);
    d=0.;for(j=0;j<nsigma;j++)d+=pow(MFNV_CStar(IMFExpansionNVGetSigma(c[0]))[j]-
                                     MFNV_CStar(IMFExpansionNVGetSigma(c[i+1]))[j],2.);
    if(sqrt(d)>1.e-14)good=1;
   }

  if(!good)
   {
    MFFreeNVector(diffi);
    MFFreeNVector(diffj);
    MFFreeNVector(diffz);
    MFFreeKVector(vs);
    MFFreeKVector(s0);

    return (MFNVector)NULL;
   }

  MFNSpaceDirection(space,z,c[0],diffz);
  for(i=0;i<k-1;i++)
   {
    MFNSpaceDirection(space,c[i+1],c[0],diffi);
    a[i+(k-1)*i]=MFNSpaceInner(space,diffi,diffi);
    for(j=i+1;j<k-1;j++)
     {
      MFNSpaceDirection(space,c[j+1],c[0],diffj);
      a[i+(k-1)*j]=MFNSpaceInner(space,diffi,diffj);
      a[i+(k-1)*j]=a[j+(k-1)*i];
     }
    t[i+1]=MFNSpaceInner(space,diffz,diffi);
   }

  MFSolveFullSmall(k-1,a,t+1);

  t[0]=1.;for(i=1;i<k;i++)t[0]-=t[i];
  if(verbose)
   {
    printf("Barycentric coordinates (%lf",t[0]);
    for(i=1;i<k;i++)printf(",%lf",t[i]);
    printf(")\n");
   }

  good=1.;
  for(i=0;i<k;i++)if(t[i]<0.||t[i]>1.)good=0;
  for(i=0;i<k;i++)if(t[i]<.1||t[i]>.9)good=0;  /* EXPERIMENT */
  if(!good)
   {
    MFFreeNVector(diffi);
    MFFreeNVector(diffj);
    MFFreeNVector(diffz);
    MFFreeKVector(vs);
    MFFreeKVector(s0);
    return (MFNVector)NULL;
   }

  if(k!=2)
   {
    fprintf(stderr,"Interpolation of more than 2d is not yet supported\n");
    fflush(stderr);
    abort();
   }

  u=(double*)malloc(n*sizeof(double));
  du=(double*)malloc(n*k*sizeof(double));
  ds=(double*)malloc(n*sizeof(double));
  dds=(double*)malloc(n*sizeof(double));

/*
  position is straightforward, as are t and sigma.
  tangent and second derivatives get as scalars along direction from c[0] to c[1]
    then interpolate u_s and u_ss 
  then use the flow to build the full expansion. (same as initial curve)
*/

  sigma=MFCreateNVector(nsigma);
  V0=IMFExpansionNVGetE(c[0]);
  V1=IMFExpansionNVGetE(c[1]);
  for(j=0;j<n;j++)u[j]=t[0]*(IMFExpansionU(V0))[j]+t[1]*(IMFExpansionU(V1))[j];
  time=t[0]*IMFExpansionNVGetT(c[0])+t[1]*IMFExpansionNVGetT(c[1]);
  for(j=0;j<nsigma;j++)MFNV_CStar(sigma)[j]=t[0]*MFNV_CStar(IMFExpansionNVGetSigma(c[0]))[j];
                                           +t[1]*MFNV_CStar(IMFExpansionNVGetSigma(c[1]))[j];

#ifdef DOINTERPANIM
  for(j=0;j<n;j++)fprintf(IMFInterpee," %lf",(IMFExpansionU(V0))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);IMFNInterpee++;
  for(j=0;j<n;j++)fprintf(IMFInterpee," %lf",(IMFExpansionU(V1))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);IMFNInterpee++;
#endif
  if(verbose)
   {
    printf("Barycentric coordinates (%lf",t[0]);
    for(i=1;i<k;i++)printf(",%lf",t[i]);
    printf(")\n");
   }

  MFKV_CStar(s0)[0]=0.;
  MFKV_CStar(s0)[1]=0.;
  MFNSpaceDirection(space,c[0],c[1],diffi);
  d=0.;for(j=0;j<n;j++)d+=MFNV_CStar(diffi)[j]*MFNV_CStar(diffi)[j];d=sqrt(d);
  if(verbose)
   {
    printf("Vector from 0 to 1 (%lf",MFNV_CStar(diffi)[0])/d;
    for(i=1;i<n;i++)printf(",%lf",MFNV_CStar(diffi)[i]/d);
    printf(")\n");
   }
  MFMVMulT(MFIMFNSpace(MFAtlasMF(A)),MFAtlasChartTangentSpace(A,chart),diffi,vs);
  d=0.;for(j=0;j<k;j++)d+=s[j]*s[j];d=sqrt(d);
  for(j=0;j<k;j++)s[j]=s[j]/d;
  IMFEvaluateExpansionDirectionalDerivative(V0,s0,vs,diffj);
  for(j=0;j<n;j++)ds[j]=t[0]*MFNV_CStar(diffj)[j];
  if(verbose)
   {
    printf("Tangent 0          (%lf",MFNV_CStar(diffj)[0]);
    for(i=1;i<n;i++)printf(",%lf",MFNV_CStar(diffj)[i]);
    printf(")\n");
   }
  IMFEvaluateExpansionSecondDirectionalDerivative(V0,s0,vs,vs,diffj);
  for(j=0;j<n;j++)dds[j]=t[0]*MFNV_CStar(diffj)[j];

  j=MFAtlasHalfSpaceRightChart(A,inter[0]);
  if(j==chart)j=MFAtlasHalfSpaceLeftChart(A,inter[0]);
  MFMVMulT(MFIMFNSpace(MFAtlasMF(A)),MFAtlasChartTangentSpace(A,j),diffi,vs);
  d=0.;for(j=0;j<k;j++)d+=s[j]*s[j];d=sqrt(d);
  for(j=0;j<k;j++)s[j]=s[j]/d;
  IMFEvaluateExpansionDirectionalDerivative(V1,s0,vs,diffj);
  for(j=0;j<n;j++)ds[j]+=t[1]*MFNV_CStar(diffj)[j];
  if(verbose)
   {
    printf("Tangent 1          (%lf",MFNV_CStar(diffj)[0]);
    for(i=1;i<n;i++)printf(",%lf",MFNV_CStar(diffj)[i]);
    printf(")\n");
    printf("Interpolated       (%lf",ds[0]);
    for(i=1;i<n;i++)printf(",%lf",ds[i]);
    printf(")\n");
   }
  IMFEvaluateExpansionSecondDirectionalDerivative(V1,s0,vs,vs,diffj);
  for(j=0;j<n;j++)dds[j]+=t[1]*MFNV_CStar(diffj)[j];

  uu=IMFCreateExpansion(n,1);
  IMFExpansionSetDerivatives(uu,u,ds,dds,(double*)NULL);
  U=IMFInflateExpansionWithFlow(uu,F);

/*printf("Interpolated u, (%lf",u[0]);for(j=1;j<n;j++)printf(",%lf",u[j]);printf(")");fflush(stdout);
  printf(", time %lf",time);
  printf(", sigma");MFPrintNVector(stdout,sigma);printf("\n");fflush(stdout);*/

  result=IMFCreateExpansionNVector(U,time,sigma,chart,2);

  IMFExpansionNVSetChart0(result,chart);
  IMFExpansionNVSetS0(result,s0);

  MFFreeNVector(diffi);
  MFFreeNVector(diffj);
  MFFreeNVector(diffz);
  IMFFreeExpansion(U);
  IMFFreeExpansion(uu);
  MFFreeNVector(sigma);
  MFFreeKVector(vs);
  MFFreeKVector(s0);

  free(u);
  free(du);
  free(ds);
  free(dds);

  return result;
 }

MFNVector IMFGetInterpolationPointOnList(MFAtlas A,IMFFlow F, MFAtlas c, double tmax, MFNRegion Omega, int nList, int *list)
 {
/*
    edges whose index is all >m give points where edge crosses sphere
    clip points against all faces with index >m not in index set of edge
    mark intersection point
    index of edge is ind1,ind2
*/

  static char RoutineName[]={"IMFGetInterpolationPointOnList"};
  int verbose=0;

  MFPolytope P;
  int iBndChart,nBndCharts;
  int chart;
  int doit;

  int i,j,k,l;
  int n,m;
  int n1,n2;
  double d0,d1,t;
  double r;
  int good0;
  int good1;
  int inter[kmax];
  MFNVector result;
  double R;

  int ii,jj;
  int noncube;
  double a[(kmax-1)*kmax],b[kmax-1],aa[(kmax-1)*(kmax-1)],asave[(kmax-1)*kmax];
  int iface[kmax-1];
  double d[kmax];
  double org[kmax],nrm[kmax];
  double oo;
  double *nn;
  double *f;
  MFKVector vf;
  MFNVector z;
  MFNVector vp;
  double *p;
  MFNVector vF;
  double *pF;
  int nv;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

#ifdef DOINTERPANIM
  if(IMFInterper!=(FILE*)NULL)
   {
    fclose(IMFInterper);
    IMFInterper=(FILE*)NULL;
   }
  if(IMFInterper==(FILE*)NULL)
   {
    IMFInterper=fopen("Interp1.dx","w");
    IMFNInterper=0;
   }
  if(IMFInterpee==(FILE*)NULL)
   {
    IMFInterpee=fopen("Interp2.dx","w");
    IMFNInterpee=0;
   }
  if(IMFCircle!=(FILE*)NULL)
   {
    fclose(IMFCircle);
    IMFCircle=(FILE*)NULL;
   }
  if(IMFCircle==(FILE*)NULL)
   {
    IMFCircle=fopen("IMFCircle.dx","w");
    IMFNCircle=0;
   }
#endif

  nBndCharts=MFAtlasNumberOfChartsWithBoundary(A);
  if(verbose){printf("There are %d charts on the boundary list\n",nBndCharts);fflush(stdout);}

  k=MFAtlasK(A);
  if(k>kmax)
   {
    fprintf(stderr,"%s, I statically allocated some things here for k<10! Complain.\n");
    fflush(stderr);
    abort();
   }

  m=1;
  for(i=0;i<k;i++)m=m*2;
  m=m-1;

  vf=MFCreateKVector(k);
  f=MFKV_CStar(vf);
  z=MFCreateNVector(MFAtlasN(A));
  vp=MFCreateNVector(MFAtlasN(A));
  p=MFNV_CStar(vp);
  vF=MFCreateNVector(MFAtlasN(A));
  pF=MFNV_CStar(vF);

/* Clean boundary list */

  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));

    d0=0.;
    for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
     {
      d1=MFPolytopeRadiusOfVertex(MFChartPolytope(MFAtlasChart(A,chart)),i);
      if(d1>d0)d0=d1;
     }

    R=MFChartRadius(MFAtlasChart(A,chart));
    if(MFChartIsSingular(MFAtlasChart(A,chart))
           ||d0<R
           ||IMFExpansionNVGetT(MFChartCenter(MFAtlasChart(A,chart)))>tmax
           ||(Omega!=(MFNRegion)NULL&&!MFNRegionInterior(Omega,IMFExpansionNVGetSigma(MFChartCenter(MFAtlasChart(A,chart))))) )
     {
      MFAtlasRemoveChartFromBoundaryList(A,iBndChart);
      iBndChart--;
      nBndCharts--;
     }
   }

#ifdef DOINTERPANIM
  nBndCharts=MFAtlasNumberOfChartsWithBoundary(A);
  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));
    R=MFChartRadius(MFAtlasChart(A,chart));

    if(0)
     { /* Draw Polygon */
      for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
       {
        n1=MFPolytopeVertexNumberOfIndices(P,i);
        for(j=i+1;j<MFPolytopeNumberOfVertices(P);j++)
         {
          n2=MFPolytopeVertexNumberOfIndices(P,j);
          m=n1;if(n2>m)m=n2;

          if(m>0)
           {
            n=MFPolytopeIntersectIndexSets(P,i,j,inter);
           }else{
            n=0;
           }

          if(n==k-1)
           {
            MFPolytopeVertex(P,i,vf);
            MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
            for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
            IMFNInterpee++;
            MFPolytopeVertex(P,j,vf);
            MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
            for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);
            IMFNInterpee++;
           }
         }
       }
     }
    if(1)
     { /* Draw Ball */
      for(i=0;i<301;i++)
       {
        f[0]=R*cos(2*3.1415926*i/300.);
        f[1]=R*sin(2*3.1415926*i/300.);
        good0=1;
        for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
         {
          oo=MFPolytopeFaceOrigin(P,ii);
          nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
          d0=-oo;for(l=0;l<k;l++)d0+=f[l]*nn[l];
          if(d0>0)good0=0;
         }

        MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

        f[0]=R*cos(2*3.1415926*(i+1)/300.);
        f[1]=R*sin(2*3.1415926*(i+1)/300.);
        for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
         {
          oo=MFPolytopeFaceOrigin(P,ii);
          nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
          d0=-oo;for(l=0;l<k;l++)d0+=f[l]*nn[l];
          if(d0>0)good0=0;
         }

        if(good0)
         {
          for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);IMFNCircle++;
          MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
          for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
          IMFNCircle++;
         }
       }
     }
   }

  if(1)
   {
    for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
     {
      chart=MFAtlasChartWithBoundary(A,iBndChart);
      P=MFChartPolytope(MFAtlasChart(A,chart));

      R=MFChartRadius(MFAtlasChart(A,chart));
      for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
       {
        n1=MFPolytopeVertexNumberOfIndices(P,i);
        for(j=i+1;j<MFPolytopeNumberOfVertices(P);j++)
         {
          n2=MFPolytopeVertexNumberOfIndices(P,j);
          m=n1;if(n2>m)m=n2;

          if(m>0)
           {
            n=MFPolytopeIntersectIndexSets(P,i,j,inter);
           }else{
            n=0;
           }

          noncube=1;for(l=0;l<n;l++)noncube=noncube&&inter[l]>m;

          if(n==k-1&&noncube)
           {
            for(ii=0;ii<k-1;ii++)
             {
              iface[ii]=-1;
              for(l=0;l<MFPolytopeNumberOfFaces(P);l++)
                if(MFPolytopeFaceIndex(P,l)==inter[ii])iface[ii]=l;
             }
  
            for(ii=0;ii<k-1;ii++)
             {
              b[ii]=MFPolytopeFaceOrigin(P,iface[ii]);
              nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
              d0;for(l=0;l<k;l++)d0+=nn[l]*nn[l];
              for(l=0;l<k;l++)
               {
                a[ii+(k-1)*l]=nn[l]/d0;
                asave[ii+(k-1)*l]=a[ii+(k-1)*l];
               }
              b[ii]=b[ii]/d0;
             }

            for(ii=0;ii<k-1;ii++)
             {
              for(jj=0;jj<k-1;jj++)
               {
                aa[ii+(k-1)*jj]=0;
                for(l=0;l<k;l++)aa[ii+(k-1)*jj]+=a[ii+(k-1)*l]*a[jj+(k-1)*l];
               }
             }

            MFNullVectorSmall(k,a,nrm);
            MFSolveFullSmall(k-1,aa,b);

            for(ii=0;ii<k;ii++)
             {
              org[ii]=0.;
              for(l=0;l<k-1;l++)org[ii]+=asave[l+(k-1)*ii]*b[l];
             }

            r=R*R;for(l=0;l<k;l++)r-=org[l]*org[l];
            if(r>0)
             {
              r=sqrt(r);

              good0=1;
              good1=1;
              for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
               {
                doit=1;for(l=0;l<k-1;l++)if(ii==iface[l])doit=0;
                if(doit)
                 {
                  oo=MFPolytopeFaceOrigin(P,ii);
                  nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
                  d0=-oo;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*nn[l];
                  if(d0>0)good0=0;
                  d1=-oo;for(l=0;l<k;l++)d1+=(org[l]-r*nrm[l])*nn[l];
                  if(d1>0)good1=0;
                 }
               }

              if(good0)
               {
                for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                IMFNInterper++;
                if(1||inter[0]<inter[1])
                 {
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                  IMFEvaluateFlow(F,vp,pF);
                  MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                  d0=0.;for(l=0;l<k;l++)d0+=f[l]*f[l];
                  for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l]+.2*f[l]/sqrt(d0);
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                 }
                for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

                IMFEvaluateFlow(F,vp,pF);
                MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
                t=-r/d1;
                if(t<0.)
                 {
                  good0=1;
                  for(ii=0;ii<k-1;ii++)
                   {
                    d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                    d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                    if(d0<0 || d0>d1)good0=0;
                   }
                  if(good0)
                   {
                    for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                    IMFNInterper++;
                   }
                 }
               }

              if(good1)
               {
                for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                IMFNInterper++;
                if(inter[0]<inter[1])
                 {
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                  IMFEvaluateFlow(F,vp,pF);
                  MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                  d0=0.;for(l=0;l<k;l++)d0+=f[l]*f[l];
                  for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l]+.2*f[l]/sqrt(d0);
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
                  for(l=0;l<MFAtlasN(A);l++)fprintf(IMFCircle," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFCircle,"\n");fflush(IMFCircle);
                  IMFNCircle++;
                 }
                for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
                MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);

                IMFEvaluateFlow(F,vp,pF);
                MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
                d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
                t=r/d1;
                if(t<0.)
                 {
                  good1=1;
                  for(ii=0;ii<k-1;ii++)
                   {
                    d0=0.;for(l=0;l<k;l++)d0+=(org[l]-r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                    d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                    if(d0<0 || d0>d1)good1=0;
                   }
                  if(good1)
                   {
                    for(l=0;l<MFAtlasN(A);l++)fprintf(IMFInterper," %lf",(MFNV_CStar(vp))[l]);fprintf(IMFInterper,"\n");fflush(IMFInterper);
                    IMFNInterper++;
                   }
                 }
               }
             }
           }
         }
       }
     }
   }
#endif

  for(iBndChart=0;iBndChart<nList;iBndChart++)
   {
    chart=list[iBndChart];
    P=MFChartPolytope(MFAtlasChart(A,chart));
    if(verbose){printf("The %dth chart on the boundary list is chart %d\n",iBndChart,chart);fflush(stdout);}
    if(!MFChartHasBoundary(MFAtlasChart(A,chart)))continue;
    nv=MFPolytopeNumberOfVertices(P);
/*  if(!MFChartChangedFlag(MFAtlasChart(A,chart)))nv=0;*/

    R=MFChartRadius(MFAtlasChart(A,chart));
    for(i=0;i<nv;i++)
     {
      n1=MFPolytopeVertexNumberOfIndices(P,i);
      for(j=i+1;j<nv;j++)
       {
        n2=MFPolytopeVertexNumberOfIndices(P,j);
        m=n1;if(n2>m)m=n2;

        if(m>0)
         {
          n=MFPolytopeIntersectIndexSets(P,i,j,inter);
         }else{
          n=0;
         }

        noncube=1;for(l=0;l<n;l++)noncube=noncube&&inter[l]>m;

        if(n==k-1&&noncube)
         {
          if(verbose){printf("  Indices of this non-cube edge:\n");fflush(stdout);}
          for(ii=0;ii<k-1;ii++)
           {
            iface[ii]=-1;
            for(l=0;l<MFPolytopeNumberOfFaces(P);l++)
             {
              if(MFPolytopeFaceIndex(P,l)==inter[ii])iface[ii]=l;
             }
            if(verbose){printf("  %d (face %d)\n",inter[ii],iface[ii]);fflush(stdout);}
           }

          for(ii=0;ii<k-1;ii++)
           {
            b[ii]=MFPolytopeFaceOrigin(P,iface[ii]);
            nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
            d0;for(l=0;l<k;l++)d0+=nn[l]*nn[l];
            for(l=0;l<k;l++)
             {
              a[ii+(k-1)*l]=nn[l]/d0;
              asave[ii+(k-1)*l]=a[ii+(k-1)*l];
             }
            b[ii]=b[ii]/d0;
           }

          for(ii=0;ii<k-1;ii++)
           {
            for(jj=0;jj<k-1;jj++)
             {
              aa[ii+(k-1)*jj]=0;
              for(l=0;l<k;l++)aa[ii+(k-1)*jj]+=a[ii+(k-1)*l]*a[jj+(k-1)*l];
             }
           }

          MFNullVectorSmall(k,a,nrm);
          MFSolveFullSmall(k-1,aa,b);

          for(ii=0;ii<k;ii++)
           {
            org[ii]=0.;
            for(l=0;l<k-1;l++)org[ii]+=asave[l+(k-1)*ii]*b[l];
           }

          if(verbose)
           {
            printf("  Done finding edge equation org=(%lf",org[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",org[ii]);
            printf(")\n");fflush(stdout);
            printf("                             nrm=(%lf",nrm[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",nrm[ii]);
            printf(")\n");fflush(stdout);
    
            printf("Check: is org on all the faces?\n");fflush(stdout);
            d0=0.;for(l=0;l<k;l++)d0+=org[l]*nrm[l];
            printf("            nrm.org=%lf\n",d0);fflush(stdout);
            for(ii=0;ii<k-1;ii++)
             {
              oo=MFPolytopeFaceOrigin(P,iface[ii]);
              nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
              d0=-oo;for(l=0;l<k;l++)d0+=nn[l]*org[l];
              printf("   face %d: org.n-o=%lf\n",ii,d0);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=nn[l]*nrm[l];
              printf("            nrm.n=%lf\n",d0);fflush(stdout);
             }
           }

          r=R*R;for(l=0;l<k;l++)r-=org[l]*org[l];
          if(r>0)
           {
            r=sqrt(r);
            if(verbose)
             {
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("  This edge crosses the ball. |pt0|=%lf, R=%lf\n",d0,R);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("                              |pt1|=%lf, R=%lf\n",d0,R);fflush(stdout);
             }

            good0=1;
            good1=1;
            for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
             {
              doit=1;for(l=0;l<k-1;l++)if(ii==iface[l])doit=0;
              if(doit)
               {
                oo=MFPolytopeFaceOrigin(P,ii);
                nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
                d0=-oo;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*nn[l];
                if(d0>0)good0=0;
                d1=-oo;for(l=0;l<k;l++)d1+=(org[l]-r*nrm[l])*nn[l];
                if(d1>0)good1=0;
               }
             }

            if(verbose)
             {
              if(good0){printf("  The +r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good0)
             {
              for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=-r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good0=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good0=0;
                 }
                if(1||good0)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
#ifdef DOINTERPANIM
                      for(j=0;j<MFNV_NC(vp);j++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterper);
                      IMFNInterpee++;
                      for(j=0;j<MFNV_NC(result);j++)fprintf(IMFInterpee," %lf",IMFExpansionU((IMFExpansionNVGetE(result)))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterper);
                      IMFNInterpee++;
#endif
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }

            if(verbose)
             {
              if(good1){printf("  The -r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good1)
             {
              for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good1=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]-r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good1=0;
                 }
                if(1||good1)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
#ifdef DOINTERPANIM
                      for(j=0;j<MFNV_NC(vp);j++)fprintf(IMFInterpee," %lf",(MFNV_CStar(vp))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterper);
                      IMFNInterpee++;
                      for(j=0;j<MFNV_NC(result);j++)fprintf(IMFInterpee," %lf",IMFExpansionU((IMFExpansionNVGetE(result)))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterper);
                      IMFNInterpee++;
#endif
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }
           }
         }
       }
     }
    MFChartResetChangedFlag(MFAtlasChart(A,chart));
   }

  result=(MFNVector)NULL;
  printf("No such points on the boundary!!!\n");

FreeAndReturn:
  MFFreeKVector(vf);
  MFFreeNVector(z);
  MFFreeNVector(vp);
  MFFreeNVector(vF);

  return result;
 }

#ifdef JJJKJK
MFNVector IMFGetInterpolationPointOnList(MFAtlas A,IMFFlow F, MFAtlas c, double tmax, MFNRegion Omega, int nList, int *list)
 {
/*
    edges whose index is all >m give points where edge crosses sphere
    clip points against all faces with index >m not in index set of edge
    mark intersection point
    index of edge is ind1,ind2
*/

  static char RoutineName[]={"IMFGetInterpolationPointOnList"};
  int verbose=0;

  MFPolytope P;
  int iBndChart,nBndCharts;
  int chart;
  int doit;

  int i,j,k,l;
  int n,m;
  int n1,n2;
  double d0,d1,t;
  double r;
  int good0;
  int good1;
  int inter[kmax];
  MFNVector result;
  double R;

  int ii,jj;
  int noncube;
  double a[(kmax-1)*kmax],b[kmax-1],aa[(kmax-1)*(kmax-1)],asave[(kmax-1)*kmax];
  int iface[kmax-1];
  double d[kmax];
  double org[kmax],nrm[kmax];
  double oo;
  double *nn;
  double *f;
  MFKVector vf;
  MFNVector z;
  MFNVector vp;
  double *p;
  MFNVector vF;
  double *pF;
  int nv;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}


  nBndCharts=MFAtlasNumberOfChartsWithBoundary(A);
  if(verbose){printf("There are %d charts on the boundary list\n",nBndCharts);fflush(stdout);}

  k=MFAtlasK(A);
  if(k>kmax)
   {
    fprintf(stderr,"%s, I statically allocated some things here for k<10! Complain.\n");
    fflush(stderr);
    abort();
   }

  m=1;
  for(i=0;i<k;i++)m=m*2;
  m=m-1;

  vf=MFCreateKVector(k);
  f=MFKV_CStar(vf);
  z=MFCreateNVector(MFAtlasN(A));
  vp=MFCreateNVector(MFAtlasN(A));
  p=MFNV_CStar(vp);
  vF=MFCreateNVector(MFAtlasN(A));
  pF=MFNV_CStar(vF);

/* Clean boundary list */

  for(iBndChart=0;iBndChart<nBndCharts;iBndChart++)
   {
    chart=MFAtlasChartWithBoundary(A,iBndChart);
    P=MFChartPolytope(MFAtlasChart(A,chart));

    d0=0.;
    for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
     {
      d1=MFPolytopeRadiusOfVertex(MFChartPolytope(MFAtlasChart(A,chart)),i);
      if(d1>d0)d0=d1;
     }

    R=MFChartRadius(MFAtlasChart(A,chart));
    if(MFChartIsSingular(MFAtlasChart(A,chart))
           ||d0<R
           ||IMFExpansionNVGetT(MFChartCenter(MFAtlasChart(A,chart)))>tmax
           ||(Omega!=(MFNRegion)NULL&&!MFNRegionInterior(Omega,IMFExpansionNVGetSigma(MFChartCenter(MFAtlasChart(A,chart))))) )
     {
      MFAtlasRemoveChartFromBoundaryList(A,iBndChart);
      iBndChart--;
      nBndCharts--;
     }
   }


  for(iBndChart=0;iBndChart<nList;iBndChart++)
   {
    chart=list[iBndChart];
    P=MFChartPolytope(MFAtlasChart(A,chart));
    if(verbose){printf("The %dth chart on the boundary list is chart %d\n",iBndChart,chart);fflush(stdout);}
    if(!MFChartHasBoundary(MFAtlasChart(A,chart)))continue;
    nv=MFPolytopeNumberOfVertices(P);
/*  if(!MFChartChangedFlag(MFAtlasChart(A,chart)))nv=0;*/

    R=MFChartRadius(MFAtlasChart(A,chart));
    for(i=0;i<nv;i++)
     {
      n1=MFPolytopeVertexNumberOfIndices(P,i);
      for(j=i+1;j<nv;j++)
       {
        n2=MFPolytopeVertexNumberOfIndices(P,j);
        m=n1;if(n2>m)m=n2;

        if(m>0)
         {
          n=MFPolytopeIntersectIndexSets(P,i,j,inter);
         }else{
          n=0;
         }

        noncube=1;for(l=0;l<n;l++)noncube=noncube&&inter[l]>m;

        if(n==k-1&&noncube)
         {
          if(verbose){printf("  Indices of this non-cube edge:\n");fflush(stdout);}
          for(ii=0;ii<k-1;ii++)
           {
            iface[ii]=-1;
            for(l=0;l<MFPolytopeNumberOfFaces(P);l++)
             {
              if(MFPolytopeFaceIndex(P,l)==inter[ii])iface[ii]=l;
             }
            if(verbose){printf("  %d (face %d)\n",inter[ii],iface[ii]);fflush(stdout);}
           }

          for(ii=0;ii<k-1;ii++)
           {
            b[ii]=MFPolytopeFaceOrigin(P,iface[ii]);
            nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
            d0;for(l=0;l<k;l++)d0+=nn[l]*nn[l];
            for(l=0;l<k;l++)
             {
              a[ii+(k-1)*l]=nn[l]/d0;
              asave[ii+(k-1)*l]=a[ii+(k-1)*l];
             }
            b[ii]=b[ii]/d0;
           }

          for(ii=0;ii<k-1;ii++)
           {
            for(jj=0;jj<k-1;jj++)
             {
              aa[ii+(k-1)*jj]=0;
              for(l=0;l<k;l++)aa[ii+(k-1)*jj]+=a[ii+(k-1)*l]*a[jj+(k-1)*l];
             }
           }

          MFNullVectorSmall(k,a,nrm);
          MFSolveFullSmall(k-1,aa,b);

          for(ii=0;ii<k;ii++)
           {
            org[ii]=0.;
            for(l=0;l<k-1;l++)org[ii]+=asave[l+(k-1)*ii]*b[l];
           }

          if(verbose)
           {
            printf("  Done finding edge equation org=(%lf",org[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",org[ii]);
            printf(")\n");fflush(stdout);
            printf("                             nrm=(%lf",nrm[0]);fflush(stdout);
            for(ii=1;ii<k;ii++)printf(",%lf",nrm[ii]);
            printf(")\n");fflush(stdout);
    
            printf("Check: is org on all the faces?\n");fflush(stdout);
            d0=0.;for(l=0;l<k;l++)d0+=org[l]*nrm[l];
            printf("            nrm.org=%lf\n",d0);fflush(stdout);
            for(ii=0;ii<k-1;ii++)
             {
              oo=MFPolytopeFaceOrigin(P,iface[ii]);
              nn=MFKV_CStar(MFPolytopeFaceNormal(P,iface[ii]));
              d0=-oo;for(l=0;l<k;l++)d0+=nn[l]*org[l];
              printf("   face %d: org.n-o=%lf\n",ii,d0);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=nn[l]*nrm[l];
              printf("            nrm.n=%lf\n",d0);fflush(stdout);
             }
           }

          r=R*R;for(l=0;l<k;l++)r-=org[l]*org[l];
          if(r>0)
           {
            r=sqrt(r);
            if(verbose)
             {
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("  This edge crosses the ball. |pt0|=%lf, R=%lf\n",d0,R);fflush(stdout);
              d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*(org[l]+r*nrm[l]);d0=sqrt(d0);
              printf("                              |pt1|=%lf, R=%lf\n",d0,R);fflush(stdout);
             }

            good0=1;
            good1=1;
            for(ii=0;ii<MFPolytopeNumberOfFaces(P);ii++)
             {
              doit=1;for(l=0;l<k-1;l++)if(ii==iface[l])doit=0;
              if(doit)
               {
                oo=MFPolytopeFaceOrigin(P,ii);
                nn=MFKV_CStar(MFPolytopeFaceNormal(P,ii));
                d0=-oo;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l])*nn[l];
                if(d0>0)good0=0;
                d1=-oo;for(l=0;l<k;l++)d1+=(org[l]-r*nrm[l])*nn[l];
                if(d1>0)good1=0;
               }
             }

            if(verbose)
             {
              if(good0){printf("  The +r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good0)
             {
              for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=-r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good0=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]+r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good0=0;
                 }
                if(1||good0)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]+r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }

            if(verbose)
             {
              if(good1){printf("  The -r point is in the polytope.\n");fflush(stdout);}
                  else {printf("  The +r point is not in the polytope.\n");fflush(stdout);}
             }

            if(good1)
             {
              for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l];
              MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,vp);
              IMFEvaluateFlow(F,vp,pF);
              MFChartProjectIntoTangentSpace(MFAtlasChart(A,chart),vF,vf);
              d1=0.;for(l=0;l<k;l++)d1+=f[l]*nrm[l];
              t=r/d1;
              if(t<0.&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,vp)))
               {
                good1=1;
                for(ii=0;ii<k-1;ii++)
                 {
                  d0=0.;for(l=0;l<k;l++)d0+=(org[l]-r*nrm[l]+t*f[l])*asave[ii+(k-1)*l];
                  d1=0.;for(l=0;l<k;l++)d1+= org[l]                 *asave[ii+(k-1)*l];
                  if(d0<0 || d0>d1)good1=0;
                 }
                if(1||good1)
                 {
                  for(l=0;l<k;l++)f[l]=org[l]-r*nrm[l]+t*f[l];
                  MFChartPointInTangentSpace(MFAtlasChart(A,chart),vf,z);
                  result=IMFInterpolateOnDualFace(A,F,z,chart,inter);
                  if(result!=(MFNVector)NULL)
                   {
                    if(IMFExpansionNVGetT(result)<tmax&&(Omega==(MFNRegion)NULL||MFNRegionInterior(Omega,IMFExpansionNVGetSigma(result))))
                     {
                      if(verbose){printf("Result is on chart %d\n",chart);fflush(stdout);}
                      MFChartResetChangedFlag(MFAtlasChart(A,chart));
                      goto FreeAndReturn;
                     }
                   }
                 }
               }
             }
           }
         }
       }
     }
    MFChartResetChangedFlag(MFAtlasChart(A,chart));
   }

  result=(MFNVector)NULL;
  printf("No such points on the boundary!!!\n");

FreeAndReturn:
  MFFreeKVector(vf);
  MFFreeNVector(z);
  MFFreeNVector(vp);
  MFFreeNVector(vF);

  return result;
 }
#endif

#define nmax 10000

IMFFlow IMFTPBVPFlow=(IMFFlow)NULL;
MFChart IMFTPBVPChart=(MFChart)NULL;
MFAtlas IMFTPBVPC=(MFAtlas)NULL;
int IMFTPBVPCChart=-1;
double *IMFTPBVPs=(double*)NULL;

void IMFTPBVPf (double,int,double*,int,double*,double*,double*,double*);
void IMFTPBVPfu(double,int,double*,int,double*,double*,double*,double*);
void IMFTPBVPfl(double,int,double*,int,double*,double*,double*,double*);
void IMFTPBVPa (int,int,double*,double*,int,double*,double*,double*,double*,double*);
void IMFTPBVPau(int,int,double*,double*,int,double*,double*,double*,double*,double*);
void IMFTPBVPal(int,int,double*,double*,int,double*,double*,double*,double*,double*);

MFNVector IMFInterpolateOnDualFace2(MFAtlas A, IMFFlow F, MFAtlas c, MFKVector s, MFNVector z, int chart, int *inter)
 {
  static char RoutineName[]={"IMFInterpolateOnDualFace2"};
  int i,j,l,n,m;
  int prev;
  MFNVector result;
  MFNVector u[nmax];
  int k;
  double a;
  MFImplicitMF M;
  MFKVector s0;
  MFNVector u0,ug;
  double lambda;
  IMFExpansion E;
  MFNKMatrix TS;
  double R,Rf,Rmax;
  double epsilon=.01;
  int nx,nu,np;
  int nE;
  int verbose=1;
  IMFFlow Fp;
  IMFExpansion E0,E1;
  MFNVector c0,c1;
  double d;
  int good;
  int nsigma;
  MFNVector diffi;
  MFNVector diffj;
  MFNVector diffz;
  double t[kmax];
  MFNVector C[kmax];
  MFNSpace space;
  double aa[(kmax-1)*(kmax-1)];
 
  if(verbose){printf("%s-------------------------------------------\n",RoutineName);fflush(stdout);}

  k=MFAtlasK(A);
  n=MFAtlasN(A);

  if(k>kmax)
   {
    fprintf(stderr,"%s, I statically allocated some things here for k<10! Complain.\n");
    fflush(stderr);
    abort();
   }

  C[0]=MFAtlasChartCenter(A,chart);
  printf("Chart containing the edge is %d",chart);MFPrintNVector(stdout,C[0]);printf("\n");fflush(stdout);
  nsigma=MFNV_NC(IMFExpansionNVGetSigma(C[0]));
  good=0;
  for(i=0;i<k-1;i++)
   {
    j=MFAtlasHalfSpaceRightChart(A,inter[i]);
    if(j==chart)j=MFAtlasHalfSpaceLeftChart(A,inter[i]);
    C[i+1]=MFAtlasChartCenter(A,j);
    printf(" adj Chart %d is %d",i,j);fflush(stdout);MFPrintNVector(stdout,C[i+1]);printf("\n");fflush(stdout);
    d=0.;for(j=0;j<nsigma;j++)d+=pow(MFNV_CStar(IMFExpansionNVGetSigma(C[0]))[j]-
                                     MFNV_CStar(IMFExpansionNVGetSigma(C[i+1]))[j],2.);
    if(sqrt(d)>1.e-14)good=1;
   }

  if(!good)return (MFNVector)NULL;

  diffi=MFCreateNVector(n);
  diffj=MFCreateNVector(n);
  diffz=MFCreateNVector(n);
  space=MFIMFNSpace(MFAtlasMF(A));

  MFNSpaceDirection(space,z,C[0],diffz);
  for(i=0;i<k-1;i++)
   {
    MFNSpaceDirection(space,C[i+1],C[0],diffi);
    aa[i+(k-1)*i]=MFNSpaceInner(space,diffi,diffi);
    for(j=i+1;j<k-1;j++)
     {
      MFNSpaceDirection(space,C[j+1],C[0],diffj);
      aa[i+(k-1)*j]=MFNSpaceInner(space,diffi,diffj);
      aa[i+(k-1)*j]=aa[j+(k-1)*i];
     }
    t[i+1]=MFNSpaceInner(space,diffz,diffi);
   }

  MFSolveFullSmall(k-1,aa,t+1);

  t[0]=1.;for(i=1;i<k;i++)t[0]-=t[i];
  if(verbose)
   {
    printf("Barycentric coordinates (%lf",t[0]);
    for(i=1;i<k;i++)printf(",%lf",t[i]);
    printf(")\n");
   }

  MFFreeNVector(diffi);
  MFFreeNVector(diffj);
  MFFreeNVector(diffz);

  good=1.;
  for(i=0;i<k;i++)if(t[i]<0.||t[i]>1.)good=0;
  for(i=0;i<k;i++)if(t[i]<.1||t[i]>.9)good=0;  /* EXPERIMENT */
  if(!good)return (MFNVector)NULL;

#ifdef DOINTERPANIM
  E0=IMFExpansionNVGetE(MFAtlasChartCenter(A,chart));
  for(j=0;j<3;j++)fprintf(IMFInterpee," %lf",(IMFExpansionU(E0))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);IMFNInterpee++;
  for(i=0;i<2-1;i++)
   {
    m=MFAtlasHalfSpaceRightChart(A,inter[i]);
    if(m==chart)m=MFAtlasHalfSpaceLeftChart(A,inter[i]);
   }
  E1=IMFExpansionNVGetE(MFAtlasChartCenter(A,m));
  for(j=0;j<3;j++)fprintf(IMFInterpee," %lf",(IMFExpansionU(E1))[j]);fprintf(IMFInterpee,"\n");fflush(IMFInterpee);IMFNInterpee++;
#endif

/* Get trajectory from the initial curve to the center of chart */

  if(verbose){printf("Get trajectory from the initial curve to the center of chart\n");fflush(stdout);}
  n=0;
  prev=IMFExpansionNVGetPrev(MFAtlasChartCenter(A,chart));
  if(verbose){printf(" n=%d, chart=%d, prev=%d\n",n,chart,prev);fflush(stdout);}
  n++;
  while(prev>0)
   {
    if(verbose){printf(" n=%d, chart=%d",n,prev);}
    prev=IMFExpansionNVGetPrev(MFAtlasChartCenter(A,prev));
    if(verbose){printf(", prev=%d\n",prev);fflush(stdout);}
    n++;
   }
  if(verbose){printf("n=%d\n",n);fflush(stdout);}

  i=n-1;
  u[i]=MFCloneNVector(MFAtlasChartCenter(A,chart));
  prev=IMFExpansionNVGetPrev(u[i]);
  if(verbose){printf("u[%d]=chart %d, prev=%d\n",i,chart,prev);fflush(stdout);}
  i--;
  while(prev>0)
   {
    if(verbose){printf("u[%d]=chart %d",i,prev);fflush(stdout);}
    u[i]=MFCloneNVector(MFAtlasChartCenter(A,prev));
    prev=IMFExpansionNVGetPrev(u[i]);
    if(verbose){printf(", prev=%d\n",prev);fflush(stdout);}
    i--;
   }
  if(verbose){printf("Done getting trajectory\n");fflush(stdout);}

/* Do the homotopy for the interpolated trajectory */

  k=MFAtlasK(c)+1;
  Fp=IMFCreateFatFlow(F,k);
  if(verbose){printf("c(0x%8.8x) is an %d-manifold in %d space\n",c,MFAtlasK(c),MFAtlasN(c));fflush(stdout);}
  m=1;
  nx=m*(n-1)+2;
  M=MFIMFCreateTPBVP(1,nx,IMFFlowN(Fp),k+1,IMFTPBVPf,IMFTPBVPfu,IMFTPBVPfl,
                            IMFFlowN(Fp)+k,IMFTPBVPa,IMFTPBVPau,IMFTPBVPal,
                            0,(MFTPBVPLFUNCTION)NULL,(MFTPBVPLFUNCTION)NULL,(MFTPBVPLFUNCTION)NULL,
                            (MFTPBVPMFUNCTION)NULL,(MFTPBVPMFUNCTION)NULL);
  if(verbose){printf("trajectory starts at chart %d on c, s(0x%8.8x)=",IMFExpansionNVGetChart0(u[0]),IMFExpansionNVGetS0(u[0]));fflush(stdout);}
  if(verbose){MFPrintKVector(stdout,IMFExpansionNVGetS0(u[0]));printf("\n");fflush(stdout);}
  IMFTPBVPFlow=Fp;
  IMFTPBVPChart=MFAtlasChart(A,chart);
  k=MFKV_NC(IMFExpansionNVGetS0(u[0]))+1;
  IMFTPBVPs=(double*)malloc((k-1)*sizeof(double));if(IMFTPBVPs==(double*)NULL)abort();
  for(i=0;i<k-1;i++)IMFTPBVPs[i]=MFKV_C(IMFExpansionNVGetS0(u[0]),i);
  IMFTPBVPC=c;
  IMFTPBVPCChart=IMFExpansionNVGetChart0(u[0]);

  nx=MFTPBVPGetNX(M);
  nu=MFTPBVPGetNU(M);
  np=MFTPBVPGetNP(M);
  if(verbose){printf("TPBVP: n=%d, nx=%d, nu=%d, np=%d\n",MFIMF_N(M),nx,nu,np);fflush(stdout);}

  TestEm(nu,np,MFTPBVPGetNBC(M));

  TS=MFCreateNKMatrixWithData(MFIMF_N(M),1,(double*)NULL);
  for(i=0;i<MFIMF_N(M);i++)MFNKMSetC(TS,i,0,0.);
  MFNKMSetC(TS,nu*nx+k-1,0,1.);

  if(verbose){printf("Construct initial guess\n");fflush(stdout);}
  u0=MFCreateNVector(nu*nx+np+nx+1);

  E0=IMFExpansionNVGetE(u[0]);E1=IMFExpansionNVGetE(u[1]);
  a=-.5/m;
  printf("E0 dataLn=%d\n",IMFExpansionDataLn(E0));fflush(stdout);
  printf("E1 dataLn=%d\n",IMFExpansionDataLn(E1));fflush(stdout);
  nE=IMFExpansionDataLn(E1);
  for(j=0;j<nE;j++)MFNVSetC(u0,j,(1-a)*IMFExpansionData(E0)[j]+a*IMFExpansionData(E1)[j]);
  MFNVSetC(u0,nu*nx+np,(1-a)*IMFExpansionNVGetT(u[0])+a*IMFExpansionNVGetT(u[1]));
  printf("left most point, t=%lf, u=(%lf,%lf,%lf,...)\n",MFNV_C(u0,nu*nx+np),MFNV_C(u0,0),MFNV_C(u0,1),MFNV_C(u0,2));
    fflush(stdout);
#ifdef DOTRAJ
  for(j=0;j<3;j++)fprintf(IMFInterpT," %lf",MFNV_C(u0,j));fprintf(IMFInterpT,"\n");fflush(IMFInterpT);IMFNInterpT++;
#endif

  for(i=1;i<n;i++)
   {
    E0=IMFExpansionNVGetE(u[i-1]);
    E1=IMFExpansionNVGetE(u[i  ]);
    printf("  u[%d] t=%lf, u=(%lf,%lf,%lf,...)\n",i-1,IMFExpansionNVGetT(u[i-1]),IMFExpansionData(E0)[0],IMFExpansionData(E0)[1],IMFExpansionData(E0)[2]);
    fflush(stdout);
    for(l=0;l<m;l++)
     {
      a=(.5+l)/m;
      for(j=0;j<nE;j++)MFNVSetC(u0,j+nu*(1+l+m*(i-1)),(1-a)*IMFExpansionData(E0)[j]+a*IMFExpansionData(E1)[j]);
#ifdef DOTRAJ
  for(j=0;j<3;j++)fprintf(IMFInterpT," %lf",MFNV_C(u0,j+nu*(l+m*(i-1))));fprintf(IMFInterpT,"\n");fflush(IMFInterpT);IMFNInterpT++;
#endif
      MFNVSetC(u0,nu*nx+np+1+l+m*(i-1),(1-a)*IMFExpansionNVGetT(u[i-1])+a*IMFExpansionNVGetT(u[i]));
      printf("   l=%d, t=%lf, u=(%lf,%lf,%lf,...)\n",l,MFNV_C(u0,nu*nx+np+l+m*(i-1)),MFNV_C(u0,0+nu*(l+m*(i-1))),MFNV_C(u0,1+nu*(l+m*(i-1))),MFNV_C(u0,2+nu*(l+m*(i-1))));
        fflush(stdout);
     }
   }
  E0=IMFExpansionNVGetE(u[n-1]);
  printf("  u[%d] t=%lf, u=(%lf,%lf,%lf,...)\n",n-1,IMFExpansionNVGetT(u[n-1]),IMFExpansionData(E0)[0],IMFExpansionData(E0)[1],IMFExpansionData(E0)[2]);

  a=(.5+m+1)/m;
  E0=IMFExpansionNVGetE(u[n-2]);
  E1=IMFExpansionNVGetE(u[n-1]);
  for(j=0;j<nE;j++)MFNVSetC(u0,j+nu*(1+m*(n-1)),(1-a)*IMFExpansionData(E0)[j]+a*IMFExpansionData(E1)[j]);
  MFNVSetC(u0,nu*nx+np+m*(n-1)+1,(1-a)*IMFExpansionNVGetT(u[n-2])+a*IMFExpansionNVGetT(u[n-1]));
  printf("right most point, t=%lf, u=(%lf,%lf,%lf,...)\n",MFNV_C(u0,nu*nx+np+1+m*(n-1)),MFNV_C(u0,0+nu*(1+m*(n-1))),MFNV_C(u0,1+nu*(1+m*(n-1))),MFNV_C(u0,2+nu*(1+m*(n-1))));
        fflush(stdout);
#ifdef DOTRAJ
  for(j=0;j<3;j++)fprintf(IMFInterpT," %lf",MFNV_C(u0,j+nu*(1+m*(n-1))));fprintf(IMFInterpT,"\n");fflush(IMFInterpT);IMFNInterpT++;
#endif

  MFNVSetC(u0,nu*nx+0,1.); /* alpha */
  for(j=0;j<k-1;j++)
   {
    MFNVSetC(u0,nu*nx+j+1,0.); /* sigma_j */
   }
  MFNVSetC(u0,nu*nx+k,0.); /* lambda */

  ug=MFCreateNVector(MFIMF_N(M));

/* Simple homotopy */

  for(lambda=0.;lambda<=1.;lambda+=.1);
   {
    MFNVSetC(u0,nu*nx+k,lambda);

    if(verbose){printf("Project at lambda=%lf\n",lambda);fflush(stdout);}
    MFIMFProject(M,u0,TS,ug);
    for(i=0;i<MFIMF_N(M);i++)MFNVSetC(u0,i,MFNV_C(ug,i));
   }

/* Add the interpolated trajectory to the atlas */
  if(verbose){printf("Add the interpolated trajectory to the atlas\n");fflush(stdout);}

  IMFExpansionNVSetChart0(u[0],IMFTPBVPCChart);
  s0=MFCreateKVector(k-1);
  for(i=0;i<k-1;i++)MFKVSetC(s0,i,IMFTPBVPs[i]);
  IMFExpansionNVSetS0(u[0],s0);
  MFFreeKVector(s0);
  free(IMFTPBVPs);

  prev=-1;
  for(i=0;i<n;i++)
   {
    IMFExpansionNVSetPrev(u[i],prev);
    if(i>0)IMFExpansionNVSetType(u[i],3);
     else IMFExpansionNVSetType(u[i],1);
    E=IMFExpansionNVGetE(u[i]);
    TS=IMFExpansionTS(E);
    R=IMFExpansionR(E,epsilon);
    Rf=IMFFlowR(F,epsilon,u[i],TS);
    if(Rf<R)R=Rf;
    if(R>Rmax||R!=R)R=Rmax;
    prev=MFAtlasAddChartWithAll(A,u[i],TS,R);
   }
  
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  IMFFreeFlow(Fp);
  return u[n-1];
 }

void IMFTPBVPf(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *f)
 {
  int i;
  MFNVector vu;
  int verbose=0;

  if(verbose)
   {
    printf("In IMFTPBVPf\n");fflush(stdout);
    printf("nu is %d, FlowN is %d\n",nu,IMFFlowN(IMFTPBVPFlow));fflush(stdout);
   }
  vu=MFCreateWrappedNVector(nu,u);

  IMFEvaluateFlow(IMFTPBVPFlow,vu,f);
  if(verbose){printf("p[0]=%lf,f:(%lf,",p[0],f[0]);for(i=1;i<nu;i++)printf(",%lf",f[i]);printf(")\n");fflush(stdout);}
  for(i=0;i<nu;i++)f[i]=p[0]*f[i];

  MFFreeNVector(vu);
  if(verbose){printf("Done IMFTPBVPf\n");fflush(stdout);}

  return;
 }

void IMFTPBVPfu(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *fu)
 {
  int i;
  MFNVector vu;

  vu=MFCreateWrappedNVector(nu,u);

  IMFEvaluateDerivativeOfFlow(IMFTPBVPFlow,vu,fu);
  for(i=0;i<nu*nu;i++)fu[i]=p[0]*fu[i];

  MFFreeNVector(vu);

  return;
 }

void IMFTPBVPfl(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *fl)
 {
  int i;
  MFNVector vu;

  vu=MFCreateWrappedNVector(nu,u);
  IMFEvaluateFlow(IMFTPBVPFlow,vu,fl);
  for(i=0;i<nu;i++)fl[i+nu*0]=fl[i];
  for(i=0;i<nu*(np-1);i++)fl[i+nu]=0.;

  MFFreeNVector(vu);

  return;
 }

/* Parms:
      p[0] : alpha
      p[1],...,p[k-1] : sigma
      p[k] : lambda
*/

void IMFTPBVPa(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *a)
 {
  int i,k;
  MFKVector vsigma;
  MFKVector vs;
  MFNVector vuR;
  MFNVector vc;
  double *c;
  MFChart cchart;
  double nrm;
  double *sigma,*s;
  IMFExpansion E;
  static int verbose=1;

  if(verbose){printf("In IMFTPBVPa\n");fflush(stdout);}

  k=MFAtlasK(IMFTPBVPC);
  vsigma=MFCreateKVector(k);
  sigma=MFKV_CStar(vsigma);
  vs=MFCreateKVector(k+1);
  s=MFKV_CStar(vs);
  vuR=MFCreateWrappedNVector(nu,uR);
  c=(double*)malloc(MFAtlasN(IMFTPBVPC)*sizeof(double));
  for(i=0;i<MFAtlasN(IMFTPBVPC);i++)c[i]=0.;
  vc=MFCreateWrappedNVector(MFAtlasN(IMFTPBVPC),c);

  cchart=MFAtlasChart(IMFTPBVPC,IMFTPBVPCChart);
  nrm=0.;for(i=0;i<k-1;i++)nrm+=p[i]*p[i];nrm=sqrt(nrm);
  if(nrm>MFAtlasChartRadius(IMFTPBVPC,IMFTPBVPCChart))
   {
    abort();
/*  pivot to next, change cchart and s
    this is going to make it difficult to take derivatives.*/
   }

  for(i=0;i<k-1;i++)sigma[i]=p[i+1];

/* c is n+k, look up last k -> s on expansion, evaluate expansion and derivatives at this point to 
     get a vector of lenght nu */

  MFChartPointInTangentSpace(cchart,vsigma,vc);
  E=IMFSphereOnExpansionGetLocal(IMFTPBVPC,vc);

/* At t=0. */

  if(verbose)
   {
    printf("Conditions at t=0. (0..%d):\n",nu-1);
    printf(" uL=(%lf,%lf,%lf,...\n",uL[0],uL[1],uL[2]);fflush(stdout);
    printf(" s=(%lf), c(s)=(%lf,%lf,%lf), chart %d\n",c[4],c[0],c[1],c[2],cchart);fflush(stdout);
    printf(" E(c(s))=(%lf,%lf,%lf,...)\n",IMFExpansionData(E)[0],IMFExpansionData(E)[1],IMFExpansionData(E)[2]);fflush(stdout);
   }

  for(i=0;i<nu;i++)a[i]=uL[i];
  for(i=0;i<IMFExpansionDataLn(E);i++)a[i]-=IMFExpansionData(E)[i];

/* At t=1. */

  if(verbose)
   {
    printf("Conditions at t=1. (%d..%d):\n",nu,nbc);fflush(stdout);
   }
  MFChartProjectIntoTangentSpace(IMFTPBVPChart,vuR,vs);
  for(i=nu;i<nbc;i++)a[i]=s[i-nu]-p[k]*IMFTPBVPs[i-nu];
  
  MFFreeKVector(vsigma);
  MFFreeKVector(vs);
  MFFreeNVector(vuR);
  free(c);
  MFFreeNVector(vc);
  IMFFreeExpansion(E);
  if(verbose){printf("Done IMFTPBVPa\n");fflush(stdout);}
  verbose=0;

  return;
 }

void IMFTPBVPau(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *au)
 {
  int i,j,k;
  MFNKMatrix cTS;
  double *TS;
  MFKVector vsigma;
  MFNVector vc;
  double *c;
  double *sigma;
  IMFExpansion E;
  MFChart cchart;
  int n1;

  k=MFAtlasK(IMFTPBVPC);
  for(i=0;i<2*nbc*nu;i++)au[i]=0.;

  cTS=MFChartTangentSpace(IMFTPBVPChart);
  TS=MFNKM_CStar(cTS);

  vsigma=MFCreateKVector(k);
  sigma=MFKV_CStar(vsigma);
  for(i=0;i<k-1;i++)sigma[i]=p[i+1];

  c=(double*)malloc(MFAtlasN(IMFTPBVPC)*sizeof(double));if(c==(double*)NULL)abort();
  for(i=0;i<MFAtlasN(IMFTPBVPC);i++)c[i]=0.;
  vc=MFCreateWrappedNVector(MFAtlasN(IMFTPBVPC),c);

  cchart=MFAtlasChart(IMFTPBVPC,IMFTPBVPCChart);
  MFChartPointInTangentSpace(cchart,vsigma,vc);
  E=IMFSphereOnExpansionGetLocal(IMFTPBVPC,vc);
  n1=MFChartN(IMFTPBVPChart);

  for(i=0;i<nu;i++)
      au[i+nbc*(i+0*nu)]=1.;

  for(i=nu;i<nbc;i++)
   {
    for(j=0;j<n1;j++)
      au[i+nbc*(j+nu)]=TS[j+n1*(i-nu)];
    for(j=n1;j<nu;j++)
      au[i+nbc*(j+nu)]=0.;
   }

  MFFreeKVector(vsigma);
  free(c);
  MFFreeNVector(vc);
  
  return;
 }

void IMFTPBVPal(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *al)
 {
  int i,j,k;
  MFNKMatrix cTS;
  double *TS;
  MFKVector vsigma;
  MFNVector vc;
  double *c;
  double *sigma;
  double *du;
  IMFExpansion E;
  int nc,n1;

  for(i=0;i<nbc*np;i++)al[i]=0.;

  k=MFAtlasK(IMFTPBVPC);
  nc=MFAtlasN(IMFTPBVPC);
  cTS=MFAtlasChartTangentSpace(IMFTPBVPC,IMFTPBVPCChart);
  TS=MFNKM_CStar(cTS);

  vsigma=MFCreateKVector(k);
  sigma=MFKV_CStar(vsigma);
  for(i=0;i<k-1;i++)sigma[i]=p[i+1];

  c=(double*)malloc(MFAtlasN(IMFTPBVPC)*sizeof(double));if(c==(double*)NULL)abort();
  for(i=0;i<MFAtlasN(IMFTPBVPC);i++)c[i]=0.;
  vc=MFCreateWrappedNVector(MFAtlasN(IMFTPBVPC),c);

  du=(double*)malloc(nu*sizeof(double));if(du==(double*)NULL)abort();

  E=IMFSphereOnExpansionGetLocal(IMFTPBVPC,vc);
  n1=IMFExpansionDataLn(E);

  for(j=0;j<k;j++)
   {
    IMFEvaluateExpansionDirectionalDerivativeE(E,c+nc-k-1,TS+nc-k-1+nc*j,du);
    for(i=0;i<n1;i++)
      al[i+nbc*(j+1)]+=du[i];
   }

  for(i=nu;i<nbc;i++)
   al[i+nbc*k]=-IMFTPBVPs[i-nu];

  MFFreeKVector(vsigma);
  free(c);
  free(du);
  MFFreeNVector(vc);
  
  return;
 }

#define nbcMAX 100
#define nuMAX 100
#define npMAX 100

void TestA(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0,
                               void (*a)(int,int,double*,double*,int,double*,double*,double*,double*,double*),
                               void (*au)(int,int,double*,double*,int,double*,double*,double*,double*,double*),
                               void (*al)(int,int,double*,double*,int,double*,double*,double*,double*,double*))
 {
  int i;
  int j;
  double Au[nbcMAX*2*nuMAX];
  double Al[nbcMAX*npMAX];
  double Ap[nbcMAX];
  double Am[nbcMAX];
  double save;
  double eps=1.e-4;
  double err;

  printf("Test derivatives of boundary conditions\n");fflush(stdout);

  (*au)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Au);
  (*al)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Al);
  for(i=0;i<nu;i++)
   {
    save=uL[i];
    uL[i]=save+eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Ap);
    uL[i]=save-eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Am);
    uL[i]=save;
    for(j=0;j<nbc;j++)
     {
      err=fabs(Au[j+nbc*i]-.5*(Ap[j]-Am[j])/eps);
      if(err>eps)
        printf(" da[%d]/duL[%d] = %le (exact) %le (diffed) error %le\n",j,i,Au[j+nbc*i],.5*(Ap[j]-Am[j])/eps,err);
     }
   }
  for(i=0;i<nu;i++)
   {
    save=uR[i];
    uR[i]=save+eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Ap);
    uR[i]=save-eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Am);
    uR[i]=save;
    for(j=0;j<nbc;j++)
     {
      err=fabs(Au[j+nbc*(i+nu)]-.5*(Ap[j]-Am[j])/eps);
      if(err>eps)
        printf(" da[%d]/duR[%d] = %le (exact) %le (diffed) error %le\n",j,i,Au[j+nbc*(i+nu)],.5*(Ap[j]-Am[j])/eps,err);
     }
   }
  for(i=0;i<np;i++)
   {
    save=p[i];
    p[i]=save+eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Ap);
    p[i]=save-eps;
    (*a)(nbc,nu,uL,uR,np,p,u0L,u0R,l0,Am);
    p[i]=save;
    for(j=0;j<nbc;j++)
     {
      err=fabs(Al[j+nbc*i]-.5*(Ap[j]-Am[j])/eps);
      if(err>eps)
        printf(" da[%d]/dp[%d] = %le (exact) %le (diffed) error %le\n",j,i,Al[j+nbc*i],.5*(Ap[j]-Am[j])/eps,err);
     }
   }
 }


void TestF(double r, int nu, double *u, int np, double *p, double *u0,double *l0,
                               void (*f)(double,int,double*,int,double*,double*,double*,double*),
                               void (*fu)(double,int,double*,int,double*,double*,double*,double*),
                               void (*fl)(double,int,double*,int,double*,double*,double*,double*))
 {
  int i;
  int j;
  double Fu[nuMAX*nuMAX];
  double Fl[nuMAX*npMAX];
  double Fp[nuMAX];
  double Fm[nuMAX];
  double save;
  double eps=1.e-4;
  double err;

  printf("Test derivatives of flow\n");fflush(stdout);

  (*fu)(r,nu,u,np,p,u0,l0,Fu);
  (*fl)(r,nu,u,np,p,u0,l0,Fl);
  for(i=0;i<nu;i++)
   {
    save=u[i];
    u[i]=save+eps;
    (*f)(r,nu,u,np,p,u0,l0,Fp);
    u[i]=save-eps;
    (*f)(r,nu,u,np,p,u0,l0,Fm);
    u[i]=save;
    for(j=0;j<nu;j++)
     {
      err=fabs(Fu[j+nu*i]-.5*(Fp[j]-Fm[j])/eps);
      if(err>eps)
        printf(" df[%d]/du[%d] = %le (exact) %le (diffed) error %le\n",j,i,Fu[j+nu*i],.5*(Fp[j]-Fm[j])/eps,err);
     }
   }
  for(i=0;i<np;i++)
   {
    save=p[i];
    p[i]=save+eps;
    (*f)(r,nu,u,np,p,u0,l0,Fp);
    p[i]=save-eps;
    (*f)(r,nu,u,np,p,u0,l0,Fm);
    p[i]=save;
    for(j=0;j<nu;j++)
     {
      eps=fabs(Fl[j+nu*i]-.5*(Fp[j]-Fm[j])/eps);
      if(err>eps)
        printf(" df[%d]/dp[%d] = %le (exact) %le (diffed) error %le\n",j,i,Fl[j+nu*i],.5*(Fp[j]-Fm[j])/eps,err);
     }
   }
 }

void TestEm(int nu, int np, int nbc)
 {
  int i;
  double r;
  double u[nuMAX];
  double u0[nuMAX];
  double uL[nuMAX];
  double u0L[nuMAX];
  double uR[nuMAX];
  double u0R[nuMAX];
  double p[npMAX];
  double l0[npMAX];

  r=1.;
  for(i=0;i<nu;i++)
   {
    u[i]=1.;
    uL[i]=1.;
    uR[i]=1.;
    u0[i]=1.;
    u0L[i]=1.;
    u0R[i]=1.;
   }
  for(i=0;i<np;i++)
   {
    p[i]=1.;
    l0[i]=1.;
   }

  printf("Test derivatives\n");fflush(stdout);
  TestF(r,nu,u,np,p,u0,l0,IMFTPBVPf,IMFTPBVPfu,IMFTPBVPfl);
  TestA(nbc,nu,uL,uR,np,p,u0L,u0R,l0,IMFTPBVPa,IMFTPBVPau,IMFTPBVPal);
 }
