#include <math.h>
#include <stdio.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

#define sigma 10.
#define rho 28.
#define beta (8./3.)

void lorenz(double *u, double *F)
 {
  F[0]=sigma*(u[1]-u[0]);
  F[1]=rho*u[0]-u[1]-u[0]*u[2];
  F[2]=u[0]*u[1]-beta*u[2];

  return;
 }

void dlorenz(double *u, double *dF)
 {
  dF[0+3*0]=-sigma;
  dF[0+3*1]=sigma;
  dF[0+3*2]=0.;

  dF[1+3*0]=rho-u[2];
  dF[1+3*1]=-1.;
  dF[1+3*2]=-u[0];

  dF[2+3*0]=u[1];
  dF[2+3*1]=u[0];
  dF[2+3*2]=-beta;

  return;
 }

void ddlorenz(double *u, double *ddF)
 {
  ddF[0+3*(0+3*0)]=0.;
  ddF[0+3*(1+3*0)]=0.;
  ddF[0+3*(2+3*0)]=0.;

  ddF[1+3*(0+3*0)]=0.;
  ddF[1+3*(1+3*0)]=0.;
  ddF[1+3*(2+3*0)]=-1;

  ddF[2+3*(0+3*0)]=0.;
  ddF[2+3*(1+3*0)]=1.;
  ddF[2+3*(2+3*0)]=0.;

  ddF[0+3*(0+3*1)]=0.;
  ddF[0+3*(1+3*1)]=0.;
  ddF[0+3*(2+3*1)]=0.;

  ddF[1+3*(0+3*1)]=0.;
  ddF[1+3*(1+3*1)]=0.;
  ddF[1+3*(2+3*1)]=0.;

  ddF[2+3*(0+3*1)]=1.;
  ddF[2+3*(1+3*1)]=0.;
  ddF[2+3*(2+3*1)]=0.;

  ddF[0+3*(0+3*2)]=0.;
  ddF[0+3*(1+3*2)]=0.;
  ddF[0+3*(2+3*2)]=0.;

  ddF[1+3*(0+3*2)]=-1.;
  ddF[1+3*(1+3*2)]=0.;
  ddF[1+3*(2+3*2)]=0.;

  ddF[2+3*(0+3*2)]=0.;
  ddF[2+3*(1+3*2)]=0.;
  ddF[2+3*(2+3*2)]=0.;

  return;
 }

void dddlorenz(double *u, double *dddF)
 {
  int i;

  for(i=0;i<3*3*3*3;i++)dddF[i]=0.;
  return;
 }
