/* 
    @(#)MFAnimate.c	1.6
    03/02/19 10:59:22
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFImplicitMF.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFPrint.h>
#include <stdio.h>
#include <MFTime.h>
#include <MFDraw.h>
#include <MFEnumPolytope.h>
#include <MFEnumDualPolytope.h>
#include <math.h>
#include <sh.h>
#include <string.h>

MFAtlas MFAnimateAtlas(MFImplicitMF M, MFNRegion Omega, MFNVector u0, double epsilon, int kmax, int verbose, char *stub, int DrawSimp, int DrawAfter, int DrawEvery)
 {
  int kkk;
  double delta=0.;
  int chart;
  int n;
  MFAtlas S;
  MFNVector u;
  MFNKMatrix Phi;
  FILE *fid;
  double volume;
  MFEnumDualPolytope P;
  int Draw;
  char *filename="";

#ifdef MFTIMINGS
  MFTimeReset();
#endif

/*shSetOutputResolution(4096,4096);*/
/*shSetOutputResolution(8192,8192);*/
  shSetOutputResolution(800,800);
  filename=(char*)malloc((strlen(stub)+5)*sizeof(char));
  shSetOutputFormat("tiff");
  MFDrawInitializeFromFile(stub);
  if(MFIMF_N(M)==MFIMF_K(M) && MFIMF_N(M)==2){
   float a,d,s;
   int n;
   a=1.;d=0.;s=0.;n=1;
   shsrfp(&a,&d,&s,&n);
  }

  kkk=0;
  S=MFCreateAtlas(M);
  MFAtlasSetEpsilon(S,epsilon);
  if(verbose){printf("%d) Add initial point to Atlas ",kkk);MFPrintNVector(stdout,u0);printf("\n");fflush(stdout);}
  MFAtlasAddChart(S,u0);
  if(kkk>=DrawAfter)
   {
    sprintf(filename,"%s%4.4d",stub,0);shSetOutputFilename(filename);
    MFDrawAtlasOnce(S);MFDrawDisplay();
   }
#ifdef MFVOLUME
  volume=MFVolumeOfAtlas(S,Omega);
  if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
  kkk++;

#ifdef MFTIMINGS
  MFTimeReport(stdout);
  MFTimeReset();
#endif

  n=MFAtlasN(S);
  u=MFCreateNVector(n);
  while((kmax<0||kkk<kmax)&&(chart=MFAtlasPointOnBoundaryInsideRegion(S,Omega,u,&Phi,&delta))>-1)
   {
    if(verbose){printf("%d) Add boundary point to Atlas ",kkk);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);}
    MFAtlasAddChartWithApproxTS(S,u,MFAtlasChartTangentSpace(S,chart),delta,MFAtlasChartRadius(S,chart));
#ifdef MFVOLUME
    volume=MFVolumeOfAtlas(S,Omega);
    if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
#ifdef MFTIMINGS
    MFTimeReport(stdout);
    MFTimeReset();
#endif
    if(kkk%DrawEvery==0 && kkk>=DrawAfter)
     {
      MFDrawClear();
      MFDrawAtlasOnce(S);
      if(DrawSimp)
       {
        P=MFEnumDualOfAtlas(S);
        MFDrawEnumDualPolytopeEdges(P);
        MFFreeEnumDualPolytope(P);
       }
      sprintf(filename,"%s%4.4d",stub,kkk);shSetOutputFilename(filename);
      MFDrawDisplay();
     }
    kkk++;
/*  if(kkk%1000==0)
     {
      printf("Check pointing, step %d\n",kkk);fflush(stdout);
      fid=fopen("Checkpoint.atlas","w");
      MFWriteAtlas(fid,S);
      fclose(fid);
      printf("Done check pointing\n");fflush(stdout);
     }*/

    MFFreeNVector(u);
    u=MFCreateNVector(n);
   }
  if(verbose&&chart<0){printf("No more points on the boundary\n");fflush(stdout);}
   else if(verbose){printf("Maximum number of charts added (%d)\n",kkk);fflush(stdout);}

  MFDrawClear();
  MFDrawAtlasOnce(S);
  if(DrawSimp)
   {
    P=MFEnumDualOfAtlas(S);
    MFDrawEnumDualPolytopeEdges(P);
    MFFreeEnumDualPolytope(P);
   }
  sprintf(filename,"%s%4.4d",stub,kkk);shSetOutputFilename(filename);
  MFDrawDisplay();
  MFDrawClose();

  return S;
 }
