/* 
    %W%
    %D% %T%

    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFImplicitMF.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFPrint.h>
#include <stdio.h>
#include <MFTime.h>
void MFAtlasPageOutChartsNotNearBoundary(MFAtlas,int,int,char*);
void MFAtlasWriteOutChartsNotPaged(MFAtlas,int,int,char*);
void MFReportMemory();
int MFNVectorGetNRefs(MFNVector);
void MFHendersonMethodCopyClipF(MFContinuationMethod,MFAtlas);
MFChart MFAtlasChart(MFAtlas,int);
void MFUpdateNeighborsReferenceMarks(MFAtlas,int,int);

static FILE *restartReg=(FILE*)NULL;
static FILE *restartSing=(FILE*)NULL;
static double *x=(double*)NULL;

MFAtlas MFComputeAtlas(MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, MFNVector u0)
 {
  static char RoutineName[]={"MFComputeAtlas"};
  MFAtlas S;

  S=MFCreateAtlas(M);
  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,1,&u0,(MFNKMatrix*)NULL);
  return S;
 }

MFAtlas MFComputeAtlasWithTangent(MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, MFNVector u0, MFNKMatrix Phi0)
 {
  static char RoutineName[]={"MFComputeAtlasWithTangent"};
  MFAtlas S;

  S=MFCreateAtlas(M);
  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,1,&u0,&Phi0);
  
  return S;
 }

MFAtlas MFComputeAtlasMultiple(MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0)
 {
  static char RoutineName[]={"MFComputeAtlasMultiple"};
  MFAtlas S;

  S=MFCreateAtlas(M);
  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,m,u0,(MFNKMatrix*)NULL);
  return S;
 }

MFAtlas MFComputeAtlasMultipleWithTangents(MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0, MFNKMatrix *Phi0)
 {
  static char RoutineName[]={"MFComputeAtlasMultipleWithTangents"};
  MFAtlas S;

  S=MFCreateAtlas(M);
  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,m,u0,Phi0);
  return S;
 }

void MFExtendAtlas(MFAtlas S, MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, MFNVector u0)
 {
  static char RoutineName[]={"MFExtendAtlas"};

  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,1,&u0,(MFNKMatrix*)NULL);
  return;
 }

void MFExtendAtlasMultiple(MFAtlas S, MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0)
 {
  static char RoutineName[]={"MFExtendAtlasMultiple"};
  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,m,u0,(MFNKMatrix*)NULL);
  return;
 }

void MFExtendAtlasWithTangent(MFAtlas S, MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, MFNVector u0,MFNKMatrix Phi0)
 {
  static char RoutineName[]={"MFExtendAtlasWithTangent"};

  MFExtendAtlasMultipleWithTangents(S,H,M,Omega,1,&u0,&Phi0);
  return;
 }

void MFExtendAtlasMultipleWithTangents(MFAtlas S, MFContinuationMethod H, MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0, MFNKMatrix *Phi0)
 {
  static char RoutineName[]={"MFExtendAtlasMultipleWithTangents"};
  MFContinuationMethodExtendAtlasMultiple(H,S,M,Omega,m,u0,Phi0);
  return;
 }

void MFFlushAtlas(MFContinuationMethod H, MFAtlas S)
 {
  static char RoutineName[]={"MFFlushAtlas"};

  MFContinuationMethodFlushAtlas(H,S);
  return;
 }

void MFCloseAtlas(MFContinuationMethod H, MFAtlas S)
 {
  static char RoutineName[]={"MFCloseAtlas"};

  MFContinuationMethodCloseAtlas(H,S);
  return;
 }

void MFHendersonExtendAtlasMultiple(MFContinuationMethod H, MFAtlas S, MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0, MFNKMatrix *Phi0)
 {
  static char RoutineName[]={"MFExtendAtlasMultiple"};
  int i,j;
  int kkk;
  double delta=0.;
  int chart;
  int n,N;
  MFNVector u;
  MFNKMatrix Phi;
  double R,Rs;
  FILE *fid;
  double volume;
  double epsilon;
  double dotmin;
  int kmax;
  int page,pageEvery;
  int checkPoint,checkPointEvery;
  int branchSwitch;
  int dumpToPlotFile;
  int dumpToCenterFile;
  int rc;
  int verbose;
  int switched;
  int dumpToRestartFile;
  int dumpToRestartFileEvery;
  char *name;
  char FileName[4096];

#ifdef MFTIMINGS
  MFTimeReset();
#endif

/* If M!=AtlasM change it. What if operations between n vectors don't work? */

  if(m==0)return;

  epsilon=MFHendersonsMethodGetEpsilon(H);
  dotmin=MFHendersonsMethodGetDotMin(H);
  kmax=MFHendersonsMethodGetMaxCharts(H);
  verbose=MFHendersonsMethodGetVerbose(H);
  page=MFHendersonsMethodGetPage(H);
  pageEvery=MFHendersonsMethodGetPageEvery(H);
  checkPoint=MFHendersonsMethodGetCheckPoint(H);
  checkPointEvery=MFHendersonsMethodGetCheckPointEvery(H);
  branchSwitch=MFHendersonsMethodGetBranchSwitch(H);
  dumpToPlotFile=MFHendersonsMethodGetDumpToPlotFile(H);
  dumpToCenterFile=MFHendersonsMethodGetDumpToCenterFile(H);
  name=MFHendersonsMethodGetFilename(H);
  switched=0;
  MFHendersonMethodCopyClipF(H,S);
  dumpToRestartFile=MFHendersonsMethodGetDumpToRestartFile(H);
  dumpToRestartFileEvery=MFHendersonsMethodGetDumpToRestartFileEvery(H);

  if(verbose){printf("%s, epsilon=%lf, kmax=%d, verbose=%d\n",RoutineName,epsilon,kmax,verbose);fflush(stdout);}

  kkk=0;
  MFAtlasSetEpsilon(S,epsilon);
  MFAtlasSetDotMin(S,dotmin);
  MFAtlasSetVerbose(S,verbose);
  for(i=0;i<m;i++)
   {
    if(Phi0==(MFNKMatrix*)NULL||Phi0[i]==(MFNKMatrix)NULL)
      rc=MFAtlasAddChart(S,u0[i]);
     else{
      R=MFIMFScale(M,u0[i],Phi0[i]);
      rc=MFAtlasAddChartWithAll(S,u0[i],Phi0[i],R);
     }
    if(verbose){printf("%d) Add initial points to Atlas ",rc);MFPrintNVector(stdout,u0[i]);printf(" index %d, bif? %d\n",MFNVGetIndex(u0[i]),MFNVGetIndex2(u0[i]));fflush(stdout);}
    MFChartSetReferenceNumber(MFAtlasChart(S,rc),0);
#ifdef MFVOLUME
    volume=MFVolumeOfAtlas(S,Omega);
    if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
    kkk++;

    if(MFChartReferenceNumber(MFAtlasChart(S,rc))>dumpToRestartFileEvery)
     {
      printf("Reference Number has exceeded the threshold\n");fflush(stdout);
      MFChartSetReferenceNumber(MFAtlasChart(S,rc),0);
      MFUpdateNeighborsReferenceMarks(S,rc,dumpToRestartFileEvery);
     }
/*  printf(" Reference Number is %d\n",MFChartReferenceNumber(MFAtlasChart(S,rc)));
    printf(" Index Number is %d\n",MFChartReferenceNumber(MFAtlasChart(S,rc)));*/
    if(dumpToRestartFile && MFChartReferenceNumber(MFAtlasChart(S,rc))==0)
     {
      if(MFChartIsSingular(MFAtlasChart(S,rc)))
       {
        if(restartSing==(FILE*)NULL)
         {
          strcpy(FileName,name);
          strcat(FileName,".singular");
          restartSing=fopen(FileName,"w");
          if(restartSing==(FILE*)NULL)abort();
         }
        N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
        if(x==(double*)NULL)
          x=(double*)malloc(N*sizeof(double));
        MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
        for(j=0;j<N;j++)fprintf(restartSing," %lf",x[j]);
        fprintf(restartSing,"\n");
       }else{
        if(restartReg==(FILE*)NULL)
         {
          strcpy(FileName,name);
          strcat(FileName,".regular");
          restartReg=fopen(FileName,"w");
          if(restartReg==(FILE*)NULL)abort();
         }
        N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
        if(x==(double*)NULL)
          x=(double*)malloc(N*sizeof(double));
        MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
        for(j=0;j<N;j++)fprintf(restartReg," %lf",x[j]);
        fprintf(restartReg,"\n");
       }
     }

    if(page && kkk%pageEvery==0)
     {
      if(verbose){printf("Before Paging\n");MFReportMemory();fflush(stdout);}
      MFAtlasPageOutChartsNotNearBoundary(S,dumpToPlotFile,dumpToCenterFile,name);
      if(verbose){printf("After Paging\n");MFReportMemory();fflush(stdout);}
     }
    if(checkPoint && kkk%checkPointEvery==0)
     {
      if(verbose){printf("Check pointing, step %d\n",kkk);fflush(stdout);}
      fid=fopen("Checkpoint.atlas","w");
      MFWriteAtlas(fid,S);
      fclose(fid);
     }
   }

#ifdef MFTIMINGS
  MFTimeReport(stdout);
  MFTimeReset();
#endif

  n=MFAtlasN(S);
  u=MFCloneNVector(u0[0]);

GoOn:
  rc=1;
  while(rc>0&&(kmax<0||kkk<kmax)&&(chart=MFAtlasPointOnBoundaryInsideRegion(S,Omega,u,&Phi,&delta))>-1)
   {
    R=MFIMFScale(MFAtlasMF(S),u,Phi);
    Rs=2*MFAtlasChartSuggestedRadius(S,chart);
    if(0&&R>Rs)R=Rs;

    if((rc=MFAtlasAddChartWithAll(S,u,Phi,R))>0)
     {
      if(verbose){printf("%d) Add boundary point to Atlas ",rc);MFPrintNVector(stdout,u);printf(" index %d, bif? %d\n",MFNVGetIndex(u),MFNVGetIndex2(u));fflush(stdout);}
      if(MFChartReferenceNumber(MFAtlasChart(S,rc))>dumpToRestartFileEvery)
       {
        printf("Reference Number has exceeded the threshold\n");fflush(stdout);
        MFChartSetReferenceNumber(MFAtlasChart(S,rc),0);
        MFUpdateNeighborsReferenceMarks(S,rc,dumpToRestartFileEvery);
       }
/*    printf(" Reference Number is %d\n",MFChartReferenceNumber(MFAtlasChart(S,rc)));*/
      if(dumpToRestartFile && MFChartReferenceNumber(MFAtlasChart(S,rc))==0)
       {
        if(MFChartIsSingular(MFAtlasChart(S,rc)))
         {
          if(restartSing==(FILE*)NULL)
           {
            strcpy(FileName,name);
            strcat(FileName,".singular");
            restartSing=fopen(FileName,"w");
            if(restartSing==(FILE*)NULL)abort();
           }
          N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
          if(x==(double*)NULL)
            x=(double*)malloc(N*sizeof(double));
          MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
          for(j=0;j<N;j++)fprintf(restartSing," %lf",x[j]);
          fprintf(restartSing,"\n");
         }else{
          if(restartReg==(FILE*)NULL)
           {
            strcpy(FileName,name);
            strcat(FileName,".regular");
            restartReg=fopen(FileName,"w");
            if(restartReg==(FILE*)NULL)abort();
           }
          N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
          if(x==(double*)NULL)
            x=(double*)malloc(N*sizeof(double));
          MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
          for(j=0;j<N;j++)fprintf(restartReg," %lf",x[j]);
          fprintf(restartReg,"\n");
         }
       }
#ifdef MFVOLUME
      volume=MFVolumeOfAtlas(S,Omega);
      if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
#ifdef MFTIMINGS
      MFTimeReport(stdout);
      MFTimeReset();
#endif
      kkk++;
      if(page && kkk%pageEvery==0)
       {
        if(verbose){printf("Before Paging\n");MFReportMemory();fflush(stdout);}
        MFAtlasPageOutChartsNotNearBoundary(S,dumpToPlotFile,dumpToCenterFile,name);
        if(verbose){printf("After Paging\n");MFReportMemory();fflush(stdout);}
       }
      if(checkPoint && kkk%checkPointEvery==0)
       {
        if(verbose){printf("Check pointing, step %d\n",kkk);fflush(stdout);}
        fid=fopen("Checkpoint.atlas","w");
        MFWriteAtlas(fid,S);
        fclose(fid);
       }
     }

    MFFreeNVector(u);
    MFFreeNKMatrix(Phi);
    u=MFCloneNVector(u0[0]);
   }

  if(verbose&&chart<0){printf("No more points on the boundary, switch=%d, switched=%d\n",branchSwitch,switched);fflush(stdout);}
   else if(verbose){printf("Maximum number of charts added (%d)\n",kkk);fflush(stdout);}

  if((switched<branchSwitch) && branchSwitch!=0 && chart<0)
   {
    int schart,bchart;
    MFNVector u1,u;
    MFNKMatrix Phi;

    schart=MFAtlasGetSingularChartWithBoundary(S,Omega);
    if(schart>0)
     {
      if(verbose){printf("Branch switching, singular chart with boundary is %d, index %d \n",schart,MFNVGetIndex2(MFAtlasChartCenter(S,schart)));fflush(stdout);}

/* This is the branch that passes on through. */

      u1=MFAtlasGetPointOnBoundaryChart(S,Omega,schart,1.);
      if(u1==(MFNVector)NULL)goto GoOn;
/*    MFAtlasAddChart(S,u1);*/
      Phi=MFAtlasChartTangentSpace(S,schart);
      R=MFIMFScale(MFAtlasMF(S),u1,Phi);
      MFIMFSetStability(MFAtlasMF(S),u1,Phi);
      bchart=MFAtlasAddChartWithAll(S,u1,Phi,R);
      if(verbose){printf("%d) Add bifurcating chart to Atlas ",bchart);MFPrintNVector(stdout,u1);printf(" index %d, bif? %d\n",MFNVGetIndex(u1),MFNVGetIndex2(u1));fflush(stdout);}
/*    MFFreeNKMatrix(Phi);*/
      if(MFChartReferenceNumber(MFAtlasChart(S,rc))>dumpToRestartFileEvery)
       {
        printf("Reference Number has exceeded the threshold\n");fflush(stdout);
        MFChartSetReferenceNumber(MFAtlasChart(S,rc),0);
        MFUpdateNeighborsReferenceMarks(S,rc,dumpToRestartFileEvery);
       }
      printf(" Reference Number is %d\n",MFChartReferenceNumber(MFAtlasChart(S,rc)));
      if(dumpToRestartFile && MFChartReferenceNumber(MFAtlasChart(S,rc))==0)
       {
        if(MFChartIsSingular(MFAtlasChart(S,rc)))
         {
          if(restartSing==(FILE*)NULL)
           {
            strcpy(FileName,name);
            strcat(FileName,".singular");
            restartSing=fopen(FileName,"w");
            if(restartSing==(FILE*)NULL)abort();
           }
          N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
          if(x==(double*)NULL)
          x=(double*)malloc(N*sizeof(double));
          MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
          for(j=0;j<N;j++)fprintf(restartSing," %lf",x[j]);
          fprintf(restartSing,"\n");
         }else{
          if(restartReg==(FILE*)NULL)
           {
            strcpy(FileName,name);
            strcat(FileName,".regular");
            restartReg=fopen(FileName,"w");
            if(restartReg==(FILE*)NULL)abort();
           }
          N=MFIMFProjectToDraw(MFAtlasMF(S),(MFNVector)NULL,(double*)NULL);
          if(x==(double*)NULL)
            x=(double*)malloc(N*sizeof(double));
          MFIMFProjectToDraw(MFAtlasMF(S),MFChartCenter(MFAtlasChart(S,rc)),x);
          for(j=0;j<N;j++)fprintf(restartReg," %lf",x[j]);
          fprintf(restartReg,"\n");
         }
       }
      switched+=1;
      goto GoOn;
     }
   }

  MFFreeNVector(u);

  return;
 }

void MFHendersonCloseAtlas(MFContinuationMethod H, MFAtlas S)
 {
  static char RoutineName[]={"MFCloseAtlas"};
  int page;
  int dumpToPlotFile;
  int dumpToCenterFile;
  char *name;

  page=MFHendersonsMethodGetPage(H);
  dumpToPlotFile=MFHendersonsMethodGetDumpToPlotFile(H);
  dumpToCenterFile=MFHendersonsMethodGetDumpToCenterFile(H);
  name=MFHendersonsMethodGetFilename(H);

  if(dumpToPlotFile)MFAtlasWriteOutChartsNotPaged(S,dumpToPlotFile,dumpToCenterFile,name);
  if(restartSing!=(FILE*)NULL){fclose(restartSing);restartSing=(FILE*)NULL;}
  if(restartReg!=(FILE*)NULL){fclose(restartReg);restartReg=(FILE*)NULL;}
  if(x!=(double*)NULL){free(x);x=(double*)NULL;}

  MFAtlasClosePlotfile(S);
  MFAtlasCloseCenterfile(S);

  return;
 }

void MFHendersonFlushAtlas(MFContinuationMethod H, MFAtlas S)
 {
  static char RoutineName[]={"MFFlushAtlas"};
  int page;
  int dumpToPlotFile;
  int dumpToCenterFile;
  char *name;

  page=MFHendersonsMethodGetPage(H);
  dumpToPlotFile=MFHendersonsMethodGetDumpToPlotFile(H);
  dumpToCenterFile=MFHendersonsMethodGetDumpToCenterFile(H);
  name=MFHendersonsMethodGetFilename(H);

  if(page)MFAtlasWriteOutChartsNotPaged(S,dumpToPlotFile,dumpToCenterFile,name);

  return;
 }

void MFUpdateNeighborsReferenceMarks(MFAtlas A, int ichart, int maxdepth)
 {
  static char RoutineName[]={"MFUpdateNeighborsReferenceMarks"};
  int i,n,hs,j,l;
  int minmark,mark;
  int neighbor;
  MFPolytope P;
  MFChart chart;
  MFChart nchart;
  static int verbose=0;
  int nNeighs;
  int mNeighs;
  int *neigh;
  int *bneigh;
  int N;
  int found;
  int depth;

  if(verbose){printf("%s: chart=%d, maxdepth=%d\n",RoutineName,ichart,maxdepth);fflush(stdout);}

  nNeighs=1;
  mNeighs=100;
  neigh=(int*)malloc(mNeighs*sizeof(int));
  bneigh=(int*)malloc(mNeighs*sizeof(int));
  if(neigh==(int*)NULL)abort();
  if(bneigh==(int*)NULL)abort();
  neigh[0]=ichart;
  bneigh[0]=1;

  for(depth=0;depth<maxdepth;depth++)
   {
    N=nNeighs;
    for(i=0;i<N;i++)
     {
      if(bneigh[i])
       {
        chart=MFAtlasChart(A,neigh[i]);
        P=MFChartPolytope(chart);
        if(P!=(MFPolytope)NULL)
         {
          n=MFPolytopeNumberOfFaces(P);
          bneigh[i]=0;
          for(j=0;j<n;j++)
           {
            hs=MFPolytopeFaceIndex(P,j);
            neighbor=-1;
            if(hs>=MFAtlasOffset(A) && hs< MFAtlasNumberOfHalfSpaces(A))neighbor=MFAtlasRightPolytope(A,hs);
  
            if(neighbor>-1 && neighbor<MFAtlasNumberOfCharts(A))
             {
              found=0;
              for(l=0;l<nNeighs;l++)if(neigh[l]==neighbor)found=1;
              if(!found)
               {
                if(nNeighs+1>mNeighs)
                 {
                  mNeighs+=100;
                  neigh=(int*)realloc((void*)neigh,mNeighs*sizeof(int));
                  bneigh=(int*)realloc((void*)bneigh,mNeighs*sizeof(int));
                  if(neigh==(int*)NULL)abort();
                  if(bneigh==(int*)NULL)abort();
                 }
                neigh[nNeighs]=neighbor;
                bneigh[nNeighs]=1;
                nNeighs++;
               }
             }
           }
         }
       }
     }
   }

  if(verbose){printf("  %d neighbors\n",nNeighs);fflush(stdout);}
  for(i=1;i<nNeighs;i++)
   {
    if(verbose){printf("  %d: chart=%d\n",i,neigh[i]);fflush(stdout);}
    chart=MFAtlasChart(A,neigh[i]);
    P=MFChartPolytope(chart);
    if(P!=(MFPolytope)NULL)
     {
      n=MFPolytopeNumberOfFaces(P);
      minmark=-1;
      for(j=0;j<n;j++)
       {
        hs=MFPolytopeFaceIndex(P,j);
        
        neighbor=-1;
        if(hs>MFAtlasOffset(A) && hs< MFAtlasNumberOfHalfSpaces(A))neighbor=MFAtlasRightPolytope(A,hs);
    
        if(neighbor>-1 && neighbor<MFAtlasNumberOfCharts(A))
         {
          nchart=MFAtlasChart(A,neighbor);
          mark=MFChartReferenceNumber(nchart);
          if((MFChartIsSingular(chart)&&MFChartIsSingular(nchart))||(!MFChartIsSingular(chart)&&!MFChartIsSingular(nchart)))
           {
            if(minmark==-1||mark<minmark)minmark=mark;
           }
         }
       }
      if(minmark!=-1)
       {
        mark=MFChartReferenceNumber(chart);
        if(verbose){printf("    old mark:%d new mark:%d\n",mark,minmark+1);fflush(stdout);}
        if(minmark+1<mark)MFChartSetReferenceNumber(chart,minmark+1);
       }
     }
   }

  free(neigh);
  free(bneigh);
  return;
 }
