/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFChart.h>
#include <MFImplicitMF.h>
#include <MFError.h>
#include <MFNVector.h>
#include <MFKVector.h>
#include <MFEnumPolytope.h>
#include <math.h>
#include <stdlib.h>
#include <sh.h>
#include <MFPrint.h>
#include <MFDraw.h>
#include <MFDrawClippedSphere.h>
#include <string.h>
#include <MFDrawChartState.h>
#include <MFEnumPolytope.h>
#include <MFEnumDualPolytope.h>

void MFSetError(int,char*,char*,int,char*);
static char MFDrawErrorMsg[256]="";
MFAtlas MFMFA;
static int MFPendulaPeriodicFlag=0;

#define PERIODIC
#define sqrt3 1.73205080
#define MFPI 3.14159265358979323846264338327950288

void MFDrawLineOnChart(MFChart,MFKVector,MFKVector,MFChartState);
void MFDrawLineOnChartTS(MFChart,MFKVector,MFKVector);
void MFDrawLine(MFNVector,MFNVector,MFChartState);
void MFMarkPoint(MFChart,MFNVector,int,MFChartState);
void MFMarkPointOnChart(MFChart,MFKVector,int,MFChartState);
void MFDrawProjectTo3d(MFChart,MFNVector,float*,float*,float*);
void MFDrawPolytope(MFPolytope,MFChart,MFChartState);
/* void MFDrawEnumPolytope(MFEnumPolytope,MFChart,MFChartState); */
/* void MFDrawEnumDualPolytope(MFEnumDualPolytope); */
/* void MFDrawEnumDualPolytopeEdges(MFEnumDualPolytope); */
MFChart MFAtlasChart(MFAtlas,int);
void MFDrawMakeListOfTriangles(MFChart,int*,int*,double**,double);
double *MFNV_CStar(MFNVector);
double *MFKV_CStar(MFKVector);
void MFDrawChartTS(MFChart);
void MFDraw1dChartTS(MFChart);
void MFDraw2dChartTS(MFChart);
void MFDrawChartBoundaryTS(MFChart);
void MFDraw1dChartBoundaryTS(MFChart);
void MFDraw2dChartBoundaryTS(MFChart);
void MFDrawPolytopeTS(MFPolytope,MFChart);
void MFDrawEnumPolytopeTS(MFEnumPolytope,MFChart);

MFKVector MFTmpK=(MFKVector)NULL;
MFNVector MFTmpN=(MFNVector)NULL;

int MFDrawSetExt=1;
double MFDrawXMax=-1.e20;
double MFDrawXMin= 1.e20;
double MFDrawYMax=-1.e20;
double MFDrawYMin= 1.e20;
double MFDrawZMax=-1.e20;
double MFDrawZMin= 1.e20;

double MFDrawCubeXMax=1.;
double MFDrawCubeXMin=-1.;
double MFDrawCubeYMax=1.;
double MFDrawCubeYMin=-1.;
double MFDrawCubeZMax=1.;
double MFDrawCubeZMin=-1.;
int MFDrawCube=0;

void MFDrawInitialize(float alpha, float beta)
 {
  static char RoutineName[]={"MFDrawInitialize"};
  int grey=100;
  int full=255;
  float dist,xmin,xmax;

/*shinit(&grey,&grey,&grey);*/
  shinit(&full,&full,&full);
  dist=100.;
  xmin=-1.;
  xmax= 1.;
  shview(&dist,&alpha,&beta,&xmin,&xmax,&xmin,&xmax,&xmin,&xmax);
  MFDrawCube=0;
  return;
 }

void MFDrawInitializeFromFile(char *name)
 {
  float alpha=0.;
  float beta=0.;
  float xmin=0.;
  float xmax=0.;
  float ymin=0.;
  float ymax=0.;
  float zmin=0.;
  float zmax=0.;
  int cube=0;
  FILE *fid=(FILE*)NULL;
  char *fullname=(char*)NULL;
  int zero=0;
  int one=1;
  float x=0.;
  float y=0.;
  float z=0.;
  float rone=1.;
  float rzero=0.;
  printf("MFDrawInitializeFromFile\n");fflush(stdout);

  fullname=(char*)malloc((strlen(name)+1+5)*sizeof(char));
  strcpy(fullname,name);
  strcat(fullname,".view");
  fid=fopen(fullname,"r");
  free(fullname);
  if(fid==(FILE*)NULL)
   {
    alpha=0.;
    beta=0.;
    xmin=-1.;
    xmax= 1.;
    ymin=-1.;
    ymax= 1.;
    zmin=-1.;
    zmax= 1.;
    cube= 0;
    printf("Couldn't find file, using defaults\n");
   }else{
    fscanf(fid,"%f %f %f %f %f %f %f %f %d",&alpha,&beta,&xmin,&xmax,&ymin,&ymax,&zmin,&zmax,&cube);
    fclose(fid);
   }
  printf("%f %f %f %f %f %f %f %f %d\n",alpha,beta,xmin,xmax,ymin,ymax,zmin,zmax,cube);fflush(stdout);

  x=0.;
  y=0.;
  z=0.;

  if(!cube)MFDrawInitializeNoCube(alpha,beta,xmin,xmax,ymin,ymax,zmin,zmax);
   else MFDrawInitializeCube(alpha,beta,xmin,xmax,ymin,ymax,zmin,zmax);

  MFDrawCube=cube;

  return;
 }

void MFDrawInitializeCube(float alpha, float beta, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax)
 {
  static char RoutineName[]={"MFDrawInitializeCube"};
  int grey=100;
  int full=255;
  int zero=0;
  float dist;

/*shinit(&grey,&grey,&grey);*/
  shinit(&full,&full,&full);
  dist=34.*(xmax-xmin+ymax-ymin+zmax-zmin);
  shview(&dist,&alpha,&beta,&xmin,&xmax,&ymin,&ymax,&zmin,&zmax);
  shlinc(&zero,&zero,&zero);
  shcube(&xmin,&xmax,&ymin,&ymax,&zmin,&zmax);
  MFDrawCubeXMax=xmax;
  MFDrawCubeXMin=xmin;
  MFDrawCubeYMax=ymax;
  MFDrawCubeYMin=ymin;
  MFDrawCubeZMax=zmax;
  MFDrawCubeZMin=zmin;
  MFDrawCube=1;
  return;
 }

void MFDrawInitializeNoCube(float alpha, float beta, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax)
 {
  static char RoutineName[]={"MFDrawInitializeCube"};
  int grey=100;
  int full=255;
  int zero=0;
  float dist;

/*shinit(&grey,&grey,&grey);*/
  shinit(&full,&full,&full);
  dist=100.;
  shview(&dist,&alpha,&beta,&xmin,&xmax,&ymin,&ymax,&zmin,&zmax);
  shlinc(&zero,&zero,&zero);
  MFDrawCube=0;
  return;
 }

void MFDrawClose()
 {
  static char RoutineName[]={"MFDrawClose"};
  shend();
  MFFreeDrawCharts();
  if(MFTmpK!=(MFKVector)NULL)MFFreeKVector(MFTmpK);
  if(MFTmpN!=(MFNVector)NULL)MFFreeNVector(MFTmpN);
  return;
 }

void MFDrawClear()
 {
  static char RoutineName[]={"MFDrawClear"};

  shclr();

  if(MFDrawCube)
   {
    float xmin,xmax,ymin,ymax,zmin,zmax;
    int zero=0;
    shlinc(&zero,&zero,&zero);
    xmin=MFDrawCubeXMin;
    xmax=MFDrawCubeXMax;
    ymin=MFDrawCubeYMin;
    ymax=MFDrawCubeYMax;
    zmin=MFDrawCubeZMin;
    zmax=MFDrawCubeZMax;
    shcube(&xmin,&xmax,&ymin,&ymax,&zmin,&zmax);
   }

  return;
 }

void MFDrawDisplay()
 {
  static char RoutineName[]={"MFDrawDisplay"};
  shpause();
  return;
 }

void MFDrawInterval(MFChart chart)
 {
  static char RoutineName[]={"MFDrawInterval"};
  return;
 }

void MFDrawPolygon(MFChart chart)
 {
  static char RoutineName[]={"MFDrawPolygon"};
  return;
 }

void MFDrawPolyhedron(MFChart chart)
 {
  static char RoutineName[]={"MFDrawPolyhedron"};
  return;
 }

void MFDrawPolytope(MFPolytope P,MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDrawPolytope"};
  MFEnumPolytope EP;

  if(MFPolytopeDimension(P)>3)return;

  EP=MFEnumeratePolytope(P);
  MFDrawEnumPolytope(EP,chart,state);
  MFFreeEnumPolytope(EP);

  return;
 }

void MFDrawPolytopeTS(MFPolytope P,MFChart chart)
 {
  static char RoutineName[]={"MFDrawPolytopeTS"};
  MFEnumPolytope EP;

  if(MFPolytopeDimension(P)>3)return;

  EP=MFEnumeratePolytope(P);
  MFDrawEnumPolytopeTS(EP,chart);
  MFFreeEnumPolytope(EP);

  return;
 }

void MFDraw1dChart(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw1dChart"};
  int color;
  int zero=0;
  int half=128;
  int full=255;
  MFPolytope P;
  MFKVector s0,s1;
  int n,k;

  P=MFChartPolytope(chart);
  k=MFChartK(chart);
  n=MFChartN(chart);
  if(MFPolytopeNumberOfVertices(P)<2)return;
  s0=MFCreateKVector(k);
  s1=MFCreateKVector(k);

  MFPolytopeVertex(P,0,s0);
  MFPolytopeVertex(P,1,s1);

  MFDrawLineOnChart(chart,s0,s1,state);
  MFMarkPointOnChart(chart,s0,2,state);
  MFMarkPointOnChart(chart,s1,2,state);
  MFMarkPoint(chart,MFChartCenter(chart),3,state);

  MFFreeKVector(s0);
  MFFreeKVector(s1);

  return;
 }

void MFDraw2dChart(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw2dChart"};
  int color;
  int zero=0;
  int half=128;
  int full=255;
  double *tri;
  int ntri,mtri;
  static float x[3],y[3],z[3];
  double t;
  MFKVector s;
  MFNVector u;
  int j,n;
  int rc;

/* The interior of the circle */

  tri=(double*)malloc(6*6*sizeof(double)); /*done*/
  if(tri==(double*)NULL)
   {
    sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",6*6*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
    return;
   }

  t=2./sqrt3*MFChartRadius(chart);

  tri[0+6*0]=0.;
  tri[1+6*0]=0.;
  tri[2+6*0]=t;
  tri[3+6*0]=0.;
  tri[4+6*0]=.5*t;
  tri[5+6*0]=.5*sqrt3*t;

  tri[0+6*1]=0.;
  tri[1+6*1]=0.;
  tri[2+6*1]=.5*t;
  tri[3+6*1]=.5*sqrt3*t;
  tri[4+6*1]=-.5*t;
  tri[5+6*1]=.5*sqrt3*t;

  tri[0+6*2]=0.;
  tri[1+6*2]=0.;
  tri[2+6*2]=-.5*t;
  tri[3+6*2]=.5*sqrt3*t;
  tri[4+6*2]=-t;
  tri[5+6*2]=0.;

  tri[0+6*3]=0.;
  tri[1+6*3]=0.;
  tri[2+6*3]=-t;
  tri[3+6*3]=0.;
  tri[4+6*3]=-.5*t;
  tri[5+6*3]=-.5*sqrt3*t;

  tri[0+6*4]=0.;
  tri[1+6*4]=0.;
  tri[2+6*4]=-.5*t;
  tri[3+6*4]=-.5*sqrt3*t;
  tri[4+6*4]= .5*t;
  tri[5+6*4]=-.5*sqrt3*t;

  tri[0+6*5]=0.;
  tri[1+6*5]=0.;
  tri[2+6*5]= .5*t;
  tri[3+6*5]=-.5*sqrt3*t;
  tri[4+6*5]=t;
  tri[5+6*5]=0.;

  ntri=6;
  mtri=6;
  t=.2*MFChartRadius(chart);
  if(t<.01)t=.01;
  MFDrawMakeListOfTriangles(chart,&ntri,&mtri,&tri,t);

  s=MFCreateKVector(MFChartK(chart));
  u=MFCreateNVector(MFChartN(chart));

  for(j=0;j<ntri;j++)
   {
    MFKVSetC(s,0,tri[0+6*j]);
    MFKVSetC(s,1,tri[1+6*j]);
    rc=MFChartEvaluate(chart,s,u);
    MFDrawProjectTo3d(chart,u,x,y,z);
    MFKVSetC(s,0,tri[2+6*j]);
    MFKVSetC(s,1,tri[3+6*j]);
    rc=rc && MFChartEvaluate(chart,s,u);
    MFDrawProjectTo3d(chart,u,x+1,y+1,z+1);
    MFKVSetC(s,0,tri[4+6*j]);
    MFKVSetC(s,1,tri[5+6*j]);
    rc = rc && MFChartEvaluate(chart,s,u);
    MFDrawProjectTo3d(chart,u,x+2,y+2,z+2);

#ifdef PERIODIC
    if(MFPendulaPeriodicFlag)
     {
      rc=rc&&fabs(x[0]-x[1])<.1;
      rc=rc&&fabs(x[1]-x[2])<.1;
      rc=rc&&fabs(x[2]-x[0])<.1;
      rc=rc&&fabs(y[0]-y[1])<.1;
      rc=rc&&fabs(y[1]-y[2])<.1;
      rc=rc&&fabs(y[2]-y[0])<.1;
     }
#endif
    n=3;
    if(rc&&state!=(MFChartState)NULL)MFChartStateAddPolygon(state,n,x,y,z,0);
    if(rc&&state==(MFChartState)NULL)shpg(&n,x,y,z);
   }
  free(tri);
  MFFreeKVector(s);
  MFFreeNVector(u);

  MFDrawPolytope(MFChartPolytope(chart),chart,state);

  return;
 }

void MFDraw3dChart(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw3dChart"};
  int full=255;

  if(MFChartK(chart)>2)shtric(&full,&full,&full);

  MFDrawPolytope(MFChartPolytope(chart),chart,state);

  return;
 }

void MFDraw1dChartBoundary(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw1dChartBoundary"};
  MFKVector s;
  int in0;
  int zero=0;
  int full=255;

  s=MFCreateKVector(MFChartK(chart));
  MFKVSetC(s,0,MFChartRadius(chart));

  in0=MFChartInterior(chart,s);
  if(in0)MFMarkPointOnChart(chart,s,0,state);

  MFKVSetC(s,0,-MFChartRadius(chart));

  in0=MFChartInterior(chart,s);
  if(in0)MFMarkPointOnChart(chart,s,1,state);

  MFFreeKVector(s);
  return;
 }

void MFDraw1dChartBoundaryTS(MFChart chart)
 {
  static char RoutineName[]={"MFDraw1dChartBoundary"};
  MFKVector s;
  MFNVector u;
  int in0;
  int zero=0;
  int full=255;

  s=MFCreateKVector(MFChartK(chart));
  MFKVSetC(s,0,MFChartRadius(chart));

  in0=MFChartInterior(chart,s);
  if(in0)
   {
    u=MFCreateNVector(MFChartN(chart));
    MFChartPointInTangentSpace(chart,s,u);
    MFMarkPoint(chart,u,0,(MFChartState)NULL);
    MFFreeNVector(u);
   }

  MFKVSetC(s,0,-MFChartRadius(chart));

  in0=MFChartInterior(chart,s);
  if(in0)
   {
    u=MFCreateNVector(MFChartN(chart));
    MFChartPointInTangentSpace(chart,s,u);
    MFMarkPoint(chart,u,1,(MFChartState)NULL);
    MFFreeNVector(u);
   }

  MFFreeKVector(s);
  return;
 }

void MFDraw2dChartBoundary(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw2dChartBoundary"};
  MFKVector s;
  MFNVector u;
  int zero=0;
  int full=255;
  int i;
  double r;
  int in0,in1;
  double pt[4]={0.,0.,0.,0.};
  float xa=0;
  float ya=0;
  float za=0.;
  float xb=0;
  float yb=0;
  float zb=0.;
  float x[2]={0.,0.};
  float y[2]={0.,0.};
  float z[2]={0.,0.};
  int rc;
  int n;
  double dt;
  int rr,gg,bb,rs,gs,bs;

/* Draw yellow circle for boundary of ball */

  s=MFCreateKVector(MFChartK(chart));
  u=MFCreateNVector(MFChartN(chart));

  r=MFChartRadius(chart);

  n=400;
  if(r<.05)n=6;
  dt=2./n*3.1415926;
  shqlinc(&rs,&gs,&bs);
  rr=255;
  gg=24;
  bb=47;
  shlinc(&rr,&gg,&bb);

  for(i=0;i<n;i++)
   {
    pt[0]=.999999*r*cos(dt*i);
    pt[1]=.999999*r*sin(dt*i);
    MFKVSetC(s,0,pt[0]);
    MFKVSetC(s,1,pt[1]);
    in0=MFChartInterior(chart,s);
    pt[2]=.999999*r*cos(dt*(i+1));
    pt[3]=.999999*r*sin(dt*(i+1));
    MFKVSetC(s,0,pt[2]);
    MFKVSetC(s,1,pt[3]);
    in1=MFChartInterior(chart,s);

    if(in0&&in1)
     {
      MFKVSetC(s,0,pt[0]);
      MFKVSetC(s,1,pt[1]);
      rc=MFChartEvaluate(chart,s,u);
      MFDrawProjectTo3d(chart,u,&xa,&ya,&za);

      MFKVSetC(s,0,pt[2]);
      MFKVSetC(s,1,pt[3]);
      rc=rc && MFChartEvaluate(chart,s,u);

      MFDrawProjectTo3d(chart,u,&xb,&yb,&zb);

      x[0]=xa;
      y[0]=ya;
      z[0]=za;
      x[1]=xb;
      y[1]=yb;
      z[1]=zb;
#ifdef PERIODIC
      if(MFPendulaPeriodicFlag)
       {
        rc=rc&&fabs(xa-xb)<.1;
        rc=rc&&fabs(ya-yb)<.1;
       }
#endif
      if(rc&&state!=(MFChartState)NULL)MFChartStateAddLine(state,x[0],y[0],z[0],x[1],y[1],z[1]);
      if(rc&&state==(MFChartState)NULL)shline(x,y,z,x+1,y+1,z+1);
     }
   }
  MFFreeKVector(s);
  MFFreeNVector(u);
  shlinc(&rs,&gs,&bs);

  return;
 }

void MFDraw2dChartBoundaryTS(MFChart chart)
 {
  static char RoutineName[]={"MFDraw2dChartBoundaryTS"};
  MFKVector s;
  MFNVector u;
  int zero=0;
  int full=255;
  int i;
  double r;
  int in0,in1;
  double pt[4]={0.,0.,0.,0.};
  float xa=0;
  float ya=0;
  float za=0.;
  float xb=0;
  float yb=0;
  float zb=0.;
  float x[2]={0.,0.};
  float y[2]={0.,0.};
  float z[2]={0.,0.};
  int rc;
  int n;
  double dt;
  int rr,gg,bb,rs,gs,bs;

/* Draw yellow circle for boundary of ball */

  s=MFCreateKVector(MFChartK(chart));
  u=MFCreateNVector(MFChartN(chart));

  r=MFChartRadius(chart);

  n=400;
  if(r<.05)n=6;
  dt=2./n*3.1415926;

  shqlinc(&rs,&gs,&bs);
  rr=255;
  gg=24;
  bb=47;
  shlinc(&rr,&gg,&bb);

  for(i=0;i<n;i++)
   {
    pt[0]=.999999*r*cos(dt*i);
    pt[1]=.999999*r*sin(dt*i);
    MFKVSetC(s,0,pt[0]);
    MFKVSetC(s,1,pt[1]);
    in0=MFChartInterior(chart,s);
    pt[2]=.999999*r*cos(dt*(i+1));
    pt[3]=.999999*r*sin(dt*(i+1));
    MFKVSetC(s,0,pt[2]);
    MFKVSetC(s,1,pt[3]);
    in1=MFChartInterior(chart,s);

    if(in0&&in1)
     {
      MFKVSetC(s,0,pt[0]);
      MFKVSetC(s,1,pt[1]);
      MFChartPointInTangentSpace(chart,s,u);
      MFDrawProjectTo3d(chart,u,&xa,&ya,&za);

      MFKVSetC(s,0,pt[2]);
      MFKVSetC(s,1,pt[3]);
      MFChartPointInTangentSpace(chart,s,u);

      MFDrawProjectTo3d(chart,u,&xb,&yb,&zb);

      x[0]=xa;
      y[0]=ya;
      z[0]=za;
      x[1]=xb;
      y[1]=yb;
      z[1]=zb;
      rc=1;
#ifdef PERIODIC
      if(MFPendulaPeriodicFlag)
       {
        rc=rc&&fabs(xa-xb)<.1;
        rc=rc&&fabs(ya-yb)<.1;
       }
#endif
      if(rc)shline(x,y,z,x+1,y+1,z+1);
     }
   }
  MFFreeKVector(s);
  MFFreeNVector(u);
  shlinc(&rs,&gs,&bs);

  return;
 }

void MFDraw3dChartBoundary(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDraw3dChartBoundary"};
  int zero=0;
  int full=255;
  double e1,e2,e3;

  e1=sqrt(2.)/8.*MFChartRadius(chart)/2.;
  e2=e1/2.;
  e3=e2/50.;
  MFDrawClippedSphere(chart,e1,e2,e3,state);

  return;
 }

void MFTestClippedSphere(MFChart,double,double,double);

void MFTest3dChartBoundary(MFChart chart)
 {
  static char RoutineName[]={"MFTest3dChartBoundary"};
  int zero=0;
  int full=255;
  double e1,e2,e3;

  e1=sqrt(2.)/8.*MFChartRadius(chart)/2.;
  e2=e1/2.;
  e3=e2/50.;
  MFTestClippedSphere(chart,e1,e2,e3);

  return;
 }

void MFDrawChart(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDrawChart"};

  if(chart==(MFChart)NULL)return;
  if(MFChartK(chart)==1)MFDraw1dChart(chart,state);
   else if(MFChartK(chart)==2)MFDraw2dChart(chart,state);
   else if(MFChartK(chart)==3)MFDraw3dChart(chart,state);

  return;
 }

void MFDrawChartBoundaryTS(MFChart chart)
 {
  static char RoutineName[]={"MFDrawChartBoundaryTS"};
  if(chart==(MFChart)NULL)return;
  if(MFChartK(chart)==1)MFDraw1dChartBoundaryTS(chart);
   else if(MFChartK(chart)==2)MFDraw2dChartBoundaryTS(chart);
   else if(MFChartK(chart)==3)MFDraw3dChartBoundary(chart,(MFChartState)NULL);

  return;
 }

void MFDrawChartBoundary(MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDrawChartBoundary"};
  if(chart==(MFChart)NULL)return;
  if(MFChartK(chart)==1)MFDraw1dChartBoundary(chart,state);
   else if(MFChartK(chart)==2)MFDraw2dChartBoundary(chart,state);
   else if(MFChartK(chart)==3)MFDraw3dChartBoundary(chart,state);

  return;
 }

void MFDrawAtlas(MFAtlas A)
 {
  static char RoutineName[]={"MFDrawAtlas"};
  int chart,bchart;
  MFChartState state;
  int rb=243;
  int gb=223;
  int bb=192;
  int verbose;

  verbose=0;

  MFDrawSetExt=1;
  MFDrawXMin=0.;
  MFDrawXMax=0.;
  MFDrawYMin=0.;
  MFDrawYMax=0.;
  MFDrawZMin=0.;
  MFDrawZMax=0.;


  for(chart=0;chart<MFAtlasNumberOfCharts(A);chart++)
   {
    printf("Draw Chart %d\n",chart);fflush(stdout);
    if(chart<MFNChartStates()) /* !!! */
      state=MFGetChartState(chart);
     else{
      state=MFAddChartState(MFAtlasChart(A,chart));
      MFDrawChart(MFAtlasChart(A,chart),state);
      MFMFA=A;
      MFDrawChartBoundary(MFAtlasChart(A,chart),state);
     }
    MFSetChartStateColor(state,0,180,180,205);
    MFSetChartStateColor(state,1,180,180,205);
   }

  for(chart=0;chart<MFAtlasNumberOfCharts(A);chart++)
   {
    if(MFAtlasIsChartNearBoundary(A,chart))
     {
      state=MFGetChartState(chart);
      MFSetChartStateColor(state,0,100,100,150);
      if(verbose){printf("This one is near the boundary, %d\n",chart);fflush(stdout);}
     }
   }

  for(bchart=0;bchart<MFAtlasNumberOfChartsWithBoundary(A);bchart++)
   {
    chart=MFAtlasChartWithBoundary(A,bchart);
    if(bchart<MFNChartStates())
     {
      state=MFGetChartState(chart);
      MFSetChartStateColor(state,0,243,223,192);
      MFSetChartStateColor(state,1,243,223,192);
     }
   }

  for(chart=0;chart<MFAtlasNumberOfCharts(A);chart++)
   {
    state=MFGetChartState(chart);
    if(MFChartStateChanged(state))
     {
      MFClearChartState(state);
      MFDrawChart(MFAtlasChart(A,chart),state);
      MFDrawChartBoundary(MFAtlasChart(A,chart),state);
     }
    MFDrawChartFromState(state);
   }

  printf("at end of MFDrawAtlas, x in (%lf %lf %lf %lf %lf %lf)\n",MFDrawXMin,MFDrawXMax,MFDrawYMin,MFDrawYMax,MFDrawZMin,MFDrawZMax);fflush(stdout);

  return;
 }

void MFDrawAtlasOnce(MFAtlas A)
 {
  static char RoutineName[]={"MFDrawAtlasOnce"};
  int chart,bchart;
  int verbose;
  int ri=62;
  int gi=158;
  int bi=255;
  int zero=0;
  int full=255;
  int rb=255;
  int gb=211;
  int bb=62;
  int r=0;
  int g=0;
  int b=0;
  int rnb=100;
  int gnb=100;
  int bnb=150;
  int rsb=150;
  int gsb= 80;
  int bsb= 80;

  verbose=0;

  shlinc(&zero,&zero,&zero);
  for(chart=0;chart<MFAtlasNumberOfCharts(A);chart++)
   {
    if(verbose){printf("%s chart %d/%d\n",RoutineName,chart,MFAtlasNumberOfCharts(A));fflush(stdout);}
    shtric(&ri,&gi,&bi);
    r=ri;g=gi;b=bi;
    if(MFAtlasIsChartNearBoundary(A,chart))
     {
      shtric(&rnb,&gnb,&bnb);
      r=rnb;g=gnb;b=bnb;
      if(verbose){printf("    This one is not near the boundary, %d\n",chart);fflush(stdout);}
     }
    if(MFAtlasIsChartSingular(A,chart))
     {
      shtric(&rsb,&gsb,&bsb);
      r=rsb;g=gsb;b=bsb;
      if(verbose){printf("    This one is singular, %d\n",chart);fflush(stdout);}
     }
    for(bchart=0;bchart<MFAtlasNumberOfChartsWithBoundary(A);bchart++)
     {
      if(chart==MFAtlasChartWithBoundary(A,bchart))
       {
        shtric(&rb,&gb,&bb);
        r=rb;g=gb;b=bb;
        if(verbose){printf("    This one is on the boundary, %d\n",chart);fflush(stdout);}
       }
     }
    MFDrawChart(MFAtlasChart(A,chart),(MFChartState)NULL);
    shtric(&r,&g,&b);
    if(MFAtlasK(A)<3)shtric(&full,&full,&zero);
    MFDrawChartBoundary(MFAtlasChart(A,chart),(MFChartState)NULL);
   }

  return;
 }

/* ---------------------------------------------------------------------------------*/

void MFDrawMakeListOfTriangles(MFChart chart,int *ntri,int *mtri,double **ptri, double d)
 {
  static char RoutineName[]={"MFDrawMakeListOfTriangles"};
  int i,j,n;
  int in0,in1,in2;
  int allIn;
  int allOut;
  double x0,y0;
  double x1,y1;
  double x2,y2;
  double delta;
  int changed;
  int subdivide;
  int remove;
  double *tri;
  MFKVector pt0;
  MFKVector pt1;
  MFKVector pt2;

  tri=*ptri;

  pt0=MFCreateKVector(MFChartK(chart));
  pt1=MFCreateKVector(MFChartK(chart));
  pt2=MFCreateKVector(MFChartK(chart));

  changed=1;
  while(changed)
   {
    changed=0;
    n=*ntri;
    for(j=0;j<n;j++)
     {
      MFKVSetC(pt0,0,tri[  6*j]);
      MFKVSetC(pt0,1,tri[1+6*j]);
      MFKVSetC(pt1,0,tri[2+6*j]);
      MFKVSetC(pt1,1,tri[3+6*j]);
      MFKVSetC(pt2,0,tri[4+6*j]);
      MFKVSetC(pt2,1,tri[5+6*j]);

      in0=MFChartInterior(chart,pt0);
      in1=MFChartInterior(chart,pt1);
      in2=MFChartInterior(chart,pt2);

      allOut=0;
      allIn=0;
      if(in0&&in1&&in2)allIn=1;
      if(!in0&&!in1&&!in2)allOut=1;

      x0=tri[0+6*j];
      y0=tri[1+6*j];
      x1=tri[2+6*j];
      y1=tri[3+6*j];
      x2=tri[4+6*j];
      y2=tri[5+6*j];

      delta=(sqrt(pow(x1-x0,2)+pow(y1-y0,2))
            +sqrt(pow(x2-x0,2)+pow(y2-y0,2))
            +sqrt(pow(x2-x1,2)+pow(y2-y1,2)))*.3333333;

      subdivide=0;
      remove=0;
      if(delta>d)subdivide=1;
      if(!allOut&&!allIn)subdivide=1;
      if(delta<.25*d)subdivide=0;
      if(allOut)remove=1;

      if(remove && !subdivide)
       {
        changed=1;
        tri[0+6*j]=tri[0+6*(*ntri-1)];
        tri[1+6*j]=tri[1+6*(*ntri-1)];
        tri[2+6*j]=tri[2+6*(*ntri-1)];
        tri[3+6*j]=tri[3+6*(*ntri-1)];
        tri[4+6*j]=tri[4+6*(*ntri-1)];
        tri[5+6*j]=tri[5+6*(*ntri-1)];
        (*ntri)--;
        if(n>*ntri)n--;
       }

      if(subdivide)
       {
        changed=1;
        if(*ntri+3>=*mtri)
         {
          *mtri+=3*(*ntri);
          *ptri=(double*)realloc((void*)(*ptri),6*(*mtri)*sizeof(double)); /*done*/
          if(*ptri==(double*)NULL)
           {
            sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",6*(*mtri)*sizeof(double));
            printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
            MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
            return;
           }
          tri=*ptri;
         }
        tri[0+6*j]=.5*(x0+x1);
        tri[1+6*j]=.5*(y0+y1);
        tri[2+6*j]=.5*(x1+x2);
        tri[3+6*j]=.5*(y1+y2);
        tri[4+6*j]=.5*(x2+x0);
        tri[5+6*j]=.5*(y2+y0);
        tri[0+6*(*ntri)]=x0;
        tri[1+6*(*ntri)]=y0;
        tri[2+6*(*ntri)]=.5*(x0+x1);
        tri[3+6*(*ntri)]=.5*(y0+y1);
        tri[4+6*(*ntri)]=.5*(x2+x0);
        tri[5+6*(*ntri)]=.5*(y2+y0);
        tri[0+6*(*ntri+1)]=x1;
        tri[1+6*(*ntri+1)]=y1;
        tri[2+6*(*ntri+1)]=.5*(x1+x2);
        tri[3+6*(*ntri+1)]=.5*(y1+y2);
        tri[4+6*(*ntri+1)]=.5*(x0+x1);
        tri[5+6*(*ntri+1)]=.5*(y0+y1);
        tri[0+6*(*ntri+2)]=x2;
        tri[1+6*(*ntri+2)]=y2;
        tri[2+6*(*ntri+2)]=.5*(x2+x0);
        tri[3+6*(*ntri+2)]=.5*(y2+y0);
        tri[4+6*(*ntri+2)]=.5*(x1+x2);
        tri[5+6*(*ntri+2)]=.5*(y1+y2);
        *ntri+=3;
       }
     }
   }

  MFFreeKVector(pt0);
  MFFreeKVector(pt1);
  MFFreeKVector(pt2);
  return;
 }

/* Use a project for each NSpace? */

int MFTPBVPGetNX(MFImplicitMF);
int MFTPBVPGetNU(MFImplicitMF);
int MFTPBVPGetNP(MFImplicitMF);

void MFDrawProjectTo3d(MFChart chart,MFNVector u, float *x,float *y,float *z)
 {
  static char RoutineName[]={"MFDrawProjectTo3d"};
  int i,j,n;
  int nx,nu,np;
  int verbose=0;

  n=MFNV_NC(u);

  if(verbose)printf("MFDrawProjectTo3d, n=%d\n",n);

  if(n<4)
   {
    if(verbose)printf("  n<4\n");
    *x=MFNV_C(u,0);
    *y=MFNV_C(u,1);
    if(n>2)
      *z=MFNV_C(u,2);
     else
      *z=0.;
   }else if(n==4){
    if(verbose)printf("  n==4\n");
    *x=.5*(MFNV_C(u,0)+MFNV_C(u,1));
    *y=MFNV_C(u,2);
    *z=MFNV_C(u,3);
   }else{
    if(verbose)printf("n>4, looking for a way %s\n",MFImplicitMFId(MFChartGetManifold(chart)));
    if(chart!=(MFChart)NULL && !strcmp(MFImplicitMFId(MFChartGetManifold(chart)),"Pendula"))
     {
      if(verbose)printf("  Pendula\n");
      *x=MFNV_C(u,n-1); /* Ic/(k pi) */
#ifdef PERIODIC
      if(MFPendulaPeriodicFlag)
       {
        while(*x<-1.e-7)*x+=1.;
        while(*x>.1*MFPI+1.e-7)*x-=1.;
       }
#endif
      *y=.2*MFNV_C(u,n-2); /* In */
      *z=.1*2.*MFNV_C(u,n-3); /* T  */

      *x=2*MFNV_C(u,n-1); /* Ic/(k pi) */
      *y=MFNV_C(u,n-2); /* In */
      *z=MFNV_C(u,n-3); /* T  */
     }else if(chart!=(MFChart)NULL && !strcmp(MFImplicitMFId(MFChartGetManifold(chart)),"TPBVP"))
     {
      if(verbose)printf("  TPBVP\n");
      nu=MFTPBVPGetNU(MFChartGetManifold(chart));
      nx=MFTPBVPGetNX(MFChartGetManifold(chart));
      np=MFTPBVPGetNP(MFChartGetManifold(chart));
      if(np==2)
       {
        *x=0.;
        for(i=0;i<nx;i++)
         {
          for(j=0;j<nu;j++)
           {
            *x+=(MFNV_C(u,i*nu+j)+MFNV_C(u,(i+1)*nu+j))*(MFNV_C(u,nx*nu+np+i+1)-MFNV_C(u,nx*nu+np+i))/2.;
           }
         }
        *x=*x/nu;
        *y=MFNV_C(u,nx*nu);
        *z=MFNV_C(u,nx*nu+1);
       }else{
        *x=MFNV_C(u,nx*nu);
        *y=MFNV_C(u,nx*nu+1);
        *z=MFNV_C(u,nx*nu+2);
       }
     }else{
      if(verbose)printf("MF is not TPBVP or others\n");
      *x=MFNV_C(u,n-3);
      *y=MFNV_C(u,n-2);
      *z=MFNV_C(u,n-1);
     }
   }

  if(MFDrawSetExt||*x<MFDrawXMin)MFDrawXMin=*x;
  if(MFDrawSetExt||*x>MFDrawXMax)MFDrawXMax=*x;
  if(MFDrawSetExt||*y<MFDrawYMin)MFDrawYMin=*y;
  if(MFDrawSetExt||*y>MFDrawYMax)MFDrawYMax=*y;
  if(MFDrawSetExt||*z<MFDrawZMin)MFDrawZMin=*z;
  if(MFDrawSetExt||*z>MFDrawZMax)MFDrawZMax=*z;
  MFDrawSetExt=0;
/*printf("%s, ",RoutineName);MFPrintNVector(stdout,u);printf(" ->(%f,%f,%f)\n",x,y,z);fflush(stdout);*/

  return;
 }

void MFDrawEnumPolytope(MFEnumPolytope EP,MFChart chart,MFChartState state)
 {
  static char RoutineName[]={"MFDrawEnumPolytope"};
  float x0,y0,z0;
  float x1,y1,z1;
  int v[100]={0.};
  int v0,v1,v2;
  int e0,e1,e2;
  int full=255;
  int zero=0;
  int i,j,l,n,k;
  int verbose;
  float *x,*y,*z;
  MFNVector u;
  int rc;

  int r=180;
  int g=180;
  int b=205;
  shpgc(&r,&g,&b);
  shbgc(&r,&g,&b);
  shpec(&zero,&zero,&zero);
  shbec(&zero,&zero,&zero);

  k=MFChartK(chart);
  n=MFChartN(chart);
  verbose=0;
  if(verbose){printf("MFDrawEnumPolytope base dim =%d, target dim=%d\n",k,n);fflush(stdout);}

  if(k<1||k>3)return;

  u=MFCreateNVector(n);

/* Vertices */

  shlinc(&zero,&zero,&zero);
  if(0)for(i=0;i<MFEnumPolytopeNumberOfCells(EP,0);i++)
   {
    rc=MFChartEvaluate(chart,MFEnumPolytopeVertex(EP,i),u);
    if(rc)MFMarkPoint(chart,u,3,state);
   }
  
/* Edges */

  for(i=0;i<MFEnumPolytopeNumberOfCells(EP,1);i++)
   {
    if(MFEnumPolytopeNumberOfCellFaces(EP,1,i)>1)
     {
      v0=MFEnumPolytopeCellFace(EP,1,i,0);
      v1=MFEnumPolytopeCellFace(EP,1,i,1);
      if(verbose){printf("edge %d joins vertices %d and %d\n",i,v0,v1);fflush(stdout);}
      MFDrawLineOnChart(chart,MFEnumPolytopeVertex(EP,v0),MFEnumPolytopeVertex(EP,v1),state);
     }
   }

/* faces */

  shtric(&full,&full,&full);
  if(k==3 && n==3)
   {
    x=(float*)NULL;
    y=(float*)NULL;
    z=(float*)NULL;
    for(i=0;i<MFEnumPolytopeNumberOfCells(EP,2);i++)
     {
      n=MFEnumPolytopeNumberOfCellFaces(EP,2,i);
  
      if(n>2)
       {
        e0=MFEnumPolytopeCellFace(EP,2,i,0);
        v[0]=MFEnumPolytopeCellFace(EP,1,e0,0);
        v[1]=MFEnumPolytopeCellFace(EP,1,e0,1);
        for(j=1;j<n;j++)
         {
          for(l=0;l<n;l++)
           {
            e1=MFEnumPolytopeCellFace(EP,2,i,l);
            if(e1!=e0&&MFEnumPolytopeCellFace(EP,1,e1,0)==v[j])
             {
              v[j+1]=MFEnumPolytopeCellFace(EP,1,e1,1);
              l=n;
             }else if(e1!=e0&&MFEnumPolytopeCellFace(EP,1,e1,1)==v[j])
             {
              v[j+1]=MFEnumPolytopeCellFace(EP,1,e1,0);
              l=n;
             }
           }
          e0=e1;
         }

        x=(float*)realloc((void*)x,n*sizeof(float)); /*done*/
        if(x==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        y=(float*)realloc((void*)y,n*sizeof(float)); /*done*/
        if(y==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        z=(float*)realloc((void*)z,n*sizeof(float)); /*done*/
        if(z==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        rc=1;
        for(j=0;j<n;j++)
         {
          rc=rc&&MFChartEvaluate(chart,MFEnumPolytopeVertex(EP,v[j]),u);
          MFDrawProjectTo3d(chart,u,x+j,y+j,z+j);
#ifdef PERIODIC
          if(MFPendulaPeriodicFlag&&j>0)
           {
            rc=rc&&fabs(x[j-1]-x[j])<.1;
            rc=rc&&fabs(y[j-1]-y[j])<.1;
           }
#endif
         }
#ifdef PERIODIC
        if(MFPendulaPeriodicFlag)
         {
          rc=rc&&fabs(x[n-1]-x[0])<.1;
          rc=rc&&fabs(y[n-1]-y[0])<.1;
         }
#endif
        if(rc&&MFEnumPolytopeCellIndex(EP,2,i,0)>7)
         {
          if(state!=(MFChartState)NULL)MFChartStateAddPolygon(state,n,x,y,z,0);
            else shpg(&n,x,y,z);
         }
       }
     }
    free(x);
    free(y);
    free(z);
   }

  MFFreeNVector(u);
  if(verbose){printf("done MFDrawEnumPolytope\n");fflush(stdout);}
  return;
 }

void MFMarkPoint(MFChart chart, MFNVector u, int i,MFChartState state)
 {
  static char RoutineName[]={"MFDrawMarkPoint"};
  float x=0.;
  float y=0.;
  float z=0.;
  double xt,yt,zt;
  double xn,yn,zn;

  float x0,y0,z0;
  float x1,y1,z1;

  MFDrawProjectTo3d(chart,u,&x,&y,&z);

  xt=1.;
  yt=0.;
  zt=0.;

  xn=0.;
  yn=1.;
  zn=0.;

  switch(i)
   {
    case 0:  /* [ */
     x0=x-.01*xt-.02*xn;
     y0=y-.01*yt-.02*yn;
     z0=z-.01*zt-.02*zn;
     x1=x       -.02*xn;
     y1=y       -.02*yn;
     z1=z       -.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.01*xt+.02*xn;
     y1=y-.01*yt+.02*yn;
     z1=z-.01*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    case 1: /* ]  */
     x0=x+.01*xt-.02*xn;
     y0=y+.01*yt-.02*yn;
     z0=z+.01*zt-.02*zn;
     x1=x       -.02*xn;
     y1=y       -.02*yn;
     z1=z       -.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.01*xt+.02*xn;
     y1=y+.01*yt+.02*yn;
     z1=z+.01*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    case 2: /*  O */
     x0=x;
     y0=y;
     z0=z;
     x1=x+.02*xt-.02*xn;
     y1=y+.02*yt-.02*yn;
     z1=z+.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.02*xt+.02*xn;
     y1=y+.02*yt+.02*yn;
     z1=z+.02*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.02*xt+.02*xn;
     y1=y-.02*yt+.02*yn;
     z1=z-.02*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.02*xt-.02*xn;
     y1=y-.02*yt-.02*yn;
     z1=z-.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.02*xt-.02*xn;
     y1=y+.02*yt-.02*yn;
     z1=z+.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    default: /*  +  */
     x0=x       -.02*xn;
     y0=y       -.02*yn;
     z0=z       -.02*zn;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x       -.02*xt;
     y0=y       -.02*yt;
     z0=z       -.02*zt;
     x1=x       +.02*xt;
     y1=y       +.02*yt;
     z1=z       +.02*zt;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
   }
  return;
 }

void MFMarkPointOnChart(MFChart chart,MFKVector s, int ms,MFChartState state)
 {
  static char RoutineName[]={"MFDrawMarkPointOnChart"};
  MFNVector u;
  float x=0.;
  float y=0.;
  float z=0.;
  float xt=0.;
  float yt=0.;
  float zt=0.;
  float xn,yn,zn;

  float x0,y0,z0;
  float x1,y1,z1;

  u=MFCreateNVector(MFChartN(chart));
  if(!MFChartEvaluate(chart,s,u))return;
  MFDrawProjectTo3d(chart,u,&x,&y,&z);
  MFFreeNVector(u);

  u=MFMColumn(MFChartTangentSpace(chart),0);
  MFDrawProjectTo3d(chart,u,&xt,&yt,&zt);
  MFFreeNVector(u);

  xn=-yt;
  yn= xt;
  zn= 0.;

  switch(ms)
   {
    case 0:  /* [ */
     x0=x-.01*xt-.02*xn;
     y0=y-.01*yt-.02*yn;
     z0=z-.01*zt-.02*zn;
     x1=x       -.02*xn;
     y1=y       -.02*yn;
     z1=z       -.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.01*xt+.02*xn;
     y1=y-.01*yt+.02*yn;
     z1=z-.01*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    case 1: /* ]  */
     x0=x+.01*xt-.02*xn;
     y0=y+.01*yt-.02*yn;
     z0=z+.01*zt-.02*zn;
     x1=x       -.02*xn;
     y1=y       -.02*yn;
     z1=z       -.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.01*xt+.02*xn;
     y1=y+.01*yt+.02*yn;
     z1=z+.01*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    case 2: /*  O */
     x0=x;
     y0=y;
     z0=z;
     x1=x+.02*xt-.02*xn;
     y1=y+.02*yt-.02*yn;
     z1=z+.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.02*xt+.02*xn;
     y1=y+.02*yt+.02*yn;
     z1=z+.02*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.02*xt+.02*xn;
     y1=y-.02*yt+.02*yn;
     z1=z-.02*zt+.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x-.02*xt-.02*xn;
     y1=y-.02*yt-.02*yn;
     z1=z-.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x1; y0=y1; z0=z1;
     x1=x+.02*xt-.02*xn;
     y1=y+.02*yt-.02*yn;
     z1=z+.02*zt-.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
    default: /*  +  */
     x0=x       -.02*xn;
     y0=y       -.02*yn;
     z0=z       -.02*zn;
     x1=x       +.02*xn;
     y1=y       +.02*yn;
     z1=z       +.02*zn;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     x0=x       -.02*xt;
     y0=y       -.02*yt;
     z0=z       -.02*zt;
     x1=x       +.02*xt;
     y1=y       +.02*yt;
     z1=z       +.02*zt;
     if(state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
     if(state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);
     break;
   }
  return;
 }

void MFDrawLineOnChart(MFChart chart,MFKVector s0,MFKVector s1,MFChartState state)
 {
  static char RoutineName[]={"MFDrawLineOnChart"};
  float x0=0.;
  float y0=0.;
  float z0=0.;
  float x1=0.;
  float y1=0.;
  float z1=0.;
  double dx,dy,dz;
  double ox,oy,oz;
  int k;
  double t,dt;
  int rc,rc0;

  if(MFTmpK==(MFKVector)NULL)MFTmpK=MFCreateKVector(MFChartK(chart));
  if(MFTmpN==(MFNVector)NULL)MFTmpN=MFCreateNVector(MFChartN(chart));

  k=MFChartK(chart);
  ox=MFKV_C(s0,0);dx=MFKV_C(s1,0)-ox;
  dy=0.;
  oy=0.;
  if(1<k){oy=MFKV_C(s0,1);dy=MFKV_C(s1,1)-oy;}
  dz=0.;
  oz=0.;
  if(2<k){oz=MFKV_C(s0,2);dz=MFKV_C(s1,2)-oz;}

  rc0=MFChartEvaluate(chart,s0,MFTmpN);
  MFDrawProjectTo3d(chart,MFTmpN,&x0,&y0,&z0);
  rc=MFChartEvaluate(chart,s1,MFTmpN);
  MFDrawProjectTo3d(chart,MFTmpN,&x1,&y1,&z1);

  dt=.05;
  if(fabs(x0-x1)+fabs(y0-y1)+fabs(z0-z1)<dt)dt=1.;
  for(t=dt;t<1.+.5*dt;t+=dt)
   {
    MFKVSetC(MFTmpK,0,ox+t*dx);
    if(1<k)MFKVSetC(MFTmpK,1,oy+t*dy);
    if(2<k)MFKVSetC(MFTmpK,2,oz+t*dz);
    rc=MFChartEvaluate(chart,MFTmpK,MFTmpN);
    MFDrawProjectTo3d(chart,MFTmpN,&x1,&y1,&z1);
#ifdef PERIODIC
    if(MFPendulaPeriodicFlag)
     {
      rc=rc&&fabs(x1-x0)<.1;
      rc=rc&&fabs(y1-y0)<.1;
     }
#endif

    if(rc&&rc0&&state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
    if(rc&&rc0&&state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);

    x0=x1;
    y0=y1;
    z0=z1;
    rc0=rc;
   }

  return;
 }

void MFDrawEnumDualPolytope(MFEnumDualPolytope EP)
 {
  static char RoutineName[]={"MFDrawEnumDualPolytope"};
  float x0=0.;
  float y0=0.;
  float z0=0.;
  float x1=0.;
  float y1=0.;
  float z1=0.;
  int v[100]={0.};
  int v0,v1;
  int e0,e1;
  int full=255;
  int zero=0;
  int i,n,j,l,k;
  int verbose;
  float x[100]={0.};
  float y[100]={0.};
  float z[100]={0.};
  float xm[100]={0.};
  float ym[100]={0.};
  float zm[100]={0.};
  MFNVector vert;

  int r=180;
  int g=180;
  int b=205;
  int rc;

  shpgc(&r,&g,&b);
  shbgc(&r,&g,&b);
  shpec(&zero,&zero,&zero);
  shbec(&zero,&zero,&zero);

  n=MFEnumDualPolytopeDimension(EP);
  verbose=0;
  if(verbose){printf("MFDrawEnumDualPolytope dim =%d\n",n);fflush(stdout);}

  if(n<1||n>3)return;

/* Vertices */

  shlinc(&zero,&zero,&zero);
  if(0)
   {
    for(i=0;i<MFEnumDualPolytopeNumberOfCells(EP,0);i++)
     {
      MFMarkPoint((MFChart)NULL,MFEnumDualPolytopeVertex(EP,i),3,(MFChartState)NULL);
     }
   }

/* Edges */

  k=MFEnumDualPolytopeDimension(EP);
  if(1)for(i=0;i<MFEnumDualPolytopeNumberOfCells(EP,1);i++)
   {
    if(k==2 && MFEnumDualPolytopeNumberOfFaceCells(EP,1,i)<2)
      shlinc(&full,&zero,&zero);
     else
      shlinc(&zero,&zero,&zero);

    if(MFEnumDualPolytopeNumberOfCellFaces(EP,1,i)>1)
     {
      v0=MFEnumDualPolytopeCellFace(EP,1,i,0);
      v1=MFEnumDualPolytopeCellFace(EP,1,i,1);
      if(verbose){printf("edge %d joins vertices %d and %d\n",i,v0,v1);fflush(stdout);}
      if(MFEnumDualPolytopeVertex(EP,v0)!=(MFNVector)NULL
       &&MFEnumDualPolytopeVertex(EP,v1)!=(MFNVector)NULL)
       {
        MFDrawProjectTo3d((MFChart)NULL,MFEnumDualPolytopeVertex(EP,v0),&x0,&y0,&z0);
        MFDrawProjectTo3d((MFChart)NULL,MFEnumDualPolytopeVertex(EP,v1),&x1,&y1,&z1);
        shline(&x0,&y0,&z0,&x1,&y1,&z1);
       }
     }
   }

/* faces */
  if(1)for(i=0;i<MFEnumDualPolytopeNumberOfCells(EP,2);i++)
   {
    n=MFEnumDualPolytopeNumberOfCellFaces(EP,2,i);
    if(1){

    if(n>3){printf("MFDrawEnumDualPolytope: Here's a 2-cell (%d) which has more sides than a triangle (%d)\n",i,n);fflush(stdout);}

/*  if(n>2 || ( k==3&&MFEnumDualPolytopeNumberOfFaceCells(EP,2,i)<2))*/
    if(n>2)
     {
      e0=MFEnumDualPolytopeCellFace(EP,2,i,0);
      v[0]=MFEnumDualPolytopeCellFace(EP,1,e0,0);
      v[1]=MFEnumDualPolytopeCellFace(EP,1,e0,1);
      if(n>4)printf("Edge %d is %d [%d,%d]\n",i,e0,v[0],v[1]);
      for(j=1;j<n;j++)
       {
        if(n>4)printf("Edge %d is %d [%d,%d]\n",i,MFEnumDualPolytopeCellFace(EP,2,i,j),MFEnumDualPolytopeCellFace(EP,1,MFEnumDualPolytopeCellFace(EP,2,i,j),0),MFEnumDualPolytopeCellFace(EP,1,MFEnumDualPolytopeCellFace(EP,2,i,j),1));
        for(l=0;l<n;l++)
         {
          e1=MFEnumDualPolytopeCellFace(EP,2,i,l);
/*        if(MFEnumDualPolytopeNumberOfFaceCells(EP,1,e1)>1)*/
           {
            if(e1!=e0&&MFEnumDualPolytopeCellFace(EP,1,e1,0)==v[j])
             {
              v[j+1]=MFEnumDualPolytopeCellFace(EP,1,e1,1);
              l=n;
             }else if(e1!=e0&&MFEnumDualPolytopeCellFace(EP,1,e1,1)==v[j])
             {
              v[j+1]=MFEnumDualPolytopeCellFace(EP,1,e1,0);
              l=n;
             }
           }
         }
        e0=e1;
       }

      for(j=0;j<n;j++)
       {
        vert=MFEnumDualPolytopeVertex(EP,v[j]);
        if(vert!=(MFNVector)NULL)
         {
          x[j]=0;
          y[j]=0;
          z[j]=0;
          if(MFNV_NC(vert)>0)
            MFDrawProjectTo3d((MFChart)NULL,MFEnumDualPolytopeVertex(EP,v[j]),x+j,y+j,z+j);
         }
       }
      x[n]=x[0];
      y[n]=y[0];
      z[n]=z[0];

      rc=1;
      for(j=0;j<n;j++)
       {
        xm[j]=.5*(x[j]+x[j+1]);
        ym[j]=.5*(y[j]+y[j+1]);
        zm[j]=.5*(z[j]+z[j+1]);
#ifdef PERIODIC
        if(MFPendulaPeriodicFlag&&j>0)
         {
          rc=rc&&fabs(x[j-1]-x[j])<.1;
          rc=rc&&fabs(y[j-1]-y[j])<.1;
         }
#endif
       }
#ifdef PERIODIC
      if(MFPendulaPeriodicFlag)
       {
        rc=rc&&fabs(x[n-1]-x[0])<.1;
        rc=rc&&fabs(y[n-1]-y[0])<.1;
       }
#endif

      if(rc&&n<5)
       {
        if(n<4)shtric(&r,&g,&b);
         else  shtric(&full,&full,&zero);
        if(k==3 && MFEnumDualPolytopeNumberOfFaceCells(EP,2,i)<2)
         {
          shtric(&full,&zero,&zero);
          shpec(&zero,&zero,&zero);
          shbec(&zero,&zero,&zero);
         }else{
          shtric(&r,&g,&b);
          shpec(&zero,&zero,&zero);
          shbec(&zero,&zero,&zero);
         }
        if(1)shpg(&n,x,y,z);
         else shpg(&n,xm,ym,zm);
       }else if(rc)
       {
        shlinc(&zero,&full,&zero);
        shpl(&n,x,y,z);
       }
     }    /* if(n>2 && MFEnumDualPolytopeNumberOfFaceCells(EP,2,i)<2) */
   }      /* if(FALSE) */
   }      /* for(i=0,..,MFEnumDualPolytopeNumberOfCells(EP,2)) */

  if(verbose){printf("done MFDrawEnumDualPolytope\n");fflush(stdout);}
  return;
 }

void MFDrawLine(MFNVector u0,MFNVector u1,MFChartState state)
 {
  static char RoutineName[]={"MFDrawLine"};
  float x0=0.;
  float y0=0.;
  float z0=0.;
  float x1=0.;
  float y1=0.;
  float z1=0.;
  int k;
  int rc;

  MFDrawProjectTo3d((MFChart)NULL,u0,&x0,&y0,&z0);
  MFDrawProjectTo3d((MFChart)NULL,u1,&x1,&y1,&z1);

  rc=1;
#ifdef PERIODIC
  if(MFPendulaPeriodicFlag)
   {
    rc=rc&&fabs(x0-x1)<.1;
    rc=rc&&fabs(y0-y1)<.1;
   }
#endif

  if(rc&&state!=(MFChartState)NULL)MFChartStateAddLine(state,x0,y0,z0,x1,y1,z1);
  if(rc&&state==(MFChartState)NULL)shline(&x0,&y0,&z0,&x1,&y1,&z1);

  return;
 }

void MFDrawChartTS(MFChart chart)
 {
  static char RoutineName[]={"MFDrawChartTS"};

  if(chart==(MFChart)NULL)return;
  if(MFChartK(chart)==1)MFDraw1dChartTS(chart);
   else if(MFChartK(chart)==2)MFDraw2dChartTS(chart);
/* else if(MFChartK(chart)==3)MFDraw3dChartTS(chart);*/

  return;
 }

void MFDrawAtlasTS(MFAtlas A)
 {
  static char RoutineName[]={"MFDrawAtlasTS"};
  int chart,bchart;
/*int ri=180;
  int gi=180;
  int bi=205;*/
  int ri=62;
  int gi=158;
  int bi=255;
  int zero=0;
  int full=255;
/*int rb=243;
  int gb=223;
  int bb=192;*/
  int rb=255;
  int gb=211;
  int bb=62;
  int N;
  int ibnd;
  int verbose;

  MFDrawSetExt=1;
  MFDrawXMin=0.;
  MFDrawXMax=0.;
  MFDrawYMin=0.;
  MFDrawYMax=0.;
  MFDrawZMin=0.;
  MFDrawZMax=0.;

  verbose=0;
  if(verbose){printf("In %s\n",RoutineName);fflush(stdout);}

  shlinc(&zero,&zero,&zero);
  N=MFAtlasNumberOfCharts(A);
  for(chart=0;chart<N;chart++)
   {
    ibnd=0;
    shtric(&ri,&gi,&bi);
    for(bchart=0;bchart<MFAtlasNumberOfChartsWithBoundary(A);bchart++)
     {
      if(chart==MFAtlasChartWithBoundary(A,bchart))
       {
        shtric(&rb,&gb,&bb);
        ibnd=1;
       }
     }
    MFDrawChartTS(MFAtlasChart(A,chart));
    MFDrawPolytopeTS(MFChartPolytope(MFAtlasChart(A,chart)),MFAtlasChart(A,chart));
    if(ibnd)
      MFDrawChartBoundaryTS(MFAtlasChart(A,chart));
    if(verbose&&chart%1000==0&&chart!=0){printf("     %d/%d\n",chart,N);fflush(stdout);}
   }

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  printf("at end of MFDrawAtlas, x in ([%lf,%lf],[%lf,%lf],[%lf,%lf])\n",MFDrawXMin,MFDrawXMax,MFDrawYMin,MFDrawYMax,MFDrawZMin,MFDrawZMax);fflush(stdout);
  return;
 }

void MFDraw1dChartTS(MFChart chart)
 {
  static char RoutineName[]={"MFDraw1dChartTS"};
  int color;
  int zero=0;
  int half=128;
  int full=255;
  MFPolytope P;
  MFKVector s0,s1;
  MFNVector u0,u1;
  int n,k;

  P=MFChartPolytope(chart);
  k=MFChartK(chart);
  n=MFChartN(chart);
  if(MFPolytopeNumberOfVertices(P)<2)return;
  s0=MFCreateKVector(k);
  s1=MFCreateKVector(k);
  u0=MFCreateNVector(n);
  u1=MFCreateNVector(n);

  MFPolytopeVertex(P,0,s0);
  MFChartPointInTangentSpace(chart,s0,u0);
  MFPolytopeVertex(P,1,s1);
  MFChartPointInTangentSpace(chart,s1,u1);

  MFMarkPoint((MFChart)NULL,u0,2,(MFChartState)NULL);
  MFMarkPoint((MFChart)NULL,u1,2,(MFChartState)NULL);
  MFMarkPoint((MFChart)NULL,MFChartCenter(chart),3,(MFChartState)NULL);

  MFDrawLine(u0,u1,(MFChartState)NULL);

  MFFreeKVector(s0);
  MFFreeKVector(s1);
  MFFreeNVector(u0);
  MFFreeNVector(u1);

  return;
 }

void MFDraw2dChartTS(MFChart chart)
 {
  static char RoutineName[]={"MFDraw2dChartTS"};
  int color;
  int zero=0;
  int half=128;
  int full=255;
  double *tri;
  int ntri,mtri;
  static float x[3],y[3],z[3];
  double t;
  MFKVector s;
  MFNVector u;
  int j,n;
  int rc;

/* The interior of the circle */

  tri=(double*)malloc(6*6*sizeof(double)); /*done*/
  if(tri==(double*)NULL)
   {
    sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",6*6*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
    return;
   }

  t=2./sqrt3*MFChartRadius(chart);

  tri[0+6*0]=0.;
  tri[1+6*0]=0.;
  tri[2+6*0]=t;
  tri[3+6*0]=0.;
  tri[4+6*0]=.5*t;
  tri[5+6*0]=.5*sqrt3*t;

  tri[0+6*1]=0.;
  tri[1+6*1]=0.;
  tri[2+6*1]=.5*t;
  tri[3+6*1]=.5*sqrt3*t;
  tri[4+6*1]=-.5*t;
  tri[5+6*1]=.5*sqrt3*t;

  tri[0+6*2]=0.;
  tri[1+6*2]=0.;
  tri[2+6*2]=-.5*t;
  tri[3+6*2]=.5*sqrt3*t;
  tri[4+6*2]=-t;
  tri[5+6*2]=0.;

  tri[0+6*3]=0.;
  tri[1+6*3]=0.;
  tri[2+6*3]=-t;
  tri[3+6*3]=0.;
  tri[4+6*3]=-.5*t;
  tri[5+6*3]=-.5*sqrt3*t;

  tri[0+6*4]=0.;
  tri[1+6*4]=0.;
  tri[2+6*4]=-.5*t;
  tri[3+6*4]=-.5*sqrt3*t;
  tri[4+6*4]= .5*t;
  tri[5+6*4]=-.5*sqrt3*t;

  tri[0+6*5]=0.;
  tri[1+6*5]=0.;
  tri[2+6*5]= .5*t;
  tri[3+6*5]=-.5*sqrt3*t;
  tri[4+6*5]=t;
  tri[5+6*5]=0.;

  ntri=6;
  mtri=6;
  t=.2*MFChartRadius(chart);
/*if(t<.01)t=.01;*/
  MFDrawMakeListOfTriangles(chart,&ntri,&mtri,&tri,t);

  s=MFCreateKVector(MFChartK(chart));
  u=MFCreateNVector(MFChartN(chart));
  for(j=0;j<ntri;j++)
   {
    MFKVSetC(s,0,tri[0+6*j]);
    MFKVSetC(s,1,tri[1+6*j]);
    MFChartPointInTangentSpace(chart,s,u);
    MFDrawProjectTo3d(chart,u,x,y,z);
    MFKVSetC(s,0,tri[2+6*j]);
    MFKVSetC(s,1,tri[3+6*j]);
    MFChartPointInTangentSpace(chart,s,u);
    MFDrawProjectTo3d(chart,u,x+1,y+1,z+1);
    MFKVSetC(s,0,tri[4+6*j]);
    MFKVSetC(s,1,tri[5+6*j]);
    MFChartPointInTangentSpace(chart,s,u);
    MFDrawProjectTo3d(chart,u,x+2,y+2,z+2);
    rc=1;
#ifdef PERIODIC
    if(MFPendulaPeriodicFlag)
     {
      rc=rc&&fabs(x[0]-x[1])<.1;
      rc=rc&&fabs(x[1]-x[2])<.1;
      rc=rc&&fabs(x[2]-x[0])<.1;
      rc=rc&&fabs(y[0]-y[1])<.1;
      rc=rc&&fabs(y[1]-y[2])<.1;
      rc=rc&&fabs(y[2]-y[0])<.1;
     }
#endif
    n=3;
    if(rc)shpg(&n,x,y,z);
   }
  free(tri);
  MFFreeKVector(s);
  MFFreeNVector(u);

  return;
 }

void MFDrawEnumPolytopeTS(MFEnumPolytope EP,MFChart chart)
 {
  static char RoutineName[]={"MFDrawEnumPolytopeTS"};
  float x0,y0,z0;
  float x1,y1,z1;
  int v[100]={0.};
  int v0,v1,v2;
  int e0,e1,e2;
  int full=255;
  int zero=0;
  int i,j,l,n,k;
  int verbose;
  float *x,*y,*z;
  MFNVector u;
  int rc;

  int r=180;
  int g=180;
  int b=205;
  shpgc(&r,&g,&b);
  shbgc(&r,&g,&b);
  shpec(&zero,&zero,&zero);
  shbec(&zero,&zero,&zero);

  k=MFChartK(chart);
  n=MFChartN(chart);
  verbose=0;
  if(verbose){printf("MFDrawEnumPolytope base dim =%d, target dim=%d\n",k,n);fflush(stdout);}

  if(k<1||k>3)return;

  u=MFCreateNVector(n);

/* Vertices */

  shlinc(&zero,&zero,&zero);
  if(0)
   {
    for(i=0;i<MFEnumPolytopeNumberOfCells(EP,0);i++)
     {
      MFChartPointInTangentSpace(chart,MFEnumPolytopeVertex(EP,i),u);
      MFMarkPoint(chart,u,3,(MFChartState)NULL);
     }
   }
  
/* Edges */

  for(i=0;i<MFEnumPolytopeNumberOfCells(EP,1);i++)
   {
    if(MFEnumPolytopeNumberOfCellFaces(EP,1,i)>1)
     {
      v0=MFEnumPolytopeCellFace(EP,1,i,0);
      v1=MFEnumPolytopeCellFace(EP,1,i,1);
      if(verbose){printf("edge %d joins vertices %d and %d\n",i,v0,v1);fflush(stdout);}
      MFDrawLineOnChartTS(chart,MFEnumPolytopeVertex(EP,v0),MFEnumPolytopeVertex(EP,v1));
     }
   }

/* faces */

  if(k==3 && n==3)
   {
    x=(float*)NULL;
    y=(float*)NULL;
    z=(float*)NULL;
    for(i=0;i<MFEnumPolytopeNumberOfCells(EP,2);i++)
     {
      n=MFEnumPolytopeNumberOfCellFaces(EP,2,i);
  
      if(n>2)
       {
        e0=MFEnumPolytopeCellFace(EP,2,i,0);
        v[0]=MFEnumPolytopeCellFace(EP,1,e0,0);
        v[1]=MFEnumPolytopeCellFace(EP,1,e0,1);
        for(j=1;j<n;j++)
         {
          for(l=0;l<n;l++)
           {
            e1=MFEnumPolytopeCellFace(EP,2,i,l);
            if(e1!=e0&&MFEnumPolytopeCellFace(EP,1,e1,0)==v[j])
             {
              v[j+1]=MFEnumPolytopeCellFace(EP,1,e1,1);
              l=n;
             }else if(e1!=e0&&MFEnumPolytopeCellFace(EP,1,e1,1)==v[j])
             {
              v[j+1]=MFEnumPolytopeCellFace(EP,1,e1,0);
              l=n;
             }
           }
          e0=e1;
         }

        x=(float*)realloc((void*)x,n*sizeof(float)); /*done*/
        if(x==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        y=(float*)realloc((void*)y,n*sizeof(float)); /*done*/
        if(y==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        z=(float*)realloc((void*)z,n*sizeof(float)); /*done*/
        if(z==(float*)NULL)
         {
          sprintf(MFDrawErrorMsg,"Out of memory trying to allocate %d bytes.",n*sizeof(float));
          printf("%s -- %s\n",RoutineName,MFDrawErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFDrawErrorMsg,__LINE__,__FILE__);
          return;
         }
        rc=1;
        for(j=0;j<n;j++)
         {
          MFChartPointInTangentSpace(chart,MFEnumPolytopeVertex(EP,v[j]),u);
          MFDrawProjectTo3d(chart,u,x+j,y+j,z+j);
#ifdef PERIODIC
          if(MFPendulaPeriodicFlag&&j>0)
           {
            rc=rc&&fabs(x[j-1]-x[j])<.1;
            rc=rc&&fabs(y[j-1]-y[j])<.1;
           }
#endif
         }
#ifdef PERIODIC
        if(MFPendulaPeriodicFlag)
         {
          rc=rc&&fabs(x[n-1]-x[0])<.1;
          rc=rc&&fabs(y[n-1]-y[0])<.1;
         }
#endif
        if(rc&&MFEnumPolytopeCellIndex(EP,2,i,0)>7)shpg(&n,x,y,z);
       }
     }
    free(x);
    free(y);
    free(z);
   }

  MFFreeNVector(u);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return;
 }

void MFDrawLineOnChartTS(MFChart chart,MFKVector s0,MFKVector s1)
 {
  static char RoutineName[]={"MFDrawLineOnChartTS"};
  float x0=0.;
  float y0=0.;
  float z0=0.;
  float x1=0.;
  float y1=0.;
  float z1=0.;
  int rc;

  if(MFTmpK==(MFKVector)NULL)MFTmpK=MFCreateKVector(MFChartK(chart));
  if(MFTmpN==(MFNVector)NULL)MFTmpN=MFCreateNVector(MFChartN(chart));

  MFChartPointInTangentSpace(chart,s0,MFTmpN);
  MFDrawProjectTo3d(chart,MFTmpN,&x0,&y0,&z0);
  MFChartPointInTangentSpace(chart,s1,MFTmpN);
  MFDrawProjectTo3d(chart,MFTmpN,&x1,&y1,&z1);

  rc=1;
#ifdef PERIODIC
  if(MFPendulaPeriodicFlag)
   {
    rc=rc&&fabs(x0-x1)<.1;
    rc=rc&&fabs(y0-y1)<.1;
   }
#endif

  if(rc)shline(&x0,&y0,&z0,&x1,&y1,&z1);

  return;
 }

void MFPendulaPeriodic(int flag)
 {
  MFPendulaPeriodicFlag=flag;
 }

#ifdef NEW
void MFDrawSlicedChartTS(MFChart,int,double*,double);
void MFDrawSliced1dChartTS(MFChart,int,double*,double);
void MFDrawSliced2dChartTS(MFChart,int,double*,double);
void MFCreateSectionOfChart(MFChart chart,int m,double *n,double *on,int *knew, double **s0, souble **Psi);
#endif

void MFDrawEnumDualPolytopeEdges(MFEnumDualPolytope EP)
 {
  static char RoutineName[]={"MFDrawEnumDualPolytopeEdges"};
  float x0=0.;
  float y0=0.;
  float z0=0.;
  float x1=0.;
  float y1=0.;
  float z1=0.;
  int v[100]={0.};
  int v0,v1;
  int e0,e1;
  int full=255;
  int zero=0;
  int i,n,j,l,k;
  int verbose;
  float x[100]={0.};
  float y[100]={0.};
  float z[100]={0.};
  float xm[100]={0.};
  float ym[100]={0.};
  float zm[100]={0.};
  MFNVector vert;

  int r=180;
  int g=180;
  int b=205;
  int rc;

  shpgc(&r,&g,&b);
  shbgc(&r,&g,&b);
  shpec(&zero,&zero,&zero);
  shbec(&zero,&zero,&zero);

  n=MFEnumDualPolytopeDimension(EP);
  verbose=0;
  if(verbose){printf("MFDrawEnumDualPolytope dim =%d\n",n);fflush(stdout);}

  if(n<1||n>3)return;

/* Vertices */

  shlinc(&zero,&zero,&zero);
  if(0)
   {
    for(i=0;i<MFEnumDualPolytopeNumberOfCells(EP,0);i++)
     {
      MFMarkPoint((MFChart)NULL,MFEnumDualPolytopeVertex(EP,i),3,(MFChartState)NULL);
     }
   }

/* Edges */

  k=MFEnumDualPolytopeDimension(EP);
  if(1)for(i=0;i<MFEnumDualPolytopeNumberOfCells(EP,1);i++)
   {
    if(k==2 && MFEnumDualPolytopeNumberOfFaceCells(EP,1,i)<2)
      shlinc(&full,&zero,&zero);
     else
      shlinc(&zero,&zero,&zero);

    if(MFEnumDualPolytopeNumberOfCellFaces(EP,1,i)>1)
     {
      v0=MFEnumDualPolytopeCellFace(EP,1,i,0);
      v1=MFEnumDualPolytopeCellFace(EP,1,i,1);
      if(verbose){printf("edge %d joins vertices %d and %d\n",i,v0,v1);fflush(stdout);}
      if(MFEnumDualPolytopeVertex(EP,v0)!=(MFNVector)NULL
       &&MFEnumDualPolytopeVertex(EP,v1)!=(MFNVector)NULL)
       {
        MFDrawProjectTo3d((MFChart)NULL,MFEnumDualPolytopeVertex(EP,v0),&x0,&y0,&z0);
        MFDrawProjectTo3d((MFChart)NULL,MFEnumDualPolytopeVertex(EP,v1),&x1,&y1,&z1);
        shline(&x0,&y0,&z0,&x1,&y1,&z1);
       }
     }
   }

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return;
 }

#include <MFFortran.h>
double K(double);

int MFDrawGetData(MFChart chart,MFNVector u, double *l)
 {
  static char RoutineName[]={"MFDrawGetData"};
  int i,j,n;
  int nx,nu,np;
  double a,c,t,x1,x2,x3,y1,y2,y3;
  double kp,kp0;
  int min,max;
  double d1x,d1y,d1z;
  double mx0,mz,f,m3,theta0;
  double delta,up,um,u0;
  double m,T,Omega,R,Wr;

  if(l==(double*)NULL)
   {
    if(chart!=(MFChart)NULL && !strcmp(MFImplicitMFId(MFChartGetManifold(chart)),"TPBVP"))return 6;
     else return 0;
   }

  n=MFNV_NC(u);

#ifdef HAVE_DRF
  if(chart!=(MFChart)NULL && !strcmp(MFImplicitMFId(MFChartGetManifold(chart)),"TPBVP"))
   {
    nu=MFTPBVPGetNU(MFChartGetManifold(chart));
    nx=MFTPBVPGetNX(MFChartGetManifold(chart));
    np=MFTPBVPGetNP(MFChartGetManifold(chart));
    min=0;max=0;
    a=MFNV_C(u,nx*nu+0);
    c=MFNV_C(u,nx*nu+1);
    t=MFNV_C(u,nx*nu+2);
    x1=(MFNV_C(u,0)+MFNV_C(u,nu+0))/2;
    y1=(MFNV_C(u,1)+MFNV_C(u,nu+1))/2;
    x2=(MFNV_C(u,2)+MFNV_C(u,nu+2))/2;
    y2=(MFNV_C(u,3)+MFNV_C(u,nu+3))/2;
    x3=(MFNV_C(u,4)+MFNV_C(u,nu+4))/2;
    y3=(MFNV_C(u,5)+MFNV_C(u,nu+5))/2;
    kp0=t*x1*y1+(t*x2-c)*y2;

    mx0=c;
    mz=a;
    f=t;
    theta0=MFNV_C(u,nx*nu+3);
    m3=mx0*y1+mz*y3;

    u0=cos(theta0);
    delta=sqrt((mz*mz+mx0*mx0)*(mz*mz+mx0*mx0)+8*f*(2*f+u0*(mx0*mx0-mz*mz))
                 -16*mz*mx0*f*sin(theta0));
    up=(mz*mz+mx0*mx0+delta)/4/f;
    um=(mz*mz+mx0*mx0-delta)/4/f;

    nu=up-u0;
    if(um<u0)nu=up-um;
    m=fabs(u0-um)/nu;
    T=2*sqrt(2.)*K(m)/sqrt(f*nu);
    Omega=2*3.1415926/T;
    R=0.;
    Wr=0.;

    d1x=0.;
    d1y=1.;
    d1z=0.;
    for(i=1;i<nx-1;i++)
     {
      x1=(MFNV_C(u,nu*i+0)+MFNV_C(u,nu*(i+1)+0))/2;
      y1=(MFNV_C(u,nu*i+1)+MFNV_C(u,nu*(i+1)+1))/2;
      x2=(MFNV_C(u,nu*i+2)+MFNV_C(u,nu*(i+1)+2))/2;
      y2=(MFNV_C(u,nu*i+3)+MFNV_C(u,nu*(i+1)+3))/2;
      x3=(MFNV_C(u,nu*i+4)+MFNV_C(u,nu*(i+1)+4))/2;
      y3=(MFNV_C(u,nu*i+5)+MFNV_C(u,nu*(i+1)+5))/2;
      kp=t*x1*y1+(t*x2-c)*y2;
      if(kp*kp0<=0.)
       {
        if(i==1)
         {
          if(kp>0.)min+=1;
          if(kp<0.)max+=1;
         }else{
          if(kp>0.)min+=2;
          if(kp<0.)max+=2;
         }
       }
      kp0=kp;
     }
    l[0]=theta0;
    l[1]=x1*y1+x2*y2+x3*y3;
    l[2]=min;
    l[3]=max;
    l[4]=Omega;
    l[5]=y3;
/*  l[6]=Wr;
    l[7]=R;*/
   }else if(chart!=(MFChart)NULL)
#endif
   {
    l[0]=MFChartHasBoundary(chart);
    l[1]=MFChartIsSingular(chart);
    l[2]=0.;
   }


  return 0;
 }

#ifdef HAVE_DRF
double K(double x)
 {
  double zero=0.;
  double one=1.;
  int ier;

/*x=1-x;*/
  x=1-x*x;
  return F77_FUNC(drf,DRF)(&zero,&x,&one,&ier);
 }
#endif
