/* 
    @(#)MFDrawClippedSphere.c	1.5
    02/04/19 16:29:46
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFDrawClippedSphere.h>
#include <MFAtlas.h>
#include <MFChart.h>
#include <MFNVector.h>
#include <MFKVector.h>
#include <MFEnumPolytope.h>
#include <math.h>
#include <stdlib.h>
#include <sh.h>
#include <MFPrint.h>
#include <MFDrawChartState.h>

#define sqrt3 1.73205080

struct MFDrawClippedSpherePolygonListSt { 
        int nPolygons;
        int maxPolygon;
        double **polygon;
        int *nVertices;
        float **normal;
       };
MFChart MFAtlasChart(MFAtlas,int);
int MFDrawTestInChart3d(MFChart,double,double,double);
typedef struct MFDrawClippedSpherePolygonListSt *MFDrawClippedSpherePolygonList;

MFDrawClippedSpherePolygonList MFDrawCreateClippedSpherePolygonList(void);
void MFDrawCheckClippedSpherePolygonListSize(MFDrawClippedSpherePolygonList);
void MFDrawFreeClippedSpherePolygonList(MFDrawClippedSpherePolygonList);

int MFDrawAddTriangle(MFDrawClippedSpherePolygonList,double,double,double,double,double,double,double,double,double);
int MFDrawAddPolygonList(MFDrawClippedSpherePolygonList,int,double*);
void MFDrawFreePolygonList(int,MFDrawClippedSpherePolygonList);
void MFDrawCheckTriangle(int,double,double,double,double,MFDrawClippedSpherePolygonList,MFChart,double,double,double);
void MFDrawClipTriangle(double,double,double,double,double,double,double,double,double,double,double,double,double,int*,double*,MFChart,double);
void MFDrawEdgeBisection(double,double,double,double,double,double,double,int,double,double,double,int,MFChart,double*,double*,double*,double);
void MFDrawProjectTo3d(MFChart,MFNVector,float*,float*,float*);

extern MFAtlas MFMFA;

void MFDrawClippedSphere(MFChart chart, double epsilon1, double epsilon2, double epsilon3, MFChartState state)
 {
  static char RoutineName[]={"MFDrawClippedSphere"};
  double h;
  double xx,yy,zz;
  double xa,ya,za;
  double xb,yb,zb;
  double xc,yc,zc;
  double xma,yma,zma;
  double xmb,ymb,zmb;
  double xmc,ymc,zmc;
  double d;
  int i,j,n;
  float *x,*y,*z;
  MFDrawClippedSpherePolygonList pList;
  MFKVector s;
  MFNVector u;
  double x0,y0,z0,r;

/*   Draws the portion of a sphere which satisies test(x,y,z)>0             */
/*                                                                          */
/* (x0,y0,z0) origin of sphere                                              */
/* r          radius of sphere                                              */
/* epsilon1   size of largest triangle to draw                              */
/* epsilon2   size of smallest triangle to draw                             */
/* epsilon3   size of error in points on boundary                           */
/* test       routine to use to determine if a point is to be drawn or not. */

  x0=0.;
  y0=0.;
  z0=0.;
  r =MFChartRadius(chart);
  
  pList=MFDrawCreateClippedSpherePolygonList();

  s=MFCreateKVector(3);
  u=MFCreateNVector(MFChartN(chart));

/* Initial Tetrahedron */
 
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0+r,y0  ,z0  ,x0  ,y0+r,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0  ,y0+r,z0  ,x0-r,y0  ,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0-r,y0  ,z0  ,x0  ,y0-r,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0  ,y0-r,z0  ,x0+r,y0  ,z0  );

  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0  ,y0+r,z0  ,x0+r,y0  ,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0-r,y0  ,z0  ,x0  ,y0+r,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0  ,y0-r,z0  ,x0-r,y0  ,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0+r,y0  ,z0  ,x0  ,y0-r,z0);

/* Now subdivide until smallest side is small enough */

  h=r*sqrt(2.);
  while(h>epsilon1)
   {
    n=pList->nPolygons;
    for(i=0;i<n;i++)
     {
      xa=(pList->polygon[i])[0];
      ya=(pList->polygon[i])[1];
      za=(pList->polygon[i])[2];
      xb=(pList->polygon[i])[3];
      yb=(pList->polygon[i])[4];
      zb=(pList->polygon[i])[5];
      xc=(pList->polygon[i])[6];
      yc=(pList->polygon[i])[7];
      zc=(pList->polygon[i])[8];

/* Compute midpoints of the sides, and project out onto the sphere */

      xma=(xa+xb)*.5;
      yma=(ya+yb)*.5;
      zma=(za+zb)*.5;
      d=r/sqrt(pow(xma-x0,2)+pow(yma-y0,2)+pow(zma-z0,2));
      xma=(xma-x0)*d+x0;
      yma=(yma-y0)*d+y0;
      zma=(zma-z0)*d+z0;
      xmb=(xb+xc)*.5;
      ymb=(yb+yc)*.5;
      zmb=(zb+zc)*.5;
      d=r/sqrt(pow(xmb-x0,2)+pow(ymb-y0,2)+pow(zmb-z0,2));
      xmb=(xmb-x0)*d+x0;
      ymb=(ymb-y0)*d+y0;
      zmb=(zmb-z0)*d+z0;
      xmc=(xc+xa)*.5;
      ymc=(yc+ya)*.5;
      zmc=(zc+za)*.5;
      d=r/sqrt(pow(xmc-x0,2)+pow(ymc-y0,2)+pow(zmc-z0,2));
      xmc=(xmc-x0)*d+x0;
      ymc=(ymc-y0)*d+y0;
      zmc=(zmc-z0)*d+z0;
      (pList->polygon[i])[0]=xma;
      (pList->polygon[i])[1]=yma;
      (pList->polygon[i])[2]=zma;
      (pList->polygon[i])[3]=xmb;
      (pList->polygon[i])[4]=ymb;
      (pList->polygon[i])[5]=zmb;
      (pList->polygon[i])[6]=xmc;
      (pList->polygon[i])[7]=ymc;
      (pList->polygon[i])[8]=zmc;
      MFDrawAddTriangle(pList, xa ,ya ,za ,xma,yma,zma,xmc,ymc,zmc);
      MFDrawAddTriangle(pList, xb ,yb ,zb ,xmb,ymb,zmb,xma,yma,zma);
      MFDrawAddTriangle(pList, xc ,yc ,zc ,xmc,ymc,zmc,xmb,ymb,zmb);
     }
    h=h/2;
   }

/* Now have a list of triangles covering the sphere, with no edge larger than epsilon1 */
  
  n=pList->nPolygons;
  for(i=0;i<n;i++)
   {
    MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,h,epsilon2,epsilon3);
   }
  
/* Compute the normals to the sphere at each point */

  for(i=0;i<pList->nPolygons;i++)
   {
    n=pList->nVertices[i];
    if(n>0)
     {
      pList->normal[i]=(float*)malloc(3*n*sizeof(float)); /*halfdone*/
      if(pList->normal[i]==(float*)NULL)
       {
        fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a float array of length %d\n",__LINE__,__FILE__,3*n);
        exit(12);
       }
      for(j=0;j<n;j++)
       {
        (pList->normal[i])[3*j+0]=((pList->polygon[i])[3*j+0]-x0)/r;
        (pList->normal[i])[3*j+1]=((pList->polygon[i])[3*j+1]-y0)/r;
        (pList->normal[i])[3*j+2]=((pList->polygon[i])[3*j+2]-z0)/r;
       }
     }
   }

/* draw it */

  x=(float*)NULL;
  y=(float*)NULL;
  z=(float*)NULL;
  for(i=0;i<pList->nPolygons;i++)
   {
    n=pList->nVertices[i];
    if(n>0)
     {
      x=(float*)realloc((void*)x,n*sizeof(float)); /*halfdone*/
      if(x==(float*)NULL)
       {
        fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a float array of length %d\n",__LINE__,__FILE__,n);
        exit(12);
       }
      y=(float*)realloc((void*)y,n*sizeof(float)); /*halfdone*/
      if(y==(float*)NULL)
       {
        fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a float array of length %d\n",__LINE__,__FILE__,n);
        exit(12);
       }
      z=(float*)realloc((void*)z,n*sizeof(float)); /*halfdone*/
      if(z==(float*)NULL)
       {
        fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a float array of length %d\n",__LINE__,__FILE__,n);
        exit(12);
       }
      for(j=0;j<n;j++)
       {
        MFKVSetC(s,0,(pList->polygon[i])[3*j  ]);
        MFKVSetC(s,1,(pList->polygon[i])[3*j+1]);
        MFKVSetC(s,2,(pList->polygon[i])[3*j+2]);
        MFChartEvaluate(chart,s,u);
        MFDrawProjectTo3d(chart,u,x+j,y+j,z+j);
       }
      if(state!=(MFChartState)NULL)MFChartStateAddPolygonWithNormal(state,n,x,y,z,pList->normal[i],1);
       else
        shpg(&n,x,y,z);
     }
   }
  free(x);
  free(y);
  free(z);
  MFFreeKVector(s);
  MFFreeNVector(u);

/* cleanup */

  MFDrawFreeClippedSpherePolygonList(pList);

  return;
 }

int MFDrawAddTriangle(MFDrawClippedSpherePolygonList pList, double x0, double y0,double z0, double x1, double y1,double z1, double x2, double y2,double z2)
 {
  MFDrawCheckClippedSpherePolygonListSize(pList);

  (pList->nVertices)[pList->nPolygons]=3;
  (pList->polygon)[pList->nPolygons]=(double*)malloc(3*3*sizeof(double)); /*halfdone*/
  if((pList->polygon)[pList->nPolygons]==(double*)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a double array of length %d\n",__LINE__,__FILE__,9);
    exit(12);
   }
  ((pList->polygon)[pList->nPolygons])[0+3*0]=x0;
  ((pList->polygon)[pList->nPolygons])[1+3*0]=y0;
  ((pList->polygon)[pList->nPolygons])[2+3*0]=z0;
  ((pList->polygon)[pList->nPolygons])[0+3*1]=x1;
  ((pList->polygon)[pList->nPolygons])[1+3*1]=y1;
  ((pList->polygon)[pList->nPolygons])[2+3*1]=z1;
  ((pList->polygon)[pList->nPolygons])[0+3*2]=x2;
  ((pList->polygon)[pList->nPolygons])[1+3*2]=y2;
  ((pList->polygon)[pList->nPolygons])[2+3*2]=z2;
  (pList->nPolygons)++;
  return((pList->nPolygons)-1);
 }

void MFDrawFreePolygonList(int i, MFDrawClippedSpherePolygonList pList)
 {
/* Delete polygon i by setting the number of vertices to -1*/

  if(i<0 || i>=pList->nPolygons)
   {
    fprintf(stderr,"MFDrawFreePolygon: attempt to delete a non-existant polygon, (must have 0 < i(=%d) <= nPolygons(=%d))\n",i, pList->nPolygons);
    exit(12);
   }

  (pList->nVertices)[i]=-1;
  return;
 }

int MFDrawAddPolygonList(MFDrawClippedSpherePolygonList pList, int n, double* x)
 {
  int i;
  MFDrawCheckClippedSpherePolygonListSize(pList);

  (pList->nVertices)[pList->nPolygons]=n;
  (pList->polygon)[pList->nPolygons]=(double*)malloc(3*n*sizeof(double)); /*halfdone*/
  if((pList->polygon)[pList->nPolygons]==(double*)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a double array of length %d\n",__LINE__,__FILE__,3*n);
    exit(12);
   }
  for(i=0;i<3*n;i++)((pList->polygon)[pList->nPolygons])[i]=x[i];
  (pList->nPolygons)++;
  return((pList->nPolygons)-1);
 }

void MFDrawCheckTriangle(int i, double x0, double y0, double z0, double r, MFDrawClippedSpherePolygonList pList, MFChart chart,double size, double epsilon1,double epsilon2 )
 {

/* Check and subdivide a triangle */

  double xa,ya,za;
  double xb,yb,zb;
  double xc,yc,zc;
  double xma,yma,zma;
  double xmb,ymb,zmb;
  double xmc,ymc,zmc;
  double d;
  int a,b,c,ma,mb,mc;
  int n=0;
  double *x=(double*)NULL;

  if((pList->nVertices)[i]!=3)
   {
    fprintf(stderr,"MFDrawCheckTriangle: attempt to delete a non-triangular polygon, (nVertices=%d)\n",(pList->nVertices)[i]);
    exit(12);
   }

/* Get vertices, and check for all in or all out */

  xa=((pList->polygon)[i])[0];
  ya=((pList->polygon)[i])[1];
  za=((pList->polygon)[i])[2];
  a=MFDrawTestInChart3d(chart,xa,ya,za);

  xb=((pList->polygon)[i])[3];
  yb=((pList->polygon)[i])[4];
  zb=((pList->polygon)[i])[5];
  b=MFDrawTestInChart3d(chart,xb,yb,zb);

  xc=((pList->polygon)[i])[6];
  yc=((pList->polygon)[i])[7];
  zc=((pList->polygon)[i])[8];
  c=MFDrawTestInChart3d(chart,xc,yc,zc);

  if(a && b && c )return;
  MFDrawFreePolygonList(i,pList);
  if(!a && !b && !c )return;

/* Compute midpoints of the sides, and project out onto the sphere */

  xma=(xa+xb)*.5;
  yma=(ya+yb)*.5;
  zma=(za+zb)*.5;
  d=r/sqrt(pow(xma-x0,2)+pow(yma-y0,2)+pow(zma-z0,2));
  xma=(xma-x0)*d+x0;
  yma=(yma-y0)*d+y0;
  zma=(zma-z0)*d+z0;
  ma=MFDrawTestInChart3d(chart,xma,yma,zma);

  xmb=(xb+xc)*.5;
  ymb=(yb+yc)*.5;
  zmb=(zb+zc)*.5;
  d=r/sqrt(pow(xmb-x0,2)+pow(ymb-y0,2)+pow(zmb-z0,2));
  xmb=(xmb-x0)*d+x0;
  ymb=(ymb-y0)*d+y0;
  zmb=(zmb-z0)*d+z0;
  mb=MFDrawTestInChart3d(chart,xmb,ymb,zmb);

  xmc=(xc+xa)*.5;
  ymc=(yc+ya)*.5;
  zmc=(zc+za)*.5;
  d=r/sqrt(pow(xmc-x0,2)+pow(ymc-y0,2)+pow(zmc-z0,2));
  xmc=(xmc-x0)*d+x0;
  ymc=(ymc-y0)*d+y0;
  zmc=(zmc-z0)*d+z0;
  mc=MFDrawTestInChart3d(chart,xmc,ymc,zmc);

/* Add and check the four smaller triangles */

  x=(double*)malloc(30*sizeof(double)); /*halfdone*/
  if(x==(double*)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a double array of length %d\n",__LINE__,__FILE__,30);
    exit(12);
   }

  if(a || ma || mc )
   {
    if(a && ma && mc )
      MFDrawAddTriangle(pList, xa ,ya ,za ,xma,yma,zma,xmc,ymc,zmc);
    else
     {
      if(size>epsilon1)
       {
        i=MFDrawAddTriangle(pList, xa ,ya ,za ,xma,yma,zma,xmc,ymc,zmc);
        MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,size/2.,epsilon1,epsilon2);
       }else{
        MFDrawClipTriangle(x0,y0,z0,r, xa ,ya ,za ,xma,yma,zma,xmc,ymc,zmc,&n,x,chart,epsilon2);
        MFDrawAddPolygonList(pList,n,x);
       }
     }
   }

  if(b || mb || ma )
   {
    if(b && mb && ma )
      MFDrawAddTriangle(pList, xb ,yb ,zb ,xmb,ymb,zmb,xma,yma,zma);
    else
     {
      if(size>epsilon1)
       {
        i=MFDrawAddTriangle(pList, xb ,yb ,zb ,xmb,ymb,zmb,xma,yma,zma);
        MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,size/2.,epsilon1,epsilon2);
       }else{
        MFDrawClipTriangle(x0,y0,z0,r, xb ,yb ,zb ,xmb,ymb,zmb,xma,yma,zma,&n,x,chart,epsilon2);
        MFDrawAddPolygonList(pList,n,x);
       }
     }
   }

  if(c || mc || mb )
   {
    if(c && mc && mb )
      MFDrawAddTriangle(pList, xc ,yc ,zc ,xmc,ymc,zmc,xmb,ymb,zmb);
    else
     {
      if(size>epsilon1)
       {
        i=MFDrawAddTriangle(pList, xc ,yc ,zc ,xmc,ymc,zmc,xmb,ymb,zmb);
        MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,size/2.,epsilon1,epsilon2);
       }else{
        MFDrawClipTriangle(x0,y0,z0,r, xc ,yc ,zc ,xmc,ymc,zmc,xmb,ymb,zmb,&n,x,chart,epsilon2);
        MFDrawAddPolygonList(pList,n,x);
       }
     }
   }

  if(ma || mb || mc )
   {
    if(ma && mb && mc )
      MFDrawAddTriangle(pList,xma,yma,zma,xmb,ymb,zmb,xmc,ymc,zmc);
    else
     {
      if(size>epsilon1)
       {
        i=MFDrawAddTriangle(pList, xma,yma,zma,xmb,ymb,zmb,xmc,ymc,zmc);
        MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,size/2.,epsilon1,epsilon2);
       }else{
        MFDrawClipTriangle(x0,y0,z0,r, xma,yma,zma,xmb,ymb,zmb,xmc,ymc,zmc,&n,x,chart,epsilon2);
        MFDrawAddPolygonList(pList,n,x);
       }
     }
   }
  free(x);
  return;
 }

void MFDrawClipTriangle(double x0, double y0, double z0, double r, double xa, double ya, double za, double xb, double yb, double zb, double xc, double yc, double zc, int *n, double *x2, MFChart chart,double epsilon)
 {

/*    clip a triangle     */

   int i,i1,j,N;
   int in0,in1;
   double *x;
   double xx=0.;
   double yy=0.;
   double zz=0.;

   x=(double*)malloc(9*sizeof(double)); /*halfdone*/
   if(x==(double*)NULL)
    {
     fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a double array of length %d\n",__LINE__,__FILE__,9);
     exit(12);
    }

   x[0]=xa;
   x[1]=ya;
   x[2]=za;
   x[3]=xb;
   x[4]=yb;
   x[5]=zb;
   x[6]=xc;
   x[7]=yc;
   x[8]=zc;

   in0=MFDrawTestInChart3d(chart,x[0],x[1],x[2]);

   j=0;
   for(i=0;i<3;i++)
    {
     i1=(i+1)%3;
     in1=MFDrawTestInChart3d(chart,x[0+3*i1],x[1+3*i1],x[2+3*i1]);

     if(in0 && in1)
      {
       x2[3*j]  =x[3*i];
       x2[3*j+1]=x[3*i+1];
       x2[3*j+2]=x[3*i+2];
       if(j==0 || fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])>1.e-5)j++;

       x2[3*j]  =x[3*i1];
       x2[3*j+1]=x[3*i1+1];
       x2[3*j+2]=x[3*i1+2];
       if(j==0 || fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])>1.e-5)j++;
      }else if(!in0 && in1)
      {
       MFDrawEdgeBisection(x0,y0,z0,r,x[3*i],x[3*i+1],x[3*i+2],in0,x[3*i1],x[3*i1+1],x[3*i1+2],in1,chart,&xx,&yy,&zz,epsilon);

       x2[3*j]  =xx;
       x2[3*j+1]=yy;
       x2[3*j+2]=zz;
       if(j==0 || fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])>1.e-5)j++;

       x2[3*j]  =x[3*i1];
       x2[3*j+1]=x[3*i1+1];
       x2[3*j+2]=x[3*i1+2];
       if(j==0 || fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])>1.e-5)j++;
      }else if(in0&&!in1)
      {
       MFDrawEdgeBisection(x0,y0,z0,r,x[3*i],x[3*i+1],x[3*i+2],in0,x[3*i1],x[3*i1+1],x[3*i1+2],in1,chart,&xx,&yy,&zz,epsilon);
       x2[3*j]  =xx;
       x2[3*j+1]=yy;
       x2[3*j+2]=zz;
       if(j==0 || fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])>1.e-5)j++;
      }
     in0=in1;
    }

   if(j>0 && fabs(x2[3*j]-x2[3*j-2])+fabs(x2[3*j+1]-x2[3*j-1])+fabs(x2[3*j+2]-x2[3*j])<=1.e-5)j--;
   *n=j;

   free(x);
   return;
 }

void MFDrawEdgeBisection(double xo,double yo,double zo,double r,double x0,double y0,double z0,int l0,double x1,double y1,double z1,int l1, MFChart chart, double *x, double *y,double *z, double epsilon)
 {
  int l;
  double d;

  epsilon=3*epsilon;
  if(l0)
   {
    *x=x0;
    *y=y0;
    *z=z0;
   }else{
    *x=x1;
    *y=y1;
    *z=z1;
   }

  while(fabs(x1-x0)+fabs(y1-y0)+fabs(z1-z0)>epsilon)
   {
    *x=.5*(x0+x1);
    *y=.5*(y0+y1);
    *z=.5*(z0+z1);
    d=r/sqrt(pow(*x-xo,2)+pow(*y-yo,2)+pow(*z-zo,2));
    *x=(*x-xo)*d+xo;
    *y=(*y-yo)*d+yo;
    *z=(*z-zo)*d+zo;
    l=MFDrawTestInChart3d(chart,*x,*y,*z);

    if(l0&&!l || !l0&&l)
     {
      x1=*x;
      y1=*y;
      z1=*z;
      l1=l;
     }else{
      x0=*x;
      y0=*y;
      z0=*z;
      l0=l;
     }
   }
  return;
 }

MFDrawClippedSpherePolygonList MFDrawCreateClippedSpherePolygonList()
 {
  int i;
  MFDrawClippedSpherePolygonList pList;

  pList=(MFDrawClippedSpherePolygonList)malloc(sizeof(struct MFDrawClippedSpherePolygonListSt)); /*halfdone*/
  if(pList==(MFDrawClippedSpherePolygonList)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate a struct MFDrawClippedSpherePolygonList of length %d\n",__LINE__,__FILE__,sizeof(struct MFDrawClippedSpherePolygonListSt));
    exit(12);
   }
  pList->maxPolygon=100;
  pList->polygon=(double**)malloc(pList->maxPolygon*sizeof(double*)); /*halfdone*/
  if(pList->polygon==(double**)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of double* of length %d\n",__LINE__,__FILE__,pList->maxPolygon);
    exit(12);
   }
  pList->nVertices=(int*)malloc(pList->maxPolygon*sizeof(int)); /*halfdone*/
  if(pList->nVertices==(int*)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of int of length %d\n",__LINE__,__FILE__,pList->maxPolygon);
    exit(12);
   }
  for(i=0;i<pList->maxPolygon;i++)pList->nVertices[i]=-1;
  pList->nPolygons=0;
  pList->normal=(float**)malloc(pList->maxPolygon*sizeof(float*)); /*halfdone*/
  if(pList->normal==(float**)NULL)
   {
    fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of float* of length %d\n",__LINE__,__FILE__,pList->maxPolygon);
    exit(12);
   }

  return(pList);
 }

void MFDrawCheckClippedSpherePolygonListSize(MFDrawClippedSpherePolygonList pList)
 {
  int i;

  if(pList->nPolygons>=pList->maxPolygon)
   {
    pList->maxPolygon+=100;
    pList->polygon=(double**)realloc(pList->polygon,3*(pList->maxPolygon)*sizeof(double*)); /*halfdone*/
    if(pList->polygon==(double**)NULL) /*halfdone*/
     {
      fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of double* of length %d\n",__LINE__,__FILE__,3*pList->maxPolygon);
      exit(12);
     }
    pList->normal=(float**)realloc(pList->normal,3*(pList->maxPolygon)*sizeof(float*)); /*halfdone*/
    if(pList->normal==(float**)NULL)
     {
      fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of float* of length %d\n",__LINE__,__FILE__,3*pList->maxPolygon);
      exit(12);
     }
    pList->nVertices=(int*)realloc(pList->nVertices,(pList->maxPolygon)*sizeof(int)); /*halfdone*/
    if(pList->nVertices==(int*)NULL)
     {
      fprintf(stderr,"Out of memory at line %d in file %s, trying to allocate an array of int of length %d\n",__LINE__,__FILE__,pList->maxPolygon);
      exit(12);
     }
    for(i=(pList->maxPolygon)-100;i<(pList->maxPolygon);i++)(pList->nVertices)[i]=-1;
   }
 }

void MFDrawFreeClippedSpherePolygonList(MFDrawClippedSpherePolygonList pList)
 {
   int i;
   
  for(i=0;i<pList->nPolygons;i++)
   {
    free(pList->polygon[i]);
    if(pList->nVertices[i]>0)free(pList->normal[i]);
   }
  free(pList->polygon);
  free(pList->normal);
  free(pList->nVertices);
  free(pList);
 
  return;
 }

MFKVector MFTESTPT=(MFKVector)NULL;

int MFDrawTestInChart3d(MFChart chart,double x,double y,double z)
 {
  if(MFTESTPT==(MFKVector)NULL)MFTESTPT=MFCreateKVector(MFChartK(chart));
  MFKVSetC(MFTESTPT,0,x);
  MFKVSetC(MFTESTPT,1,y);
  MFKVSetC(MFTESTPT,2,z);
  return MFPolytopeInterior(MFChartPolytope(chart),MFTESTPT);
 }

void MFTestClippedSphere(MFChart chart, double epsilon1, double epsilon2, double epsilon3)
 {
  double h;
  double xx,yy,zz;
  double xa,ya,za;
  double xb,yb,zb;
  double xc,yc,zc;
  double xma,yma,zma;
  double xmb,ymb,zmb;
  double xmc,ymc,zmc;
  double d;
  int i,j,n;
  float *x=(float*)NULL;
  float *y=(float*)NULL;
  float *z=(float*)NULL;
  MFDrawClippedSpherePolygonList pList;
  MFKVector s;
  MFNVector u;
  double x0,y0,z0,r;
  int jj;

/*   Draws the portion of a sphere which satisies test(x,y,z)>0             */
/*                                                                          */
/* (x0,y0,z0) origin of sphere                                              */
/* r          radius of sphere                                              */
/* epsilon1   size of largest triangle to draw                              */
/* epsilon2   size of smallest triangle to draw                             */
/* epsilon3   size of error in points on boundary                           */
/* test       routine to use to determine if a point is to be drawn or not. */

  x0=0.;
  y0=0.;
  z0=0.;
  r =MFChartRadius(chart);
  
  pList=MFDrawCreateClippedSpherePolygonList();

  s=MFCreateKVector(3);
  u=MFCreateNVector(MFChartN(chart));

/* Initial Tetrahedron */
 
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0+r,y0  ,z0  ,x0  ,y0+r,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0  ,y0+r,z0  ,x0-r,y0  ,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0-r,y0  ,z0  ,x0  ,y0-r,z0  );
  MFDrawAddTriangle(pList, x0  ,y0  ,z0+r,x0  ,y0-r,z0  ,x0+r,y0  ,z0  );

  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0  ,y0+r,z0  ,x0+r,y0  ,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0-r,y0  ,z0  ,x0  ,y0+r,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0  ,y0-r,z0  ,x0-r,y0  ,z0);
  MFDrawAddTriangle(pList, x0  ,y0  ,z0-r,x0+r,y0  ,z0  ,x0  ,y0-r,z0);

/* Now subdivide until smallest side is small enough */

  h=r*sqrt(2.);
  while(h>epsilon1)
   {
    n=pList->nPolygons;
    for(i=0;i<n;i++)
     {
      xa=(pList->polygon[i])[0];
      ya=(pList->polygon[i])[1];
      za=(pList->polygon[i])[2];
      xb=(pList->polygon[i])[3];
      yb=(pList->polygon[i])[4];
      zb=(pList->polygon[i])[5];
      xc=(pList->polygon[i])[6];
      yc=(pList->polygon[i])[7];
      zc=(pList->polygon[i])[8];

/* Compute midpoints of the sides, and project out onto the sphere */

      xma=(xa+xb)*.5;
      yma=(ya+yb)*.5;
      zma=(za+zb)*.5;
      d=r/sqrt(pow(xma-x0,2)+pow(yma-y0,2)+pow(zma-z0,2));
      xma=(xma-x0)*d+x0;
      yma=(yma-y0)*d+y0;
      zma=(zma-z0)*d+z0;
      xmb=(xb+xc)*.5;
      ymb=(yb+yc)*.5;
      zmb=(zb+zc)*.5;
      d=r/sqrt(pow(xmb-x0,2)+pow(ymb-y0,2)+pow(zmb-z0,2));
      xmb=(xmb-x0)*d+x0;
      ymb=(ymb-y0)*d+y0;
      zmb=(zmb-z0)*d+z0;
      xmc=(xc+xa)*.5;
      ymc=(yc+ya)*.5;
      zmc=(zc+za)*.5;
      d=r/sqrt(pow(xmc-x0,2)+pow(ymc-y0,2)+pow(zmc-z0,2));
      xmc=(xmc-x0)*d+x0;
      ymc=(ymc-y0)*d+y0;
      zmc=(zmc-z0)*d+z0;
      (pList->polygon[i])[0]=xma;
      (pList->polygon[i])[1]=yma;
      (pList->polygon[i])[2]=zma;
      (pList->polygon[i])[3]=xmb;
      (pList->polygon[i])[4]=ymb;
      (pList->polygon[i])[5]=zmb;
      (pList->polygon[i])[6]=xmc;
      (pList->polygon[i])[7]=ymc;
      (pList->polygon[i])[8]=zmc;
      MFDrawAddTriangle(pList, xa ,ya ,za ,xma,yma,zma,xmc,ymc,zmc);
      MFDrawAddTriangle(pList, xb ,yb ,zb ,xmb,ymb,zmb,xma,yma,zma);
      MFDrawAddTriangle(pList, xc ,yc ,zc ,xmc,ymc,zmc,xmb,ymb,zmb);
     }
    h=h/2;
   }

/* Now have a list of triangles covering the sphere, with no edge larger than epsilon1 */
  
  n=pList->nPolygons;
  for(i=0;i<n;i++)
    MFDrawCheckTriangle(i,x0,y0,z0,r,pList,chart,h,epsilon2,epsilon3);

/* now have a list of triangles which are on the boundary of the chart, and inside the polytope */
/* Are these points inside any other chart? */

  for(i=0;i<pList->nPolygons;i++)
   {
    n=pList->nVertices[i];
    if(n>0)
     {
      for(j=0;j<n;j++)
       {
        MFKVSetC(s,0,(pList->polygon[i])[3*j  ]);
        MFKVSetC(s,1,(pList->polygon[i])[3*j+1]);
        MFKVSetC(s,2,(pList->polygon[i])[3*j+2]);
        MFChartEvaluate(chart,s,u);
        for(jj=0;jj<MFAtlasNumberOfCharts(MFMFA);jj++)
         {
          if(MFAtlasChart(MFMFA,j)!=chart)
           {
            MFChartProjectIntoTangentSpace(MFAtlasChart(MFMFA,j),u,s);
            if(MFChartInterior(MFAtlasChart(MFMFA,j),s))
              printf("ERROR: here's a point that is supposedly on the boundary, but which is in chart %d\n",j);fflush(stdout);
           }
         }
       }
     }
   }
  free(x);
  free(y);
  free(z);
  MFFreeKVector(s);
  MFFreeNVector(u);

/* cleanup */

  MFDrawFreeClippedSpherePolygonList(pList);

  return;
 }
