/* 
    %W%
    %D% %T%

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
   
*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   April 1, 2002                     */

#include <MFAtlas.h>
#include <MFContinuationMethod.h>

struct MFHendersonParmBlockST {
                             int verbose;
                             int maxCharts;
                             double minR;
                             double maxR;
                             double epsilon;
                             double dotmin;
                             int page;
                             int pageEvery;
                             int useBB;
                             int dumpToPlotFile;
                             int dumpToCenterFile;
                             int dumpToRestartFile;
                             int dumpToRestartFileEvery;
                             char *fileName;
                             int checkPoint;
                             int checkPointEvery;
                             int branchSwitch;
                             int nClipF;
                             int mClipF;
                             double (**clipF)(MFNVector);
                            };

typedef struct MFHendersonParmBlockST *MFHendersonParmBlock;
MFAtlas MFHendersonComputeAtlas(MFContinuationMethod,MFImplicitMF,MFNRegion,int,MFNVector*);
void MFHendersonMethodCopyClipF(MFContinuationMethod,MFAtlas);

char MFHendersonErrorMsg[256];
void MFSetError(int,char*,char*,int,char*);

MFContinuationMethod MFCreateHendersonsMethod()
 {
  static char RoutineName[]={"MFCreateHendersonsMethod"};
  MFContinuationMethod result;
  MFHendersonParmBlock parms;

  result=(MFContinuationMethod)malloc(sizeof(struct MFContinuationMethodST));
  if(result==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"Out of space trying to allocate %d bytes.",sizeof(struct MFContinuationMethodST));
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return (MFContinuationMethod)NULL;
   }

  result->type=(char*)malloc(10*sizeof(char));
  if(result->type==(char*)NULL)
   {
    sprintf(MFHendersonErrorMsg,"Out of space trying to allocate %d bytes.",10*sizeof(char));
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    free(result);
    return (MFContinuationMethod)NULL;
   }

  strcpy(result->type,"Henderson");
  parms=(MFHendersonParmBlock)malloc(sizeof(struct MFHendersonParmBlockST));
  if(parms==(MFHendersonParmBlock)NULL)
   {
    sprintf(MFHendersonErrorMsg,"Out of space trying to allocate %d bytes.",sizeof(struct MFHendersonParmBlockST));
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    free(result);
    free(result->type);
    return (MFContinuationMethod)NULL;
   }

  parms->verbose=0;
  parms->maxCharts=-1;
  parms->minR=.01;
  parms->maxR=1.;
  parms->epsilon=1.e-7;
  parms->dotmin=.2;
  parms->page=1;
  parms->pageEvery=1000;
  parms->useBB=1;
  parms->dumpToPlotFile=1;
  parms->dumpToCenterFile=1;
  parms->dumpToRestartFile=0;
  parms->dumpToRestartFileEvery=100;
  parms->checkPoint=0;
  parms->checkPointEvery=100;
  parms->branchSwitch=1;
  parms->fileName=(char*)malloc(6*sizeof(char));
  parms->nClipF=0;
  parms->mClipF=0;
  parms->clipF=(double (**)(MFNVector))NULL;
  strcpy(parms->fileName,"Atlas");
  result->parmBlock=parms;

  result->ComputeAtlas=MFHendersonComputeAtlas;
  result->ExtendAtlas=MFHendersonExtendAtlasMultiple;
  result->CloseAtlas=MFHendersonCloseAtlas;
  result->FlushAtlas=MFHendersonFlushAtlas;

  return result;
 }

void MFFreeHendersonsMethod(MFContinuationMethod H)
 { 
  MFHendersonParmBlock parms;

  free(H->type);
  parms=(MFHendersonParmBlock)(H->parmBlock);
  if(parms->fileName!=(char*)NULL)free(parms->fileName);
  if(parms->clipF!=(double(**)(MFNVector))NULL)free(parms->clipF);
  free(parms);
  free(H);

  return;
 }

void MFHendersonsMethodSetVerbose(MFContinuationMethod H,int verbose)
 {
  static char RoutineName[]={"MFHendersonsMethodSetVerbose"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->verbose=verbose;
  return;
 }

int MFHendersonsMethodGetVerbose(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetVerbose"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->verbose;
 }

void MFHendersonsMethodSetMaxCharts(MFContinuationMethod H,int maxCharts)
 {
  static char RoutineName[]={"MFHendersonsMethodSetMaxCharts"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->maxCharts=maxCharts;
  return;
 }

int MFHendersonsMethodGetMaxCharts(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetMaxCharts"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->maxCharts;
 }

void MFHendersonsMethodSetMinR(MFContinuationMethod H,int minR)
 {
  static char RoutineName[]={"MFHendersonsMethodSetMinR"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->minR=minR;
  return;
 }

double MFHendersonsMethodGetMinR(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetMinR"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->minR;
 }

void MFHendersonsMethodSetMaxR(MFContinuationMethod H,int maxR)
 {
  static char RoutineName[]={"MFHendersonsMethodSetMaxR"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->maxR=maxR;
  return;
 }

double MFHendersonsMethodGetMaxR(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetMaxR"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->maxR;
 }
void MFHendersonsMethodSetEpsilon(MFContinuationMethod H,double epsilon)
 {
  static char RoutineName[]={"MFHendersonsMethodSetEpsilon"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->epsilon=epsilon;
  return;
 }

double MFHendersonsMethodGetEpsilon(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetEpsilon"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->epsilon;
 }

void MFHendersonsMethodSetDotMin(MFContinuationMethod H,double dotmin)
 {
  static char RoutineName[]={"MFHendersonsMethodSetDotMin"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->dotmin=dotmin;
  return;
 }

double MFHendersonsMethodGetDotMin(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetDotMin"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->dotmin;
 }

void MFHendersonsMethodSetPage(MFContinuationMethod H,int page)
 {
  static char RoutineName[]={"MFHendersonsMethodSetPage"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->page=page;
  return;
 }

int MFHendersonsMethodGetPage(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetPage"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->page;
 }

void MFHendersonsMethodSetPageEvery(MFContinuationMethod H,int pageEvery)
 {
  static char RoutineName[]={"MFHendersonsMethodSetPageEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->pageEvery=pageEvery;
  return;
 }

int MFHendersonsMethodGetPageEvery(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetPageEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->pageEvery;
 }

void MFHendersonsMethodSetDumpToPlotFile(MFContinuationMethod H,int dumpToPlotFile)
 {
  static char RoutineName[]={"MFHendersonsMethodSetDumpToPlotFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->dumpToPlotFile=dumpToPlotFile;
  return;
 }

int MFHendersonsMethodGetDumpToPlotFile(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetDumpToPlotFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->dumpToPlotFile;
 }

void MFHendersonsMethodSetDumpToCenterFile(MFContinuationMethod H,int dumpToCenterFile)
 {
  static char RoutineName[]={"MFHendersonsMethodSetDumpToCenterFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->dumpToCenterFile=dumpToCenterFile;
  return;
 }

int MFHendersonsMethodGetDumpToCenterFile(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetDumpToCenterFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->dumpToCenterFile;
 }

void MFHendersonsMethodSetCheckPoint(MFContinuationMethod H,int checkPoint)
 {
  static char RoutineName[]={"MFHendersonsMethodSetCheckPoint"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->checkPoint=checkPoint;
  return;
 }

int MFHendersonsMethodGetCheckPoint(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetCheckPoint"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->checkPoint;
 }

void MFHendersonsMethodSetCheckPointEvery(MFContinuationMethod H,int checkPointEvery)
 {
  static char RoutineName[]={"MFHendersonsMethodSetCheckPointEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->checkPointEvery=checkPointEvery;
  return;
 }

int MFHendersonsMethodGetCheckPointEvery(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetCheckPointEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->checkPointEvery;
 }

void MFHendersonsMethodSetBranchSwitch(MFContinuationMethod H,int branchSwitch)
 {
  static char RoutineName[]={"MFHendersonsMethodSetBranchSwitch"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  parms->branchSwitch=branchSwitch;
  return;
 }

int MFHendersonsMethodGetBranchSwitch(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetBranchSwitch"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }


  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->branchSwitch;
 }

#include <MFAtlas.h>
#include <MFImplicitMF.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFPrint.h>
#include <stdio.h>
#include <MFTime.h>
void MFAtlasPageOutAllCharts(MFAtlas);

MFAtlas MFHendersonComputeAtlas(MFContinuationMethod H,MFImplicitMF M, MFNRegion Omega, int m, MFNVector *u0)
 {
  static char RoutineName[]={"MFHendersonComputeAtlas"};
  int i;
  int kkk;
  double delta=0.;
  int chart;
  int n;
  MFAtlas S;
  MFNVector u;
  FILE *fid;
  double volume;
  double epsilon;
  int kmax;
  MFNKMatrix Phi;
  double R;
  int verbose;

#ifdef MFTIMINGS
  MFTimeReset();
#endif

  epsilon=MFHendersonsMethodGetEpsilon(H);
  kmax=MFHendersonsMethodGetMaxCharts(H);
  verbose=MFHendersonsMethodGetVerbose(H);

  printf("%s\n",RoutineName);fflush(stdout);

  kkk=0;
  S=MFCreateAtlas(M);
  printf("call MFHendersonMethodCopyClipF\n");fflush(stdout);
  MFHendersonMethodCopyClipF(H,S);
  MFAtlasSetEpsilon(S,epsilon);
  for(i=0;i<m;i++)
   {
    if(verbose){printf("%d) Add initial points to Atlas ",kkk);MFPrintNVector(stdout,u0[i]);printf(" index %d, bif? %d\n",MFNVGetIndex(u0[i]),MFNVGetIndex2(u0[i]));fflush(stdout);}
    chart=MFAtlasAddChart(S,u0[i]);
    MFIMFSetStability(M,u0[i],MFAtlasChartTangentSpace(S,chart));
#ifdef MFVOLUME
    volume=MFVolumeOfAtlas(S,Omega);
    if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
    kkk++;
   }

#ifdef MFTIMINGS
  MFTimeReport(stdout);
  MFTimeReset();
#endif

  n=MFAtlasN(S);
  u=MFCreateNVector(n);
  while((kmax<0||kkk<kmax)&&(chart=MFAtlasPointOnBoundaryInsideRegion(S,Omega,u,&Phi,&delta))>-1)
   {
    if(verbose){printf("%d) Add boundary point to Atlas ",kkk);MFPrintNVector(stdout,u);printf(" index %d, bif? %d\n",MFNVGetIndex(u),MFNVGetIndex2(u));fflush(stdout);}
    R=MFIMFScale(MFAtlasMF(S),u,Phi);
    if(MFAtlasAddChartWithAll(S,u,Phi,R)>0)
     {
#ifdef MFVOLUME
      volume=MFVolumeOfAtlas(S,Omega);
      if(verbose){printf("  Volume of Atlas %lf\n",volume);fflush(stdout);}
#endif
#ifdef MFTIMINGS
      MFTimeReport(stdout);
      MFTimeReset();
#endif
      kkk++;
      if(MFHendersonsMethodGetCheckPoint(H)&&(kkk%MFHendersonsMethodGetCheckPointEvery(H))==0)
       {
        printf("Check pointing, step %d\n",kkk);fflush(stdout);
        fid=fopen("Checkpoint.atlas","w");
        MFWriteAtlas(fid,S);
        fclose(fid);
        printf("Done check pointing\n");fflush(stdout);
       }
     }

    MFFreeNVector(u);
    MFFreeNKMatrix(Phi);
    u=MFCreateNVector(n);
   }

  if(verbose&&chart<0){printf("No more points on the boundary\n");fflush(stdout);}
   else if(verbose){printf("Maximum number of charts added (%d)\n",kkk);fflush(stdout);}

/*if(chart<0)
   {
    find a singular chart instead of a boundary chart.
    branch switch.
    keep going.
   }*/

/*if(MFHendersonsMethodGetPage(H))
   {
    printf("Calling Page Out All Charts\n");fflush(stdout);
    MFAtlasPageOutAllCharts(S);
    printf("Done Page out, return)\n");fflush(stdout);
   }*/
  return S;
 }

char *MFHendersonsMethodGetFilename(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetFileName"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }
  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->fileName;
 }

void MFHendersonsMethodSetFilename(MFContinuationMethod H,char *name)
 {
  static char RoutineName[]={"MFHendersonsMethodSetFileName"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);
  free(parms->fileName);
  parms->fileName=(char*)malloc((strlen(name)+1)*sizeof(char));
  strcpy(parms->fileName,name);
  return;
 }

void MFHendersonsMethodAddClipF(MFContinuationMethod H,double (*clipF)(MFNVector))
 {
  static char RoutineName[]={"MFHendersonsMethodAddClipF"};
  MFHendersonParmBlock this;
  int verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  this=(MFHendersonParmBlock)(H->parmBlock);

  if(this->nClipF>=this->mClipF)
   {
    this->mClipF+=10;
    this->clipF=(double (**)(MFNVector))realloc(this->clipF,(this->mClipF)*sizeof(double (*)(MFNVector)));
   }
  this->clipF[this->nClipF]=clipF;
  this->nClipF++;
 }

void MFHendersonMethodCopyClipF(MFContinuationMethod H,MFAtlas A)
 {
  static char RoutineName[]={"MFHendersonsMethodCopyClipF"};
  int i;
  MFHendersonParmBlock this;
  int verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  this=(MFHendersonParmBlock)(H->parmBlock);

  MFAtlasClearClipF(A);
  for(i=0;i<this->nClipF;i++)MFAtlasAddClipF(A,this->clipF[i]);

  return;
 }

void MFContinuationMethodExtendAtlasMultiple(MFContinuationMethod this,MFAtlas A,MFImplicitMF M,MFNRegion Omega,int n,MFNVector *u0,MFNKMatrix *Phi0)
 {
  this->ExtendAtlas(this,A,M,Omega,n,u0,Phi0);
  return;
 }

void MFContinuationMethodFlushAtlas(MFContinuationMethod this,MFAtlas A)
 {
  this->FlushAtlas(this,A);
  return;
 }

void MFContinuationMethodCloseAtlas(MFContinuationMethod this,MFAtlas A)
 {
  this->CloseAtlas(this,A);
  return;
 }

int MFHendersonsMethodGetDumpToRestartFile(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetDumpToRestartFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);
  return parms->dumpToRestartFile;
 }

int MFHendersonsMethodGetDumpToRestartFileEvery(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodGetDumpToRestartFileEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);

  return parms->dumpToRestartFileEvery;
 }

void MFHendersonsMethodSetDumpToRestartFile(MFContinuationMethod H,int i)
 {
  static char RoutineName[]={"MFHendersonsMethodSetDumpToRestartFile"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);

  parms->dumpToRestartFile=i;
  return;
 }

void MFHendersonsMethodSetDumpToRestartFileEvery(MFContinuationMethod H,int i)
 {
  static char RoutineName[]={"MFHendersonsMethodSetDumpToRestartFileEvery"};
  MFHendersonParmBlock parms;

  if(H==(MFContinuationMethod)NULL)
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is NULL.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(H->type,"Henderson"))
   {
    sprintf(MFHendersonErrorMsg,"MFContinuation Method (argument 1) is not Henderson's Method.");
    MFSetError(12,RoutineName,MFHendersonErrorMsg,__LINE__,__FILE__);
    return;
   }

  parms=(MFHendersonParmBlock)(H->parmBlock);

  parms->dumpToRestartFileEvery=i;
  return;
 }


void MFHendersonsMethodClearClipF(MFContinuationMethod H)
 {
  static char RoutineName[]={"MFHendersonsMethodClearClipF"};
  MFHendersonParmBlock this;
  int i;
  int verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  this=(MFHendersonParmBlock)(H->parmBlock);

  for(i=0;i<this->nClipF;i++)this->clipF[i]=NULL;

  this->nClipF=0;

  return;
 }
