/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */

#include <MFError.h>
#include <MFKVector.h>
#include <float.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <MFTime.h>

#ifndef DBL_QNAN
#define DBL_QNAN 1.e200
#endif

void MFSetError(int,char*,char*,int,char*);
static char MFKVectorErrorMsg[256]="";

struct MFKVectorSt
 {
  int nC;
  double *data;
  int nRefs;
 };

MFKVector MFCreateKVector(int n)
 {
  static char RoutineName[]={"MFCreateKVector"};
  MFKVector this;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateKVector++;
  starttime=clock();
#endif

  if(n<1)
   {
    sprintf(MFKVectorErrorMsg,"Length of Vector %d (argument 1) is Illegal. Must be positive.",n);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVector+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

  this=(MFKVector)malloc(sizeof(struct MFKVectorSt)); /*done*/
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFKVectorSt));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVector+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }
  this->nC=n;

  this->data=(double*)malloc(n*sizeof(double)); /*done*/
  if(this->data==(double*)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVector+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }
  for(i=0;i<n;i++)this->data[i]=0.;
  this->nRefs=1;

#ifdef MFTIMINGS
    MFTimeMFCreateKVector+=clock()-starttime;
#endif
  return this;
 }

MFKVector MFCreateKVectorWithData(int n,double *vl)
 {
  static char RoutineName[]={"MFCreateKVectorWithFullData"};
  int i;
  MFKVector this;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateKVectorWithData++;
  starttime=clock();
#endif

  if(n<1)
   {
    sprintf(MFKVectorErrorMsg,"Length of Vector %d (argument 1) is Illegal. Must be positive.",n);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVectorWithData+=clock()-starttime;
#endif
    return((MFKVector)NULL);
   }

  this=(MFKVector)malloc(sizeof(struct MFKVectorSt)); /*done*/
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFKVectorSt));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVectorWithData+=clock()-starttime;
#endif
    return(this);
   }

  this->data=(double*)malloc(n*sizeof(double)); /*done*/
  if(this->data==(double*)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateKVectorWithData+=clock()-starttime;
#endif
    return(this);
   }
  if(vl==(double*)NULL)
   {
    sprintf(MFKVectorErrorMsg,"The pointer to the array of coordinates (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
    this->data=(double*)NULL;
    this->nRefs=1;
#ifdef MFTIMINGS
    MFTimeMFCreateKVectorWithData+=clock()-starttime;
#endif
    return(this);
   }
  for(i=0;i<n;i++)this->data[i]=vl[i];
  this->nC=n;

  this->nRefs=1;

#ifdef MFTIMINGS
    MFTimeMFCreateKVectorWithData+=clock()-starttime;
#endif
  return this;
 }

void MFFreeKVector(MFKVector this)
 {
  static char RoutineName[]={"MFFreeKVector"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeKVector++;
  starttime=clock();
#endif
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeKVector+=clock()-starttime;
#endif
    return;
   }

  this->nRefs--;

  if(this->nRefs<1)
   {
    if(this->data!=(double*)NULL){free(this->data);this->data=(double*)NULL;}
    free(this);
   }
#ifdef MFTIMINGS
    MFTimeMFFreeKVector+=clock()-starttime;
#endif
  return;
 }

int MFKV_NC(MFKVector this)
 {
  static char RoutineName[]={"MFKV_NC"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKV_NC++;
  starttime=clock();
#endif
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKV_NC+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
    MFTimeMFKV_NC+=clock()-starttime;
#endif
  return(this->nC);
 }

double MFKV_C(MFKVector this,int i)
 {
  static char RoutineName[]={"MFKV_C"};
  int j;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKV_C++;
  starttime=clock();
#endif

  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKV_C+=clock()-starttime;
#endif
    return DBL_QNAN;
   }
  if(i<0|| !(i<this->nC))
   {
    sprintf(MFKVectorErrorMsg,"Coordinate %d (argument 2) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKV_C+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

/*printf("\n%s (0x%8.8x,%d,%lf this->data[i]=0x%8.8x)\n",RoutineName,this,i,this->data[i],&(this->data[i]));fflush(stdout);*/
#ifdef MFTIMINGS
    MFTimeMFKV_C+=clock()-starttime;
#endif
  return this->data[i];
 }

void MFKVSetC(MFKVector this,int i,double vl)
 {
  static char RoutineName[]={"MFKVSetC"};
  int j;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVSetC++;
  starttime=clock();
#endif

  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVSetC+=clock()-starttime;
#endif
    return;
   }

  if(i<0|| !(i<this->nC))
   {
    sprintf(MFKVectorErrorMsg,"Coordinate %d (argument 2) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVSetC+=clock()-starttime;
#endif
    return;
   }

/*printf("\n%s(0x%8.8x,%d,%lf this->data[i]=0x%8.8x)\n",RoutineName,this,i,vl,&(this->data[i]));fflush(stdout);*/
  this->data[i]=vl;
#ifdef MFTIMINGS
    MFTimeMFKVSetC+=clock()-starttime;
#endif
  return;
 }

void MFRefKVector(MFKVector this)
 {
  static char RoutineName[]={"MFRefKVector"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefKVector++;
  starttime=clock();
#endif
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefKVector+=clock()-starttime;
#endif
    return;
   }

  this->nRefs++;
  return;
 }

double MFKVDot(MFKVector a,MFKVector b)
 {
  static char RoutineName[]={"MFKVDot"};
  double d;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVDot++;
  starttime=clock();
#endif

  if(a==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDot+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  if(b==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDot+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  if(a->nC!=b->nC)
   {
    sprintf(MFKVectorErrorMsg,"Vectors must be the same length a=%d, b=%d",a->nC,b->nC);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDot+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  d=0;
  for(i=0;i<a->nC;i++)d+=a->data[i]*b->data[i];

#ifdef MFTIMINGS
    MFTimeMFKVDot+=clock()-starttime;
#endif
  return d;
 }

double MFKVNorm(MFKVector a)
 {
  static char RoutineName[]={"MFKVNorm"};
  double d;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVNorm++;
  starttime=clock();
#endif

  if(a==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVNorm+=clock()-starttime;
#endif
    return -1.;
   }

  d=0;
  for(i=0;i<a->nC;i++)d+=a->data[i]*a->data[i];

#ifdef MFTIMINGS
    MFTimeMFKVNorm+=clock()-starttime;
#endif
  return sqrt(d);
 }

void MFKVScale(double s,MFKVector a)
 {
  static char RoutineName[]={"MFKVScale"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVNorm++;
  starttime=clock();
#endif

  for(i=0;i<a->nC;i++)a->data[i]*=s;

#ifdef MFTIMINGS
    MFTimeMFKVNorm+=clock()-starttime;
#endif
  return;
 }

void MFKVScaleMul(double s,MFKVector a,MFKVector b)
 {
  static char RoutineName[]={"MFKVScaleMul"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVScaleMul++;
  starttime=clock();
#endif

  if(a==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVScaleMul+=clock()-starttime;
#endif
    return;
   }

  if(b==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVScaleMul+=clock()-starttime;
#endif
    return;
   }

  if(a->nC!=b->nC)
   {
    sprintf(MFKVectorErrorMsg,"Vectors must be the same length a=%d, b=%d",a->nC,b->nC);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVScaleMul+=clock()-starttime;
#endif
    return;
   }

  for(i=0;i<a->nC;i++)b->data[i]=s*a->data[i];

#ifdef MFTIMINGS
    MFTimeMFKVScaleMul+=clock()-starttime;
#endif
  return;
 }

double *MFKV_CStar(MFKVector this)
 {
  static char RoutineName[]={"MFKV_CStar"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKV_CStar++;
  starttime=clock();
#endif
  if(this==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKV_CStar+=clock()-starttime;
#endif
    return (double*)NULL;
   }

#ifdef MFTIMINGS
    MFTimeMFKV_CStar+=clock()-starttime;
#endif
  return(this->data);
 }

void MFKVDiff(MFKVector a,MFKVector b, MFKVector c)
 {
  static char RoutineName[]={"MFKVDiff"};
  MFKVector result;
  int i,n;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVDiff++;
  starttime=clock();
#endif

  if(a==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDiff+=clock()-starttime;
#endif
    return;
   }

  if(b==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDiff+=clock()-starttime;
#endif
    return;
   }

  if(c==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDiff+=clock()-starttime;
#endif
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFKVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVDiff+=clock()-starttime;
#endif
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
    c->data[i]=a->data[i]-b->data[i];

#ifdef MFTIMINGS
    MFTimeMFKVDiff+=clock()-starttime;
#endif
  return;
 }

void MFKVAdd(MFKVector a,MFKVector b, MFKVector c)
 {
  static char RoutineName[]={"MFKVAdd"};
  MFKVector result;
  int i,n;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFKVAdd++;
  starttime=clock();
#endif

  if(a==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVAdd+=clock()-starttime;
#endif
    return;
   }

  if(b==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVAdd+=clock()-starttime;
#endif
    return;
   }

  if(c==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Vector c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVAdd+=clock()-starttime;
#endif
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFKVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFKVAdd+=clock()-starttime;
#endif
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
    c->data[i]=a->data[i]+b->data[i];

#ifdef MFTIMINGS
    MFTimeMFKVAdd+=clock()-starttime;
#endif
  return;
 }

void MFWriteKVector(FILE *fid,MFKVector s)
 {
  static char RoutineName[]={"MFWriteKVector"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteKVector++;
  starttime=clock();
#endif

  fprintf(fid,"%s\n","KVector");
  fprintf(fid,"%d %d\n",s->nC,s->nRefs);
  for(i=0;i<s->nC;i++)
   {
    if(i>0)fprintf(fid," ");
    fprintf(fid,"%lf",s->data[i]);
   }
  fprintf(fid,"\n");

#ifdef MFTIMINGS
    MFTimeMFWriteKVector+=clock()-starttime;
#endif
  return;
 }

MFKVector MFReadKVector(FILE *fid)
 {
  static char RoutineName[]={"MFReadKVector"};
  int i;
  MFKVector s;
  char tag[100]="";

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadKVector++;
  starttime=clock();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"KVector"))
   {
    sprintf(MFKVectorErrorMsg,"Next Object is not a KVector! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadKVector+=clock()-starttime;
#endif
    return;
   }

  s=(MFKVector)malloc(sizeof(struct MFKVectorSt)); /*done*/
  if(s==(MFKVector)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFKVectorSt));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadKVector+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

  fscanf(fid,"%d %d\n",&(s->nC),&(s->nRefs));

  s->data=(double*)malloc(s->nC*sizeof(double)); /*done*/
  if(s->data==(double*)NULL)
   {
    sprintf(MFKVectorErrorMsg,"Out of memory, trying to allocate %d bytes",s->nC*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFKVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFKVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadKVector+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }
  
  for(i=0;i<s->nC;i++)
   {
    if(i>0)fscanf(fid," ");
    fscanf(fid,"%lf",&(s->data[i]));
   }
  fscanf(fid,"\n");

#ifdef MFTIMINGS
    MFTimeMFReadKVector+=clock()-starttime;
#endif
  return s;
 }

int MFKVectorGetNRefs(MFKVector u)
 {
  return u->nRefs;
 }
