/* 
    @(#)MFNSpaceMF.c	1.17
    03/07/11 17:06:41
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */

#include <MFImplicitMF.h>
#include <stdlib.h>
#include <string.h>

static char MFNSpaceMFErrorMsg[256]="";

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFSetError(int,char*,char*,int,char*);

void MFFreeNSpaceData(void*);
int MFProjectNSpace(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentNSpace(int,int,MFNVector,MFNKMatrix,void*);
double MFScaleNSpace(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteNSpaceData(FILE*,void*);
MFImplicitMF MFReadNSpaceMF(FILE*);

int MFNSpaceMFProjectToSave(MFNVector,double*,void*);
int MFNSpaceMFProjectToDraw(MFNVector,double*,void*);
int MFNSpaceMFProjectForBB(MFNVector,double*,void*);

struct MFIMFNSpaceData
 {
  int n;
  double R;
 };

MFImplicitMF MFIMFCreateNSpaceWithRadius(int n, double R)
 {
  static char RoutineName[]={"MFIMFCreateNSpaceWithRadius"};
  MFImplicitMF nSpace;
  struct MFIMFNSpaceData *data;
  MFNSpace space;

  nSpace=MFIMFCreateBaseClass(n,n,"NSpaceMF");

  space=MFCreateNSpace(n);
  MFIMFSetSpace(nSpace,space);
  MFFreeNSpace(space);

  MFIMFSetFreeData(nSpace,MFFreeNSpaceData);
  MFIMFSetProject(nSpace,MFProjectNSpace);
  MFIMFSetTangent(nSpace,MFTangentNSpace);
  MFIMFSetScale(nSpace,MFScaleNSpace);
  MFIMFSetR(nSpace,R);
  MFIMFSetWriteData(nSpace,MFWriteNSpaceData);
  MFIMFSetProjectForSave(nSpace,MFNSpaceMFProjectToSave);
  MFIMFSetProjectForDraw(nSpace,MFNSpaceMFProjectToDraw);
  MFIMFSetProjectForBB(nSpace,MFNSpaceMFProjectForBB);

  data=(struct MFIMFNSpaceData*)malloc(sizeof(struct MFIMFNSpaceData));
  if(data==(struct MFIMFNSpaceData*)NULL)
   {
    sprintf(MFNSpaceMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFIMFNSpaceData));
    printf("%s -- %s\n",RoutineName,MFNSpaceMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }
  data->n=n;
  data->R=R;
  MFIMFSetData(nSpace,(void*)data);

  return nSpace;
 }

MFImplicitMF MFIMFCreateNSpace(int n)
 {
  static char RoutineName[]={"MFIMFCreateNSpace"};
  MFImplicitMF nSpace;

  nSpace=MFIMFCreateNSpaceWithRadius(n,-1.);

  return nSpace;
 }

void MFFreeNSpaceData(void *d)
 {
  static char RoutineName[]={"MFFreeNSpaceData"};
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;
  if(data!=(struct MFIMFNSpaceData*)NULL)free(data);
  return;
 }

int MFProjectNSpace(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectNSpace"};
  static int i;
  double *u,*u0,*Phi;
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  for(i=0;i<n;i++)u[i]=u0[i];
  *index=0;
  return 1;
 }

int MFTangentNSpace(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentNSpace"};
  static int i;
  double *u,*Phi;
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  for(i=0;i<n*n;i++)Phi[i]=0.;
  for(i=0;i<n;i++)Phi[i*(n+1)]=1.;

  return 1;
 }

double MFScaleNSpace(int n,int k,MFNVector u,MFNKMatrix Phi,void *d)
 {
  static char RoutineName[]={"MFScaleNSpace"};
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;
  return data->R;
 }

void MFWriteNSpaceData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWriteNSpaceData"};
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  fprintf(fid,"%d %lf\n",data->n,data->R);
  return;
 }

MFImplicitMF MFReadNSpaceMF(FILE *fid)
 {
  static char RoutineName[]={"MFReadNSpaceMF"};
  int n=0;
  double R=0.;
  MFImplicitMF nSpace;

  fscanf(fid,"%d %lf\n",&n,&R);
  nSpace=MFIMFCreateNSpaceWithRadius(n,R);

  return nSpace;
 }

int MFNSpaceMFProjectToSave(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFNSpaceMFProjectToDraw(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFNSpaceMFProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFIMFNSpaceData *data;

  data=(struct MFIMFNSpaceData*)d;

  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }
