/* 
    @(#)MFNVector.c	1.16
    03/07/10 16:58:22
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              April 29, 2002     Made into a base class */

#include <MFError.h>
#include <MFNVector.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>

#ifndef DBL_QNAN
#define DBL_QNAN 1.e200
#endif

void MFSetError(int,char*,char*,int,char*);
static char MFNVectorErrorMsg[256]="";

MFNVector MFReadDenseNVector(FILE*);
#ifdef AUTO
MFNVector MFReadAUTOBVNVector(FILE*);
#endif

struct MFNVectorSt
 {
  char *type;

  void *data;
  void (*freedata)(void*);
  void (*writedata)(FILE*,void*);

  MFNVector (*clone)(void*);
  int (*getNC)(void*);
  double (*getC)(int,void*);
  void (*setC)(int,double,void*);
  void (*add)(void*,void*,void*);
  void (*diff)(void*,void*,void*);
  void (*print)(FILE*,void*);

  int index;
  int index2;
  int nRefs;
 };

MFNVector MFCreateNVectorBaseClass(char *type)
 {
  static char RoutineName[]={"MFCreateNVectorBaseClass"};
  MFNVector this;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateNVector++;
  starttime=clock();
#endif

  this=(MFNVector)malloc(sizeof(struct MFNVectorSt)); /*done*/
  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNVectorSt));
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNVector+=clock()-starttime;
#endif
    return(this);
   }

  this->data=NULL;
  this->freedata=NULL;
  this->writedata=NULL;
  this->clone=NULL;
  this->getNC=NULL;
  this->getC=NULL;
  this->setC=NULL;
  this->add=NULL;
  this->diff=NULL;
  this->print=NULL;

  this->nRefs=1;

  this->index=0;
  this->index2=0;

  if(type!=(char*)NULL)
   {
    this->type=(char*)malloc((strlen(type)+1)*sizeof(char));
    if(this->type==(char*)NULL)
     {
      sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(type)+1)*sizeof(char));
      printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
      return;
     }
    strcpy(this->type,type);
   }else
   this->type=(char*)NULL;

#ifdef MFTIMINGS
    MFTimeMFCreateNVector+=clock()-starttime;
#endif
  return this;
 }

void MFFreeNVector(MFNVector this)
 {
  static char RoutineName[]={"MFFreeNVector"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeNVector++;
  starttime=clock();
#endif
  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeNVector+=clock()-starttime;
#endif
    return;
   }

  this->nRefs--;

  if(this->nRefs<1)
   {
    if(this==(MFNVector)0x300fc908)printf("Free of vector 0x300fc908\n");fflush(stdout);
    if(this->type!=(char*)NULL)free(this->type);
    if(this->freedata!=NULL&&this->data!=(double*)NULL)(this->freedata)(this->data);
    free(this);
   }
#ifdef MFTIMINGS
    MFTimeMFFreeNVector+=clock()-starttime;
#endif
  return;
 }

int MFNV_NC(MFNVector this)
 {
  static char RoutineName[]={"MFNV_NC"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNV_NC++;
  starttime=clock();
#endif
  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNV_NC+=clock()-starttime;
#endif
    return -1;
   }

  if(this->getNC==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to get the number of coordinates has not been provided");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  result=(this->getNC)(this->data);

#ifdef MFTIMINGS
    MFTimeMFNV_NC+=clock()-starttime;
#endif
  return result;
 }

double MFNV_C(MFNVector this,int i)
 {
  static char RoutineName[]={"MFNV_C"};
  int j;
  double result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNV_C++;
  starttime=clock();
#endif

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNV_C+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  if(this->getC==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to get a coordinate (vector type %s) has not been provided",this->type);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return DBL_QNAN;
   }

  result=(this->getC)(i,this->data);

#ifdef MFTIMINGS
    MFTimeMFNV_C+=clock()-starttime;
#endif
  return result;
 }

void MFNVSetC(MFNVector this,int i,double vl)
 {
  static char RoutineName[]={"MFNVSetC"};
  int rc;
  int j;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNVSetC++;
  starttime=clock();
#endif

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVSetC+=clock()-starttime;
#endif
    return;
   }

  if(this->setC==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to set a coordinate (vector type %s) has not been provided",this->type);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  (this->setC)(i,vl,this->data);

#ifdef MFTIMINGS
    MFTimeMFNVSetC+=clock()-starttime;
#endif
  return;
 }

void MFPrintNVector(FILE *fid, MFNVector this)
 {
  static char RoutineName[]={"MFPrintNVector"};
  int j;
  double result;

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(this->print==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to print an NVector of type %s has not been provided",this->print);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  (this->print)(fid,this->data);

  return;
 }

void MFRefNVector(MFNVector this)
 {
  static char RoutineName[]={"MFRefNVector"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefNVector++;
  starttime=clock();
#endif
  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefNVector+=clock()-starttime;
#endif
    return;
   }

  this->nRefs++;
#ifdef MFTIMINGS
    MFTimeMFRefNVector+=clock()-starttime;
#endif
  return;
 }

void MFNVDiff(MFNVector a,MFNVector b, MFNVector c)
 {
  static char RoutineName[]={"MFNVDiff"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNVDiff++;
  starttime=clock();
#endif

  if(a==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
    return;
   }

  if(b==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
    return;
   }

  if(c==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
    return;
   }

  if(a->diff==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to take the difference of two vectors not been provided");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(a->type,b->type))
   {
    sprintf(MFNVectorErrorMsg,"Can't take the difference of vectors of different type!");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(b->type,c->type))
   {
    sprintf(MFNVectorErrorMsg,"The difference of two vectors must be the same type as the vectors!");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  (a->diff)(a->data,b->data,c->data);

#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
  return;
 }

void MFNVAdd(MFNVector a,MFNVector b, MFNVector c)
 {
  static char RoutineName[]={"MFNVAdd"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNVAdd++;
  starttime=clock();
#endif

  if(a==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVAdd+=clock()-starttime;
#endif
    return;
   }

  if(b==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVAdd+=clock()-starttime;
#endif
    return;
   }

  if(c==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Vector c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVAdd+=clock()-starttime;
#endif
    return;
   }

  if(a->add==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Routine to add two vectors not been provided");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(a->type,b->type))
   {
    sprintf(MFNVectorErrorMsg,"Can't add vectors of different type!");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(strcmp(b->type,c->type))
   {
    sprintf(MFNVectorErrorMsg,"The sum of two vectors must be the same type as the vectors!");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  (a->add)(a->data,b->data,c->data);

#ifdef MFTIMINGS
    MFTimeMFNVAdd+=clock()-starttime;
#endif
  return;
 }

void MFWriteNVector(FILE *fid,MFNVector u)
 {
  static char RoutineName[]={"MFWriteNVector"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteNVector++;
  starttime=clock();
#endif

  fprintf(fid,"%s\n","NVector");
  if(u->type!=(char*)NULL)
    fprintf(fid,"%d %s\n",strlen(u->type),u->type);
   else
    fprintf(fid,"%d \n",0);
  if(u->writedata!=NULL) (u->writedata)(fid,u->data);
  fprintf(fid,"%d %d %d\n",u->index,u->index2,u->nRefs);

#ifdef MFTIMINGS
    MFTimeMFWriteNVector+=clock()-starttime;
#endif
  return;
 }

MFNVector MFReadNVector(FILE *fid)
 {
  static char RoutineName[]={"MFReadNVector"};
  int n;
  MFNVector u;
  char tag[100]="";

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadNVector++;
  starttime=clock();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"NVector"))
   {
    sprintf(MFNVectorErrorMsg,"Next Object is not a NVector! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNVector+=clock()-starttime;
#endif
    return;
   }

  u=MFCreateNVectorBaseClass((char*)NULL);

  fscanf(fid,"%d",&n);
  if(n>0)
   {
    fscanf(fid,"%s\n",tag);
   }else{
    sprintf(MFNVectorErrorMsg,"Unknown input Vector type, type length 0");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNVector+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

  if(!strcmp(tag,"DENSE"))u=MFReadDenseNVector(fid);
   else if(!strcmp(u->type,"LOCA"))n=n;
#ifdef AUTO
   else if(!strcmp(u->type,"AUTO"))u=MFReadAUTOBVNVector(fid);
#endif
   else{
    sprintf(MFNVectorErrorMsg,"Unknown input Vector type \"%s\"",u->type);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNVector+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

  fscanf(fid,"%d %d %d\n",&(u->index),&(u->index2),&(u->nRefs));

#ifdef MFTIMINGS
    MFTimeMFReadNVector+=clock()-starttime;
#endif
  return u;
 }

int MFNVGetIndex(MFNVector u)
 {
  static char RoutineName[]={"MFNVGetIndex"};

  if(u==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefNVector+=clock()-starttime;
#endif
    return;
   }

  return u->index;
 }

void MFNVSetIndex(MFNVector u, int index)
 {
  static char RoutineName[]={"MFNVSetIndex"};
  u->index=index;
 }

int MFNVGetIndex2(MFNVector u)
 {
  static char RoutineName[]={"MFNVGetIndex2"};

  if(u==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefNVector+=clock()-starttime;
#endif
    return;
   }

  return u->index2;
 }

void MFNVSetIndex2(MFNVector u, int index)
 {
  static char RoutineName[]={"MFNVSetIndex2"};

  if(u==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefNVector+=clock()-starttime;
#endif
    return;
   }

  u->index2=index;
 }

MFNVector MFCloneNVector(MFNVector this)
 {
  static char RoutineName[]={"MFCloneNVector"};

  if(this==(MFNVector)NULL)return (MFNVector)NULL;

  if(this->clone==NULL)
   {
    sprintf(MFNVectorErrorMsg,"Clone routine not provided");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return (MFNVector)NULL;
   }

  return (this->clone)(this->data);
 }

char *MFNVGetId(MFNVector this)
 {
  static char RoutineName[]={"MFNVGetId"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return (char*)NULL;
   }

  return(this->type);
 }

void MFNVectorSetPrint(MFNVector this, void (*print)(FILE*,void*))
 {
  static char RoutineName[]={"MFNVectorSetPrint"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->print=print;

  return;
 }

void MFNVectorSetData(MFNVector this, void *data)
 {
  static char RoutineName[]={"MFNVectorSetData"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->data=data;

  return;
 }

void *MFNVectorGetData(MFNVector this)
 {
  static char RoutineName[]={"MFNVectorGetData"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  return this->data;
 }

void MFNVectorSetFreeData(MFNVector this, void (*freedata)(void*))
 {
  static char RoutineName[]={"MFNVectorSetFreeData"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->freedata=freedata;

  return;
 }

void MFNVectorSetWriteData(MFNVector this, void (*writedata)(FILE*,void*))
 {
  static char RoutineName[]={"MFNVectorSetWriteData"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->writedata=writedata;

  return;
 }

void MFNVectorSetClone(MFNVector this, MFNVector (*clone)(void*))
 {
  static char RoutineName[]={"MFNVectorSetClone"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->clone=clone;

  return;
 }

void MFNVectorSetGetNC(MFNVector this, int (*getNC)(void*))
 {
  static char RoutineName[]={"MFNVectorSetGetNC"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->getNC=getNC;

  return;
 }

void MFNVectorSetGetC(MFNVector this, double (*getC)(int,void*))
 {
  static char RoutineName[]={"MFNVectorSetGetC"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->getC=getC;

  return;
 }

void MFNVectorSetSetC(MFNVector this, void (*setC)(int,double,void*))
 {
  static char RoutineName[]={"MFNVectorSetSetC"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->setC=setC;

  return;
 }

void MFNVectorSetAdd(MFNVector this, void (*add)(void*,void*,void*))
 {
  static char RoutineName[]={"MFNVectorSetAdd"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->add=add;

  return;
 }

void MFNVectorSetDiff(MFNVector this, void (*diff)(void*,void*,void*))
 {
  static char RoutineName[]={"MFNVectorSetDiff"};

  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->diff=diff;

  return;
 }
