/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              October 6, 2000   Added Polyhedron    */

#include <stdio.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <stdlib.h>
#include <math.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNRegionErrorMsg[256]="";

double *MFNV_CStar(MFNVector);

int CubeTest(MFNVector,void*);
void CubeFree(void*);
int RectangleTest(MFNVector,void*);
void RectangleFree(void*);
int HyperCubeTest(MFNVector,void*);
void HyperCubeFree(void*);
int HyperCubeTestByCorners(MFNVector,void*);
int DodIcosTest(MFNVector,void*);
void DodIcosFree(void*);
int MF3dPolygonTest(MFNVector,void*);
void MF3dPolygonFree(void*);
int MF3dEdgeTest(MFNVector,void*);
void MF3dEdgeFree(void*);
int MFPolyhedronTest(MFNVector,void*);
void MFPolyhedronFree(void*);

MFNRegion MFNRegionCreateCube(double x0,double y0,double z0,double x1,double y1,double z1)
 {
  static char RoutineName[]={"MFNRegionCreateCube"};
  MFNRegion cube;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateCube++;
  starttime=clock();
#endif

  cube=MFNRegionCreateBaseClass("Cube");
  MFNRegionSetTest(cube,CubeTest);
  MFNRegionSetFreeData(cube,CubeFree);

  data=(void*)malloc(6*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",6*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(cube);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateCube+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=x0;
  data[1]=y0;
  data[2]=z0;
  data[3]=x1;
  data[4]=y1;
  data[5]=z1;
  MFNRegionSetData(cube,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateCube+=clock()-starttime;
#endif
  return(cube);
 }

int CubeTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"CubeTest"};
  double *d;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledCubeTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  d=(double*)data;
  if(x[0]<d[0])rc=0;
   else if(x[1]<d[1])rc=0;
   else if(x[2]<d[2])rc=0;
   else if(x[0]>d[3])rc=0;
   else if(x[1]>d[4])rc=0;
   else if(x[2]>d[5])rc=0;
   else rc=1;

#ifdef MFTIMINGS
    MFTimeCubeTest+=clock()-starttime;
#endif
  return rc;
 }

void CubeFree(void *data)
 {
  static char RoutineName[]={"CubeFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledCubeFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimeCubeFree+=clock()-starttime;
#endif
  return;
 }

MFNRegion MFNRegionCreateRectangle(double x0,double y0,double x1,double y1)
 {
  static char RoutineName[]={"MFNRegionCreateRectangle"};
  MFNRegion rectangle;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateRectangle++;
  starttime=clock();
#endif

  rectangle=MFNRegionCreateBaseClass("Rectangle");
  MFNRegionSetTest(rectangle,RectangleTest);
  MFNRegionSetFreeData(rectangle,RectangleFree);

  data=(void*)malloc(4*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",4*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(rectangle);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateRectangle+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=x0;
  data[1]=y0;
  data[2]=x1;
  data[3]=y1;
  MFNRegionSetData(rectangle,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateRectangle+=clock()-starttime;
#endif
  return(rectangle);
 }

int RectangleTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"RectangleTest"};
  double *d;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledRectangleTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  d=(double*)data;
  if(x[0]<d[0])rc=0;
   else if(x[1]<d[1])rc=0;
   else if(x[0]>d[2])rc=0;
   else if(x[1]>d[3])rc=0;
   else rc=1;

#ifdef MFTIMINGS
    MFTimeRectangleTest+=clock()-starttime;
#endif
  return rc;
 }

void RectangleFree(void *data)
 {
  static char RoutineName[]={"RectangleFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledRectangleFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimeRectangleFree+=clock()-starttime;
#endif
  return;
 }

MFNRegion MFNRegionCreateHyperCube(int n, double R)
 {
  static char RoutineName[]={"MFNRegionCreateHyperCube"};
  MFNRegion hypercube;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateHyperCube++;
  starttime=clock();
#endif

  hypercube=MFNRegionCreateBaseClass("HyperCube");
  MFNRegionSetTest(hypercube,HyperCubeTest);
  MFNRegionSetFreeData(hypercube,HyperCubeFree);

  data=(void*)malloc(sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(hypercube);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateHyperCube+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=R;

  MFNRegionSetData(hypercube,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateHyperCube+=clock()-starttime;
#endif
  return(hypercube);
 }

MFNRegion MFNRegionCreateHyperCubeByCorners(int n, MFNVector LL, MFNVector UR)
 {
  static char RoutineName[]={"MFNRegionCreateHyperCubeByCorners"};
  MFNRegion hypercube;
  int i;
  double l,r;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateHyperCube++;
  starttime=clock();
#endif

  hypercube=MFNRegionCreateBaseClass("HyperCubeByCorners");
  MFNRegionSetTest(hypercube,HyperCubeTestByCorners);
  MFNRegionSetFreeData(hypercube,HyperCubeFree);

  data=(void*)malloc((2*n+1)*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(hypercube);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateHyperCube+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=n;
  for(i=0;i<n;i++)
   {
    l=MFNV_C(LL,i);
    r=MFNV_C(UR,i);
    if(l>r)
     {
      l=MFNV_C(UR,i);
      r=MFNV_C(LL,i);
     }
    data[2*i+1]=l;
    data[2*i+2]=r;
/*  printf("Direction %d range [%lf,%lf]\n",i,l,r);fflush(stdout);*/
   }

  MFNRegionSetData(hypercube,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateHyperCube+=clock()-starttime;
#endif
  return(hypercube);
 }

int HyperCubeTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"HyperCubeTest"};
  double R;
  int i;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledHyperCubeTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  R=((double*)data)[0];
  if(!strcmp(MFNVGetId(v),"DENSE"))
   {
    x=MFNV_CStar(v);
    rc=1;
    for(i=0;i<n;i++)
     {
      if(x[i]>R)rc=0;
      if(x[i]<-R)rc=0;
     }
   }else{
    rc=1;
    for(i=0;i<n&&rc;i++)
     {
      if(MFNV_C(v,i)>R)rc=0;
      if(MFNV_C(v,i)<-R)rc=0;
     }
   }

#ifdef MFTIMINGS
    MFTimeHyperCubeTest+=clock()-starttime;
#endif
  return rc;
 }

int HyperCubeTestByCorners(MFNVector v,void *data)
 {
  static char RoutineName[]={"HyperCubeTestByCorners"};
  double *c;
  int i;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledHyperCubeTest++;
  starttime=clock();
#endif
  c=(double*)data;
  n=c[0];
  if(!strcmp(MFNVGetId(v),"DENSE"))
   {
    x=MFNV_CStar(v);
    rc=1;
    for(i=0;i<n&&rc;i++)
     {
      if(x[i]<c[2*i+1])rc=0;
      if(x[i]>c[2*i+2])rc=0;
     }
   }else{
    rc=1;
    for(i=0;i<n&&rc;i++)
     {
      if(MFNV_C(v,i)<c[2*i+1])rc=0;
      if(MFNV_C(v,i)>c[2*i+2])rc=0;
     }
   }

#ifdef MFTIMINGS
    MFTimeHyperCubeTest+=clock()-starttime;
#endif
  return rc;
 }

void HyperCubeFree(void *data)
 {
  static char RoutineName[]={"HyperCubeFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledHyperCubeFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimeHyperCubeFree+=clock()-starttime;
#endif
  return;
 }

MFNRegion MFNRegionCreateDodecahedronMinusIcosahedron()
 {
  static char RoutineName[]={"MFNRegionCreateDodecahedronMinusIcosahedron"};
  MFNRegion diff;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateDodecahedronMinusIcosahedron++;
  starttime=clock();
#endif

  diff=MFNRegionCreateBaseClass("DodecahedronMinusIcosahedron");
  MFNRegionSetTest(diff,DodIcosTest);
  MFNRegionSetFreeData(diff,DodIcosFree);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateDodecahedronMinusIcosahedron+=clock()-starttime;
#endif
  return(diff);
 }

int DodIcosTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"DodIcosTest"};
#include <Dodecahedron.h>
#include <Icosahedron.h>
  int i;
  double nx,ny,nz;
  double ox,oy,oz;
  double ax,ay,az;
  double bx,by,bz;
  double t;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledDodIcosTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  for(i=0;i<DODnF;i++)
   {
    ox=DODx[3*(DODf[DODnFV*i]-1)  ];
    oy=DODx[3*(DODf[DODnFV*i]-1)+1];
    oz=DODx[3*(DODf[DODnFV*i]-1)+2];
    ax=DODx[3*(DODf[DODnFV*i+1]-1)  ]-ox;
    ay=DODx[3*(DODf[DODnFV*i+1]-1)+1]-oy;
    az=DODx[3*(DODf[DODnFV*i+1]-1)+2]-oz;
    bx=DODx[3*(DODf[DODnFV*i+3]-1)  ]-ox;
    by=DODx[3*(DODf[DODnFV*i+3]-1)+1]-oy;
    bz=DODx[3*(DODf[DODnFV*i+3]-1)+2]-oz;
    nx=ay*bz-az*by;
    ny=az*bx-ax*bz;
    nz=ax*by-ay*bx;
    t=-ox*nx-oy*ny-oz*nz;
    if(t*((x[0]-ox)*nx+(x[1]-oy)*ny+(x[2]-oz)*nz)>0)
     {
#ifdef MFTIMINGS
      MFTimeDodIcosTest+=clock()-starttime;
#endif
      return 0;
     }
   }
  for(i=0;i<ICOSnF;i++)
   {
    ox=ICOSx[3*(ICOSf[ICOSnFV*i]-1)  ];
    oy=ICOSx[3*(ICOSf[ICOSnFV*i]-1)+1];
    oz=ICOSx[3*(ICOSf[ICOSnFV*i]-1)+2];
    ax=ICOSx[3*(ICOSf[ICOSnFV*i+1]-1)  ]-ox;
    ay=ICOSx[3*(ICOSf[ICOSnFV*i+1]-1)+1]-oy;
    az=ICOSx[3*(ICOSf[ICOSnFV*i+1]-1)+2]-oz;
    bx=ICOSx[3*(ICOSf[ICOSnFV*i+3]-1)  ]-ox;
    by=ICOSx[3*(ICOSf[ICOSnFV*i+3]-1)+1]-oy;
    bz=ICOSx[3*(ICOSf[ICOSnFV*i+3]-1)+2]-oz;
    nx=ay*bz-az*by;
    ny=az*bx-ax*bz;
    nz=ax*by-ay*bx;
    t=-ox*nx-oy*ny-oz*nz;
    if(t*((x[0]-ox)*nx+(x[1]-oy)*ny+(x[2]-oz)*nz)<0)
     {
#ifdef MFTIMINGS
      MFTimeDodIcosTest+=clock()-starttime;
#endif
      return 0;
     }
   }
#ifdef MFTIMINGS
    MFTimeDodIcosTest+=clock()-starttime;
#endif
  return 1;
 }

void DodIcosFree(void *data)
 {
  static char RoutineName[]={"DodIcosFree"};
  return;
 }

struct MF3dPolygon
  {
   int n;
   double *v;
   double x0,y0,z0;
   double a0,b0,c0;
   double a1,b1,c1;
  };

MFNRegion MFNRegionCreatePolygonal3dRegion(int n, double *v)
 {
  static char RoutineName[]={"MFNRegionCreatePolygonal3dRegion"};
  MFNRegion polygon;
  struct MF3dPolygon *p;
  int i;
  double d;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreatePolygonal3dRegion++;
  starttime=clock();
#endif

  polygon=MFNRegionCreateBaseClass("3dPolygon");
  MFNRegionSetTest(polygon,MF3dPolygonTest);
  MFNRegionSetFreeData(polygon,MF3dPolygonFree);

  p=(struct MF3dPolygon*)malloc(sizeof(struct MF3dPolygon)); /*done*/
  if(p==(struct MF3dPolygon*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MF3dPolygon));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolygonal3dRegion+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }
  p->n=n;
  p->v=(double*)malloc(3*(n+1)*sizeof(double)); /*done*/
  if(p->v==(double*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",3*(n+1)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolygonal3dRegion+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }
  for(i=0;i<3*n;i++)(p->v)[i]=v[i];
  (p->v)[3*n]=0.;
  (p->v)[3*n+1]=0.;
  (p->v)[3*n+2]=0.;
  p->x0=0.;
  p->y0=0.;
  p->z0=0.;
  for(i=0;i<n;i++)
   {
    (p->v)[3*n]+=v[3*i];
    (p->v)[3*n+1]+=v[3*i+1];
    (p->v)[3*n+2]+=v[3*i+2];
    p->x0+=v[3*i];
    p->y0+=v[3*i+1];
    p->z0+=v[3*i+2];
   }
  (p->v)[3*n]/=n;
  (p->v)[3*n+1]/=n;
  (p->v)[3*n+2]/=n;
  p->x0/=n;
  p->y0/=n;
  p->z0/=n;

  p->a0=v[0]-p->x0;
  p->b0=v[1]-p->y0;
  p->c0=v[2]-p->z0;
  d=1./sqrt(p->a0*p->a0+p->b0*p->b0+p->c0*p->c0);
  p->a0*=d;
  p->b0*=d;
  p->c0*=d;

  p->a1=v[3]-p->x0;
  p->b1=v[4]-p->y0;
  p->c1=v[5]-p->z0;

  d=p->a1*p->a0+p->b1*p->b0+p->c1*p->c0;
  p->a1-=d*p->a0;
  p->b1-=d*p->b0;
  p->c1-=d*p->c0;
  d=1/sqrt(p->a1*p->a1+p->b1*p->b1+p->c1*p->c1);
  p->a1*=d;
  p->b1*=d;
  p->c1*=d;

  MFNRegionSetData(polygon,(void*)p);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolygonal3dRegion+=clock()-starttime;
#endif
  return(polygon);
 }

int MF3dPolygonTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"MF3dPolygonTest"};
  int i,j;
  double ax,ay,az;
  double bx,by,bz;
  double sum;
  double rx,ry,rz;
  struct MF3dPolygon *p;
  double a,b;
  double alpha,beta;
  double n1,n2;
  double aBar,bBar,cBar;
  double integralTmp,q,fluxX,fluxY;
  double x1,y1,x2,y2,xp,yp;
  double exact;
  int n;
  double *x;
  int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMF3dPolygonTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  p=(struct MF3dPolygon*)data;
  if(verbose){printf("%s (%lf,%lf,%lf)\n",RoutineName,x[0],x[1],x[2]);fflush(stdout);}

  xp=(x[0]-p->x0)*p->a0+(x[1]-p->y0)*p->b0+(x[2]-p->z0)*p->c0;
  yp=(x[0]-p->x0)*p->a1+(x[1]-p->y0)*p->b1+(x[2]-p->z0)*p->c1;
  if(verbose)
   {
    printf("         (%lf,%lf)\n",xp,yp);fflush(stdout);
    printf("         n=%d\n",p->n);fflush(stdout);
   }

  sum=0.;
  for(i=0;i<p->n;i++)
   {
    ax=(p->v)[3*i  ];
    ay=(p->v)[3*i+1];
    az=(p->v)[3*i+2];
    j=i+1;if(j>p->n-1)j=0;
    bx=(p->v)[3*j  ];
    by=(p->v)[3*j+1];
    bz=(p->v)[3*j+2];

    x1=(ax-p->x0)*p->a0+(ay-p->y0)*p->b0+(az-p->z0)*p->c0;
    y1=(ax-p->x0)*p->a1+(ay-p->y0)*p->b1+(az-p->z0)*p->c1;
    x2=(bx-p->x0)*p->a0+(by-p->y0)*p->b0+(bz-p->z0)*p->c0;
    y2=(bx-p->x0)*p->a1+(by-p->y0)*p->b1+(bz-p->z0)*p->c1;

    a=(x1-xp);
    b=(y1-yp);
    alpha=(x2-x1);
    beta =(y2-y1);
    n1=beta;
    n2=-alpha;

    aBar=a*a+b*b;
    bBar=2.*(a*alpha+b*beta);
    cBar=alpha*alpha+beta*beta;
    q=4.*aBar*cBar-bBar*bBar;
    if(b*alpha==a*beta && bBar>=0 && bBar<=aBar)return 1;
    if(verbose){printf("    edge %d (%lf,%lf)<->%d (%lf,%lf), q=%le\n",i,x1,y1,j,x2,y2,q);fflush(stdout);}

    if(q!=0.)
     {
      integralTmp=2./sqrt(q)*(atan2(2.*cBar+bBar,sqrt(q))
                             -atan2(bBar,sqrt(q)));
     }else{
         integralTmp=2.*(1./bBar-1./(2*cBar+bBar));
     }
    exact=a*integralTmp+.5*alpha/cBar*
          (log(fabs(aBar+bBar+cBar))-log(aBar))
        -.5*alpha*bBar/cBar*integralTmp;
    fluxX=n1*exact;
    exact=b*integralTmp+.5*beta/cBar*
          (log(fabs(aBar+bBar+cBar))-log(aBar))
        -.5*beta*bBar/cBar*integralTmp;
    fluxY=n2*exact;
    sum+=fluxX+fluxY;
   }

#ifdef MFTIMINGS
    MFTimeMF3dPolygonTest+=clock()-starttime;
#endif
  if(verbose){printf("         sum=%le, result=%d\n",sum,fabs(sum)>1.5*3.1415926);fflush(stdout);}
  return fabs(sum)>1.5*3.1415926;
 }

void MF3dPolygonFree(void *data)
 {
  static char RoutineName[]={"MF3dPolygonFree"};
  struct MF3dPolygon *p;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMF3dPolygonFree++;
  starttime=clock();
#endif

  p=(struct MF3dPolygon*)data;

  free(p->v);
  free(p);
#ifdef MFTIMINGS
  MFTimeMF3dPolygonFree+=clock()-starttime;
#endif
  return;
 }

struct MFPolyhedron
  {
   int nv;
   double *v;
   int nt;
   int *t;
  };

#ifdef TRIANGULATED
MFNRegion MFNRegionCreatePolyhedral3dRegion(int nv, double *v, int nf, int *nfv, int **fv)
 {
  static char RoutineName[]={"MFNRegionCreatePolyhedral3dRegion"};
  MFNRegion polyhedron;
  struct MFPolyhedron *p;
  int i,n;
  int j;
  double d;
  int *tmp;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreatePolyhedral3dRegion++;
  starttime=clock();
#endif

  polyhedron=MFNRegionCreateBaseClass("Polyhedron");
  MFNRegionSetTest(polyhedron,MFPolyhedronTest);
  MFNRegionSetFreeData(polyhedron,MFPolyhedronFree);

  p=(struct MFPolyhedron*)malloc(sizeof(struct MFPolyhedron)); /*done*/
  if(p==(struct MFPolyhedron*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFPolyhedron));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolyhedral3dRegion+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }
  p->nv=nv;
  p->v=(double*)malloc(3*p->nv*sizeof(double)); /*done*/
  if(p->v==(double*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",3*n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolyhedral3dRegion+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  for(i=0;i<3*nv;i++)(p->v)[i]=v[i];

/* Break faces up into triangles */

  tmp=(int*)NULL;
  p->t=(int*)NULL;
  for(i=0;i<nf;i++)
   {
    printf("Breaking up face %d, which has %d vertices",i,nfv[i]);fflush(stdout);
    printf("    [");
    for(j=0;j<nfv[i];j++)
     {
      if(j>0)printf(",");
      printf("%d",(fv[i])[j]);
     }
    printf("]\n");fflush(stdout);
    n=MFCreateTriangulatedPolygon(nv,v,nfv[i],fv[i],&tmp);
    p->t=(int*)realloc((void*)(p->t),3*(p->nt+n)*sizeof(int));
    for(j=0;j<3*n;j++)p->t[3*p->nt+j]=tmp[j];
    p->nt+=n;
   }
  free(tmp);

  MFNRegionSetData(polyhedron,(void*)p);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePolyhedral3dRegion+=clock()-starttime;
#endif
  return(polyhedron);
 }
#endif

int MFPolyhedronTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"MFPolyhedronTest"};
  int i,j;
  double sum,flux,vol;
  double r1,r1x,r1y,r1z;
  double r2,r2x,r2y,r2z;
  double r3,r3x,r3y,r3z;
  struct MFPolyhedron *p;
  double a,b;
  double area;
  int n;
  double *x;
  int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFPolyhedronTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  p=(struct MFPolyhedron*)data;
  if(verbose){printf("%s (%lf,%lf,%lf)\n",RoutineName,x[0],x[1],x[2]);fflush(stdout);}

  sum=0.;
  for(i=0;i<p->nt;i++)
   {
    r1x=p->v[3*(p->t[3*i  ])  ] -x[0];
    r1y=p->v[3*(p->t[3*i  ])+1] -x[1];
    r1z=p->v[3*(p->t[3*i  ])+2] -x[2];
    r1=sqrt(r1x*r1x+r1y*r1y+r1z*r1z);
    r2x=p->v[3*(p->t[3*i+1])  ] -x[0];
    r2y=p->v[3*(p->t[3*i+1])+1] -x[1];
    r2z=p->v[3*(p->t[3*i+1])+2] -x[2];
    r2=sqrt(r2x*r2x+r2y*r2y+r2z*r2z);
    r2x=p->v[3*(p->t[3*i+2])  ] -x[0];
    r2y=p->v[3*(p->t[3*i+2])+1] -x[1];
    r2z=p->v[3*(p->t[3*i+2])+2] -x[2];
    r3=sqrt(r3x*r3x+r3y*r3y+r3z*r3z);
    flux=(r1*r2*r3+r1*(r2x*r3x+r2y*r3y+r2z*r3z)
                  +r2*(r3x*r1x+r3y*r1y+r3z*r1z)
                  +r3*(r1x*r2x+r1y*r2y+r1z*r2z))/
         sqrt(2*(r2*r3+r2x*r3x+r2y*r3y+r2z*r3z)
               *(r3*r1+r3x*r1x+r3y*r1y+r3z*r1z)
               *(r1*r2+r1x*r2x+r1y*r2y+r1z*r2z));
    vol=r1x*(r2y*r3z-r2z*r3y)
       +r1y*(r2z*r3x-r2x*r3z)
       +r1z*(r2x*r3y-r2y*r3x);
    flux=2*acos(flux);
    if(vol<0.)flux=-flux;
    sum+=flux;
   }

#ifdef MFTIMINGS
    MFTimeMFPolyhedronTest+=clock()-starttime;
#endif
  if(verbose){printf("         sum=%le, result=%d\n",sum,fabs(sum)>1.5*3.1415926);fflush(stdout);}
  return fabs(sum)>2.*3.1415926;
 }

void MFPolyhedronFree(void *data)
 {
  static char RoutineName[]={"MFPolyhedronFree"};
  struct MFPolyhedron *p;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFPolyhedronFree++;
  starttime=clock();
#endif

  p=(struct MFPolyhedron*)data;

  free(p->v);
  free(p);
#ifdef MFTIMINGS
  MFTimeMFPolyhedronFree+=clock()-starttime;
#endif
  return;
 }

struct MF3dEdge
  {
   double x0,y0,z0;
   double a,b,c;
   double d;
  };

MFNRegion MFNRegionCreateEdge3dRegion(double *v0, double *v1)
 {
  static char RoutineName[]={"MFNRegionCreateEdge3dRegion"};
  MFNRegion edge;
  struct MF3dEdge *p;
  int i;
  double d;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateEdge3dRegion++;
  starttime=clock();
#endif

  edge=MFNRegionCreateBaseClass("3dEdge");
  MFNRegionSetTest(edge,MF3dEdgeTest);
  MFNRegionSetFreeData(edge,MF3dEdgeFree);

  p=(struct MF3dEdge*)malloc(sizeof(struct MF3dEdge)); /*done*/
  if(p==(struct MF3dEdge*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MF3dEdge));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateEdge3dRegion+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }
  p->x0=v0[0];
  p->y0=v0[1];
  p->z0=v0[2];
  p->a=v1[0]-v0[0];
  p->b=v1[1]-v0[1];
  p->c=v1[2]-v0[2];
  p->d=p->a*p->a+p->b*p->b+p->c*p->c;

  MFNRegionSetData(edge,(void*)p);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateEdge3dRegion+=clock()-starttime;
#endif
  return(edge);
 }

int MF3dEdgeTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"MF3dEdgeTest"};
  int i,j;
  double s;
  int verbose=0;
  int n;
  double *x;
  struct MF3dEdge *p;

  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  p=(struct MF3dEdge*)data;

  s=((x[0]-p->x0)*p->a+(x[1]-p->y0)*p->b+(x[2]-p->z0)*p->c)/p->d;

  return s*(1-s)>-1.e-7;
 }

void MF3dEdgeFree(void *data)
 {
  static char RoutineName[]={"MF3dEdgeFree"};
  struct MF3dEdge *p;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMF3dEdgeFree++;
  starttime=clock();
#endif

  p=(struct MF3dEdge*)data;

  free(p);
#ifdef MFTIMINGS
  MFTimeMF3dEdgeFree+=clock()-starttime;
#endif
  return;
 }
