/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFPrint.h>
#include <MFError.h>
#include <MFKVector.h>
#include <MFAtlasFriends.h>

#include <MFEnumPolytope.h>
#include <MFEnumDualPolytope.h>

void MFSetError(int,char*,char*,int,char*);
static char MFPrintErrorMsg[256]="";

#define DENSE 0
#define VECTOR 1
#define LOCA 2

void MFPrintEnumPolytope(FILE *fout,MFEnumPolytope EP)
 {
  static char RoutineName[]={"MFPrintEnumPolytope"};
  int k,d;
  int i,n;
  int j,m;
  MFKVector v;

  if(fout==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(EP==(MFEnumPolytope)NULL)
   {
    sprintf(MFPrintErrorMsg,"Enumerated Polytope (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  k=MFEnumPolytopeDimension(EP);
  for(d=0;d<k+1;d++)
   {
    fprintf(fout," %d cells\n",d);
    n=MFEnumPolytopeNumberOfCells(EP,d);
    for(i=0;i<n;i++)
     {
      fprintf(fout,"  %2d  ",i);fflush(fout);
      fprintf(fout,"[");fflush(fout);
      m=MFEnumPolytopeNumberOfCellIndices(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout," ");
        fprintf(fout,"%2d",MFEnumPolytopeCellIndex(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,"]   (");
      m=MFEnumPolytopeNumberOfCellFaces(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout,",");
        fprintf(fout,"%2d",MFEnumPolytopeCellFace(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,") (");fflush(fout);
      m=MFEnumPolytopeNumberOfFaceCells(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout,",");
        fprintf(fout,"%2d",MFEnumPolytopeFaceCell(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,") ");fflush(fout);
      v=MFEnumPolytopeVertex(EP,i);
      if(d==0&&v!=(MFKVector)NULL)MFPrintKVector(stdout,v);
      fprintf(fout,"\n");fflush(fout);
     }
   }

  return;
 }

void MFPrintEnumDualPolytope(FILE *fout,MFEnumDualPolytope EP)
 {
  static char RoutineName[]={"MFPrintEnumDualPolytope"};
  int k,d;
  int i,n;
  int j,m;
  MFNVector v;

  if(fout==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(EP==(MFEnumDualPolytope)NULL)
   {
    sprintf(MFPrintErrorMsg,"Enumerated Dual Polytope (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  k=MFEnumDualPolytopeDimension(EP);
  for(d=0;d<k+1;d++)
   {
    fprintf(fout," %d cells\n",d);
    n=MFEnumDualPolytopeNumberOfCells(EP,d);
    for(i=0;i<n;i++)
     {
      fprintf(fout,"  %2d  ",i);fflush(fout);
      fprintf(fout,"[");fflush(fout);
      m=MFEnumDualPolytopeNumberOfCellIndices(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout," ");
        fprintf(fout,"%2d",MFEnumDualPolytopeCellIndex(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,"]   (");
      m=MFEnumDualPolytopeNumberOfCellFaces(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout,",");
        fprintf(fout,"%2d",MFEnumDualPolytopeCellFace(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,") (");fflush(fout);
      m=MFEnumDualPolytopeNumberOfFaceCells(EP,d,i);
      for(j=0;j<m;j++)
       {
        if(j>0)fprintf(fout,",");
        fprintf(fout,"%2d",MFEnumDualPolytopeFaceCell(EP,d,i,j));fflush(fout);
       }
      fprintf(fout,") ");fflush(fout);
      v=MFEnumDualPolytopeVertex(EP,i);
      if(d==0&&v!=(MFNVector)NULL)MFPrintNVector(stdout,v);
      fprintf(fout,"\n");fflush(fout);
     }
   }

  return;
 }
