/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              October 6, 2000   Added Polyhedron    */

#include <stdio.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <stdlib.h>
#include <math.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNRegionErrorMsg[256]="";

double *MFNV_CStar(MFNVector);

int TPBVPTest(MFNVector,void*);
void TPBVPFree(void*);

MFNRegion MFNRegionCreateTPBVP(int nx,int nu,int np,double *p0,double *p1,double u0, double u1)
 {
  static char RoutineName[]={"MFNRegionCreateTPBVP"};
  MFNRegion tpbvp;
  int i;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateTPBVP++;
  starttime=clock();
#endif

  tpbvp=MFNRegionCreateBaseClass("TPBVP");
  MFNRegionSetTest(tpbvp,TPBVPTest);
  MFNRegionSetFreeData(tpbvp,TPBVPFree);

  data=(double*)malloc((5+2*np)*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",(5+2*np)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(tpbvp);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateTPBVP+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  ((int*)(data))[0]=nx;
  ((int*)(data))[1]=nu;
  ((int*)(data))[2]=np;
  for(i=0;i<np;i++)
   {
    ((double*)(data))[3+2*i]=p0[i];
    ((double*)(data))[3+2*i+1]=p1[i];
   }
  ((double*)(data))[3+2*np]=u0;
  ((double*)(data))[3+2*np+1]=u1;

  MFNRegionSetData(tpbvp,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateTPBVP+=clock()-starttime;
#endif
  return(tpbvp);
 }

int TPBVPTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"TPBVPTest"};
  double *d;
  double u;
  int rc;
  int n;
  double *x;
  int nx,nu,np;
  int i,j;
  static int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledTPBVPTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  d=(double*)data;
  nx=((int*)d)[0];
  nu=((int*)d)[1];
  np=((int*)d)[2];
  rc=1;
  if(verbose){printf("%s. nx=%d, nu=%d, np=%d\n",RoutineName,nx,nu,np);}
  for(i=0;i<np;i++)
   {
    rc=rc&&(x[nx*nu+i]>=d[3+2*i])&&(x[nx*nu+i]<=d[3+2*i+1]);
    if(verbose){printf("%d, %lf<=%lf<=%lf\n",i,d[3+2*i],x[nx*nu+i],d[3+2*i+1]);fflush(stdout);}
   }
  u=0;
  i=0;
  for(j=0;j<nu;j++)u+=(x[i*nu+j]+x[(i+1)*nu+j])*x[nx*nu+np+i]/2;
  for(i=0;i<nx-1;i++)
   {
    for(j=0;j<nu;j++)
     {
      u+=(x[i*nu+j]+x[(i+1)*nu+j])*(x[nx*nu+np+i+1]-x[nx*nu+np+i])/2;
     }
   }
  i=nx-1;
  for(j=0;j<nu;j++)u+=(x[i*nu+j]+x[(i+1)*nu+j])*x[nx*nu+np+i]/2;
  u=u/nu;
  rc=rc&&(u>=d[3+2*np])&&(u<=d[3+2*np+1]);
  if(verbose){printf("  %lf<=%lf<=%lf\n",d[3+2*np],u,d[3+2*np+1]);fflush(stdout);}

#ifdef MFTIMINGS
    MFTimeTPBVPTest+=clock()-starttime;
#endif
  return rc;
 }

void TPBVPFree(void *data)
 {
  static char RoutineName[]={"TPBVPFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledTPBVPFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimeTPBVPFree+=clock()-starttime;
#endif
  return;
 }
