/* 
    %W%
    %D% %T%
   

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

#define PI 3.14159265358979323846264338327950288

/*
u'=f(u,t,l);                     n equations
a(u(0),l)+b(u(1),l)+c(l)=0;    n equations
int l(u(t),t,l) dt + m(l) = 0;   nc equations
*/

void f(double *u,double *f)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;
  int nw=1;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  f[0]=T*x;
  f[1]=T*(Ic-g*x-sin(xi)*cos(eta+2*PI*nw*t)-2*k*xi);
  f[2]=T*y;
  f[3]=T*(I-g*y-sin(eta+2*PI*nw*t)*cos(xi))-g*2*PI*nw;
/*printf("t=%lf, u=[%lf,%lf,%lf,%lf], f=[%lf,%lf,%lf,%lf] T=%lf, g=%lf, k=%lf, I=%lf, Ic=%lf\n",t,xi,x,eta,y,f[0],f[1],f[2],f[3],T,g,k,I,Ic);*/
  return;
 }

void fu(double *u,double *fu)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;
  int nw=1;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  fu[0+4*0]=0.;
  fu[0+4*1]=T;
  fu[0+4*2]=0.;
  fu[0+4*3]=0.;
  fu[1+4*0]=-T*(cos(xi)*cos(eta+2*PI*nw*t)+2*k);
  fu[1+4*1]=-T*g;
  fu[1+4*2]=0.;
  fu[1+4*3]=0.;
  fu[2+4*0]=0.;
  fu[2+4*1]=0.;
  fu[2+4*2]=T;
  fu[2+4*3]=0.;
  fu[3+4*0]=0.;
  fu[3+4*1]=0.;
  fu[3+4*2]=-T*cos(eta+2*PI*nw*t)*cos(xi);
  fu[3+4*3]=-T*g;

  return;
 }

void fl(double *u,double *fl)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;
  int nw=1;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  fl[0+4*0]=x;
  fl[0+4*1]=0.;
  fl[0+4*2]=0.;
  fl[0+4*3]=0.;
  fl[0+4*4]=0.;

  fl[1+4*0]=Ic-g*x-sin(xi)*cos(eta+2*PI*nw*t)-2*k*xi;
  fl[1+4*1]=-T*x;
  fl[1+4*2]=-2*T*xi;
  fl[1+4*3]=0.;
  fl[1+4*4]=T;

  fl[2+4*0]=y;
  fl[2+4*1]=0.;
  fl[2+4*2]=0.;
  fl[2+4*3]=0.;
  fl[2+4*4]=0.;

  fl[3+4*0]=I-g*y-sin(eta+2*PI*nw*t)*cos(xi);
  fl[3+4*1]=-T*y-2*PI*nw;
  fl[3+4*2]=0.;
  fl[3+4*3]=T;
  fl[3+4*4]=0.;

  return;
 }

void a(double *u,double *a)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  a[0]=xi;
  a[1]=x;
  a[2]=eta;
  a[3]=y;

  return;
 }

void au(double *u,double *au)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  au[0+4*0]=1.;
  au[0+4*1]=0.;
  au[0+4*2]=0.;
  au[0+4*3]=0.;

  au[1+4*0]=0.;
  au[1+4*1]=1.;
  au[1+4*2]=0.;
  au[1+4*3]=0.;

  au[2+4*0]=0.;
  au[2+4*1]=0.;
  au[2+4*2]=1.;
  au[2+4*3]=0.;

  au[3+4*0]=0.;
  au[3+4*1]=0.;
  au[3+4*2]=0.;
  au[3+4*3]=1.;

  return;
 }

void al(double *u,double *al)
 {
  al[0+4*0]=0.;
  al[1+4*0]=0.;
  al[2+4*0]=0.;
  al[3+4*0]=0.;

  al[0+4*1]=0.;
  al[1+4*1]=0.;
  al[2+4*1]=0.;
  al[3+4*1]=0.;

  al[0+4*2]=0.;
  al[1+4*2]=0.;
  al[2+4*2]=0.;
  al[3+4*2]=0.;

  al[0+4*3]=0.;
  al[1+4*3]=0.;
  al[2+4*3]=0.;
  al[3+4*3]=0.;

  al[0+4*4]=0.;
  al[1+4*4]=0.;
  al[2+4*4]=0.;
  al[3+4*4]=0.;

  return;
 }

void b(double *u,double *b)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  b[0]=xi;
  b[1]=x;
  b[2]=eta;
  b[3]=y;

/*printf("u=[%lf,%lf,%lf,%lf], b=[%lf,%lf,%lf,%lf] T=%lf, g=%lf, k=%lf, I=%lf, Ic=%lf\n",xi,x,eta,y,b[0],b[1],b[2],b[3],T,g,k,I,Ic);*/

  return;
 }

void bu(double *u,double *bu)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  bu[0+4*0]=1.;
  bu[0+4*1]=0.;
  bu[0+4*2]=0.;
  bu[0+4*3]=0.;

  bu[1+4*0]=0.;
  bu[1+4*1]=1.;
  bu[1+4*2]=0.;
  bu[1+4*3]=0.;

  bu[2+4*0]=0.;
  bu[2+4*1]=0.;
  bu[2+4*2]=1.;
  bu[2+4*3]=0.;

  bu[3+4*0]=0.;
  bu[3+4*1]=0.;
  bu[3+4*2]=0.;
  bu[3+4*3]=1.;

  return;
 }

void bl(double *u,double *bl)
 {
  bl[0+4*0]=0.;
  bl[1+4*0]=0.;
  bl[2+4*0]=0.;
  bl[3+4*0]=0.;

  bl[0+4*1]=0.;
  bl[1+4*1]=0.;
  bl[2+4*1]=0.;
  bl[3+4*1]=0.;

  bl[0+4*2]=0.;
  bl[1+4*2]=0.;
  bl[2+4*2]=0.;
  bl[3+4*2]=0.;

  bl[0+4*3]=0.;
  bl[1+4*3]=0.;
  bl[2+4*3]=0.;
  bl[3+4*3]=0.;

  bl[0+4*4]=0.;
  bl[1+4*4]=0.;
  bl[2+4*4]=0.;
  bl[3+4*4]=0.;

  return;
 }

void c(double *x,double *c)
 {
  c[0]=0.;
  c[1]=0.;
  c[2]=0.;
  c[3]=0.;
  return;
 }

void cl(double *x,double *cl)
 {
  cl[0+4*0]=0.;
  cl[1+4*0]=0.;
  cl[2+4*0]=0.;
  cl[3+4*0]=0.;

  cl[0+4*1]=0.;
  cl[1+4*1]=0.;
  cl[2+4*1]=0.;
  cl[3+4*1]=0.;

  cl[0+4*2]=0.;
  cl[1+4*2]=0.;
  cl[2+4*2]=0.;
  cl[3+4*2]=0.;

  cl[0+4*3]=0.;
  cl[1+4*3]=0.;
  cl[2+4*3]=0.;
  cl[3+4*3]=0.;

  cl[0+4*4]=0.;
  cl[1+4*4]=0.;
  cl[2+4*4]=0.;
  cl[3+4*4]=0.;

  return;
 }

void l(double *u,double *l)
 {
  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  xi =u[0];
  x  =u[1];
  eta=u[2];
  y  =u[3];
  t  =u[4];
  T =u[5];
  g =u[6];
  k =u[7];
  I =u[8];
  Ic=u[9];

  l[0]=xi+eta;
  return;
 }

void lu(double *u,double *lu)
 {
  int i;

  lu[0]=1.;
  lu[1]=0.;
  lu[2]=1.;
  lu[3]=0.;

  return;
 }

void ll(double *u,double *ll)
 {
  int i;

  ll[0]=0.;
  ll[1]=0.;
  ll[2]=0.;
  ll[3]=0.;
  ll[4]=0.;

  return;
 }

void m(double *u,double *m)
 {
  int i;

  m[0]=0.;

  return;
 }

void ml(double *u,double *ml)
 {
  int i;

  ml[0]=0.;
  ml[1]=0.;
  ml[2]=0.;
  ml[3]=0.;
  ml[4]=0.;

  return;
 }

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j,n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFNVector ug;
  MFNKMatrix Tan;
  MFNVector dug[4];
  MFNKMatrix Tan0;
  int nw=1;

  double T,g,k,Ic,I;
  double t,x,y,xi,eta;

  int nx=7;
  int np=5;
  int nu=4;
  int nc=1;

  M=MFIMFCreateTPBVP(nx,nu,np,nc,f,fu,fl,a,au,al,b,bu,bl,c,cl,l,lu,ll,m,ml);
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,100.);

  ug=MFCreateNVector(n);
  dug[0]=MFCreateNVector(n);
  dug[1]=MFCreateNVector(n);
  dug[2]=MFCreateNVector(n);
  dug[3]=MFCreateNVector(n);

  g=.5;
  k =.1;
  I =100.;
  Ic=0.;
  T=g*2*PI*nw/I;

  for(i=0;i<nx;i++)
   {
    t=(i-.5)/(nx-2);

    MFNVSetC( ug,nu*nx+np+i,t);
/*  printf("x[%d]=%lf\n",i,t);*/

    MFNVSetC( ug,nu*i+0,0.);  /* xi [i] */
    MFNVSetC( ug,nu*i+1,0.);  /* x  [i] */
    MFNVSetC( ug,nu*i+2,0.);  /* eta[i] */
    MFNVSetC( ug,nu*i+3,0.);  /* y[i] */

    for(j=0;j<4;j++)
     {
      MFNVSetC(dug[j],nu*nx+np+i,t);
      MFNVSetC(dug[j],nu*i+0,.1); /* dxi [i] */
      MFNVSetC(dug[j],nu*i+1,.1); /* dx  [i] */
      MFNVSetC(dug[j],nu*i+2,-.1); /* deta[i] */
      MFNVSetC(dug[j],nu*i+3,.1); /* dy  [i] */
     }
   }

  t=(nx-.5)/(nx-2);
  MFNVSetC( ug,nu*nx+np+nx,t);
  for(j=0;j<4;j++)MFNVSetC(dug[j],nu*nx+np+nx,t);
/*printf("x[%d]=%lf\n",nx,t);*/

  MFNVSetC( ug,nu*nx  ,T);
  MFNVSetC( ug,nu*nx+1,g);
  MFNVSetC( ug,nu*nx+2,k);
  MFNVSetC( ug,nu*nx+3,I);
  MFNVSetC( ug,nu*nx+4,Ic);

  MFNVSetC(dug[3],nu*nx  ,0.);
  MFNVSetC(dug[3],nu*nx+1,0.);
  MFNVSetC(dug[3],nu*nx+2,0.);
  MFNVSetC(dug[3],nu*nx+3,0.);
  MFNVSetC(dug[3],nu*nx+4,1.);

  MFNVSetC(dug[2],nu*nx  ,-2*g*PI*nw/I/I);
  MFNVSetC(dug[2],nu*nx+1,0.);
  MFNVSetC(dug[2],nu*nx+2,0.);
  MFNVSetC(dug[2],nu*nx+3,1.);
  MFNVSetC(dug[2],nu*nx+4,0.);

  MFNVSetC(dug[1],nu*nx  ,0.);
  MFNVSetC(dug[1],nu*nx+1,0.);
  MFNVSetC(dug[1],nu*nx+2,1.);
  MFNVSetC(dug[1],nu*nx+3,0.);
  MFNVSetC(dug[1],nu*nx+4,0.);

  MFNVSetC(dug[0],nu*nx  ,2*PI*nw/I);
  MFNVSetC(dug[0],nu*nx+1,1.);
  MFNVSetC(dug[0],nu*nx+2,0.);
  MFNVSetC(dug[0],nu*nx+3,0.);
  MFNVSetC(dug[0],nu*nx+4,0.);

  Tan0=MFCreateNKMatrix(MFIMF_K(M),dug);
  Tan=MFIMFTangentSpaceWithGuess(M,ug,Tan0);
  u0=MFCreateNVector(n);
  MFIMFProject(M,ug,Tan,u0);
  MFFreeNVector(ug);

  S=MFComputeAtlas(M,Omega,u0,.1,1000,1);

  fid=fopen("PendulaTPBVP.atlas","w");
  MFWriteAtlas(fid,S);

  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
